package edu.wustl.cse.mobilab.openworkflow.auction;

import edu.wustl.cse.mobilab.openworkflow.comm.AuctionQueryable;
import edu.wustl.cse.mobilab.openworkflow.comm.HostId;
import edu.wustl.cse.mobilab.openworkflow.schedule.Time;
import edu.wustl.cse.mobilab.openworkflow.shell.SystemTimeProvider;
import edu.wustl.cse.mobilab.openworkflow.util.TimeNotifier;
import edu.wustl.cse.mobilab.openworkflow.verify.Assert;
import edu.wustl.cse.mobilab.openworkflow.verify.Require;
import edu.wustl.cse.mobilab.openworkflow.workflow.Task;
import org.apache.log4j.Logger;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.concurrent.Executor;

//--------------------------------------------------------------------
/** (description of class)
 * @author Created by Louis Thomas on Nov 6, 2008. */
public class AuctionManagerImpl implements AuctionManager {

    private static final boolean REJECTED = true;
    private static final boolean CONFIRMED = false;

    private final AuctionQueryable m_auctionQueryable;
    private final TaskBidComparator m_taskBidComparator;
    private final SystemTimeProvider m_systemTimeProvider;
    private final Logger m_logger;

    private final Map<Task, TaskDetail> m_taskToTaskDetailMap = new LinkedHashMap<Task, TaskDetail>();
    private final TimeNotifier m_timeNotifier;

    //----------------------------------------------------------------
    public AuctionManagerImpl(AuctionQueryable auctionQueryable, TaskBidComparator taskBidComparator, SystemTimeProvider systemTimeProvider, TimeNotifier timeNotifier, Logger logger) {
        Require.neqNull(auctionQueryable, "auctionQueryable");
        Require.neqNull(taskBidComparator, "taskBidComparator");
        Require.neqNull(systemTimeProvider, "systemTimeProvider");
        Require.neqNull(timeNotifier, "timeNotifier");
        Require.neqNull(logger, "logger");
        m_auctionQueryable = auctionQueryable;
        m_taskBidComparator = taskBidComparator;
        m_systemTimeProvider = systemTimeProvider;
        m_timeNotifier = timeNotifier;
        m_logger = logger;
    }

    //----------------------------------------------------------------
    // from AuctionManager
    public void dispose() {
        // nothing to do yet
    }

    //----------------------------------------------------------------
    // from AuctionManager
    public void allocateTask(Task task) {
        Require.neqNull(task, "task");
        Require.eqFalse(m_taskToTaskDetailMap.containsKey(task), "m_taskToTaskDetailMap.containsKey(task)");

        m_logger.info("AuctionManagerImpl: Task added.");

        TaskDetail taskDetail = new TaskDetail(task);
        m_taskToTaskDetailMap.put(task, taskDetail);
        m_auctionQueryable.startSolicitingBids(task, taskDetail);
    }

    //----------------------------------------------------------------
    private void allocateBids(Task task) {
        TaskDetail taskDetail = m_taskToTaskDetailMap.get(task);
        BidDetail bid = (BidDetail) m_taskBidComparator.selectBestBid(taskDetail.getTaskBids());
        if (bid != null) {
            // tentativeAllocation is the previous temporary allocation
            TaskBid tentativeAllocation = taskDetail.getTentativeAllocation();

            // If we didn't have a previous allocation, we simply allocate bid.
            if (tentativeAllocation == null) {
                doTentativeAllocate(task, bid, null);

            } else if (bid != tentativeAllocation) {
                // If we did have an old tentativeAllocation, we see which is better and update appropriately.
                // In either case, we reject the loser bid immediately.
                TaskBid winner = m_taskBidComparator.compare(bid, tentativeAllocation) < 0 ? bid : tentativeAllocation;
                if (winner != bid) {
                    taskDetail.getTaskBids().remove(bid);
                    notifyHost(bid, REJECTED);
                } else {
                    TaskBid loser = winner == bid ? tentativeAllocation : bid;
                    doTentativeAllocate(task, winner, loser);
                    taskDetail.getTaskBids().remove(loser);
                    notifyHost(loser, REJECTED);
                }
            }
        }

        // Here we notify the host or merely put the task back as NO_ATTENTION
        if (shouldCommitTentativeAllocation(task, taskDetail.getTentativeAllocation())) {
            notifyHost(taskDetail.getTentativeAllocation(), CONFIRMED);
            deleteTask(task);
        } else {
            scheduleFutureNotification(bid, bid.getDeadline());
        }
    }

    //----------------------------------------------------------------
    /** This method determines whether the tentative allocation for
     * task t should be made final. The exact behavior of this method
     * needs to be ironed out carefully.
     * @param t the task
     * @param taskBid its best bid
     * @return true iff we should finalize the allocation */
    private boolean shouldCommitTentativeAllocation(Task t, TaskBid taskBid) {
        return true;
        //return taskBid.getDeadline().isAfter(m_systemTimeProvider.getSystemTime());
    }

    //----------------------------------------------------------------
    /** In CiAN this method updates conflicts to handle the tentative
     * allocation. We now insist that all potential conflicts are
     * marked by the bidder, so this method has little to do. We may
     * still want the deadline code below though. */
    private void doTentativeAllocate(Task t, TaskBid winner, TaskBid loser) {
        m_logger.debug("AuctionManagerImpl: Tentatively allocating to " + winner.getHostId() + " task " + t + ".");
        m_taskToTaskDetailMap.get(t).setTentativeAllocation(winner);
    }

    //----------------------------------------------------------------
    /** Removes this task from the system and rejects all bids for
     * this task which did not win an allocation. This can be used to
     * shut down the system or to remove a task when its allocation is
     * official. */
    private void deleteTask(Task task) {
        m_logger.info("AuctionManagerImpl: Deleting task " + task + ".");

        AuctionManagerImpl.TaskDetail taskDetail = m_taskToTaskDetailMap.remove(task);
        // reject all non-allocated bids for this task
        for (TaskBid taskBid : taskDetail.getTaskBids()) {
            if (taskBid == taskDetail.getTentativeAllocation()) {
                continue;
            }
            notifyHost(taskBid, REJECTED);
        }

        m_auctionQueryable.stopSolicitingBids(task, taskDetail);
    }

    //----------------------------------------------------------------
    private void notifyHost(TaskBid taskBid, boolean isRejection) {
        BidDetail bidDetail = (BidDetail) taskBid;
        if (isRejection) {
            m_logger.info("AuctionManagerImpl: Officially rejecting worker " + bidDetail.getHostId() + "'s bid on task " + bidDetail.getTask() + ".");
            bidDetail.getBidder().rejectBid(bidDetail.getTaskBid());
        } else {
            m_logger.info("AuctionManagerImpl: Officially allocating worker " + bidDetail.getHostId() + " task " + bidDetail.getTask() + ".");
            bidDetail.getBidder().confirmBid(bidDetail.getTaskBid());
        }
        cancelFutureNotification(bidDetail);
    }

    //----------------------------------------------------------------
    private void scheduleFutureNotification(final BidDetail bidDetail, Time time) {
        TimeNotifier.Listener timeNotifierListener = new TimeNotifier.Listener() {
            public void notifyTimeReached(Time time) {
                cancelFutureNotification(bidDetail);
                allocateBids(bidDetail.getTask());
            }
        };
        bidDetail.setTimeListener(timeNotifierListener);
        bidDetail.setTimeListenerTime(time);
        m_timeNotifier.subscribeTimeNotification(time, timeNotifierListener);
    }

    //----------------------------------------------------------------
    private void cancelFutureNotification(BidDetail bidDetail) {
        if (bidDetail.getTimeListener() != null) {
            m_timeNotifier.unsubscribeTimeNotification(bidDetail.getTimeListenerTime(), bidDetail.getTimeListener());
        }
        bidDetail.setTimeListener(null);
        bidDetail.setTimeListenerTime(null);
    }

    //################################################################

    //----------------------------------------------------------------
    private class TaskDetail implements AuctionQueryable.Listener {

        private final Task m_task;
        private final SortedSet<TaskBid> m_taskBids;
        private TaskBid m_tentativeAllocation;

        private TaskDetail(Task task) {
            m_task = task;
            m_taskBids = new TreeSet<TaskBid>(m_taskBidComparator);
        }

        public Task getTask() {
            return m_task;
        }

        public SortedSet<TaskBid> getTaskBids() {
            return m_taskBids;
        }

        public TaskBid getTentativeAllocation() {
            return m_tentativeAllocation;
        }

        public void setTentativeAllocation(TaskBid tentativeAllocation) {
            m_tentativeAllocation = tentativeAllocation;
        }

        // from AuctionQueryable.Listener
        public void handleBid(AuctionQueryable.Bidder bidder, TaskBid taskBid) {
            Require.neqNull(bidder, "bidder");
            Require.neqNull(taskBid, "taskBid");
            Require.equals(taskBid.getTask(), "taskBid.getTask()", m_task, "m_task");

            // This implementation assumes that workers will never submit
            // more than one TaskBid per Task. It is up to the workers to
            // check this. A worker may, however, re-bid a task which he is
            // rejected from.

            // This implementation further assumes that it is always okay to
            // allocate a pair of taskbids A and B from the same worker. It
            // is up to the worker to check that none of its taskbids conflict.

            m_logger.debug("AuctionManagerImpl: Received bid " + taskBid + ".");

            BidDetail bidDetail = new BidDetail(bidder, taskBid);
            // Is this bid late for an old task?
            if (!m_taskToTaskDetailMap.containsKey(m_task)) {
                notifyHost(bidDetail, REJECTED);
                return;
            }
            // Record bid

            boolean bWasAdded = m_taskBids.add(bidDetail);
            Assert.eqTrue(bWasAdded, "bWasAdded");
            m_taskToTaskDetailMap.get(m_task).m_taskBids.add(taskBid);
            // Alert the allocator that there is work to do
            scheduleFutureNotification(bidDetail, m_systemTimeProvider.getSystemTime());
        }
    }

    //----------------------------------------------------------------
    private static class BidDetail implements TaskBid {

        private final AuctionQueryable.Bidder m_bidder;
        private final TaskBid m_taskBid;

        // Ptr to our timeListener in case we need to cancel this allocation.
        private TimeNotifier.Listener m_timeNotifierListener;
        private Time m_timeListenerTime;

        private BidDetail(AuctionQueryable.Bidder bidder, TaskBid taskBid) {
            m_bidder = bidder;
            m_taskBid = taskBid;
        }

        public AuctionQueryable.Bidder getBidder() {
            return m_bidder;
        }

        public TaskBid getTaskBid() {
            return m_taskBid;
        }

        public TimeNotifier.Listener getTimeListener() {
            return m_timeNotifierListener;
        }

        public Time getTimeListenerTime() {
            return m_timeListenerTime;
        }

        public void setTimeListener(TimeNotifier.Listener timeNotifierListener) {
            m_timeNotifierListener = timeNotifierListener;
        }

        public void setTimeListenerTime(Time listenerTime) {
            m_timeListenerTime = listenerTime;
        }

        // from TaskBid
        public Task getTask() {
            return m_taskBid.getTask();
        }

        // from TaskBid
        public double getCapabilityFraction() {
            return m_taskBid.getCapabilityFraction();
        }

        // from TaskBid
        public Time getDeadline() {
            return m_taskBid.getDeadline();
        }

        // from TaskBid
        public HostId getHostId() {
            return m_taskBid.getHostId();
        }
    }
}
