package edu.wustl.cse.mobilab.openworkflow.auction;

import edu.wustl.cse.mobilab.openworkflow.comm.HostId;
import edu.wustl.cse.mobilab.openworkflow.schedule.CommitmentImpl;
import edu.wustl.cse.mobilab.openworkflow.schedule.Duration;
import edu.wustl.cse.mobilab.openworkflow.schedule.DurationImpl;
import edu.wustl.cse.mobilab.openworkflow.schedule.Location;
import edu.wustl.cse.mobilab.openworkflow.schedule.ScheduleManager;
import edu.wustl.cse.mobilab.openworkflow.schedule.ScheduleManager.Commitment;
import edu.wustl.cse.mobilab.openworkflow.schedule.ScheduleManager.CommitmentListener;
import edu.wustl.cse.mobilab.openworkflow.schedule.ServiceInvocationDescriptor;
import edu.wustl.cse.mobilab.openworkflow.schedule.Time;
import edu.wustl.cse.mobilab.openworkflow.schedule.TimePeriod;
import edu.wustl.cse.mobilab.openworkflow.schedule.TimePeriodImpl;
import edu.wustl.cse.mobilab.openworkflow.service.ServiceDescriptor;
import edu.wustl.cse.mobilab.openworkflow.service.ServiceManager;
import edu.wustl.cse.mobilab.openworkflow.shell.SystemTimeProvider;
import edu.wustl.cse.mobilab.openworkflow.verify.Require;
import edu.wustl.cse.mobilab.openworkflow.workflow.Task;
import org.apache.log4j.Logger;

import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.PriorityQueue;

//--------------------------------------------------------------------
/** (description of class)
 * @author Created by Louis Thomas on Nov 6, 2008. */
public class AuctionParticipationManagerImpl implements AuctionParticipationManager {

    private static final Duration MAX_BID_WAIT_TIME = new DurationImpl(5000);
    private static final int MAX_REBID_TRIES = 10;

    private final ScheduleManager m_scheduleManager;
    private final ServiceManager m_serviceManager;
    private final SystemTimeProvider m_systemTimeProvider;
    private final Logger m_logger;
    private final HostId m_hostId;
    private final Map<Task, TaskSolicitationDetail> m_solicitedTasks = new HashMap<Task, TaskSolicitationDetail>();

    private final PriorityQueue<Time> m_bidWindowStarts;
    private final PriorityQueue<Time> m_bidWindowEnds;
    private Commitment m_masterBidWindow;

    //----------------------------------------------------------------
    public AuctionParticipationManagerImpl(ScheduleManager scheduleManager, ServiceManager serviceManager, SystemTimeProvider systemTimeProvider, Logger logger, HostId hostId) {
        Require.neqNull(scheduleManager, "scheduleManager");
        Require.neqNull(serviceManager, "serviceManager");
        Require.neqNull(systemTimeProvider, "systemTimeProvider");
        Require.neqNull(logger, "logger");
        Require.neqNull(hostId, "hostId");
        m_scheduleManager = scheduleManager;
        m_serviceManager = serviceManager;
        m_systemTimeProvider = systemTimeProvider;
        m_logger = logger;
        m_hostId = hostId;
        m_bidWindowStarts = new PriorityQueue<Time>(11, new Comparator<Time> () {
            public int compare(Time o1, Time o2) {
                return o1.compareTo(o2);
            }
        });
        m_bidWindowEnds = new PriorityQueue<Time>(11, new Comparator<Time> () {
            public int compare(Time o1, Time o2) {
                return -o1.compareTo(o2);
            }
        });
    }

    //----------------------------------------------------------------
    public void dispose() {
        // nothing to do yet
    }

    //----------------------------------------------------------------
    // from AuctionQueryable
    public void startSolicitingBids(Task task, Listener listener) {
        // It is safe to call contains(Task) even though m_solicitedTasks is a set of
        // TaskSolicitationDetails because of the equals method on TaskSolicitationDetail.
        Require.neqNull(task, "task");
        Require.neqNull(listener, "listener");
        Require.eqFalse(m_solicitedTasks.containsKey(task), "m_solicitedTasks.containsKey(task)");

        // Bids are being solicited from this host
        m_logger.debug("AuctionParticipationManagerImpl: Received startSolicitingBids request for " + task + ".");
        TaskSolicitationDetail solicitationDetail = new TaskSolicitationDetail(task, listener);
        solicitationDetail.m_acceptingBids = true;
        m_solicitedTasks.put(task, solicitationDetail);
        solicitationDetail.bidOnTask();
    }

    //----------------------------------------------------------------
    // from AuctionQueryable
    public void stopSolicitingBids(Task task, Listener listener) {
        // It is safe to call contains(Task) even though m_solicitedTasks is a set of
        // TaskSolicitationDetails because of the equals method on TaskSolicitationDetail.
        Require.eqTrue(m_solicitedTasks.containsKey(task), "m_solicitedTasks.containsKey(task)");

        // Bids are no longer being solicited from this host
        m_logger.debug("AuctionParticipationManagerImpl: Received stopSolicitingBids request for " + task + ".");
        m_solicitedTasks.remove(task);
    }

    /*
     * Expands the master bid window to include the new bid window if possible.
     */
    private boolean attemptToExpandBidWindow(Commitment newBidWindow) {
        if (doBidWindowExpansion(newBidWindow)) {
            m_bidWindowStarts.add(newBidWindow.getTimePeriod().getStartTime());
            m_bidWindowEnds.add(newBidWindow.getTimePeriod().getEndTime());
            return true;
        }
        return false;
    }

    private boolean doBidWindowExpansion(Commitment newBidWindow) {
        Require.neqNull(newBidWindow, "newBidWindow");
        if (null == m_masterBidWindow) {
            if (m_scheduleManager.addToSchedule(newBidWindow)) {
                m_masterBidWindow = newBidWindow;
                return true;
            }
            return false;
        }

        Time newBidWindowStart = newBidWindow.getTimePeriod().getStartTime();
        Time newBidWindowEnd = newBidWindow.getTimePeriod().getEndTime();
        Time masterWindowStart = m_masterBidWindow.getTimePeriod().getStartTime();
        Time masterWindowEnd = m_masterBidWindow.getTimePeriod().getEndTime();

        if (masterWindowStart.compareTo(newBidWindowStart) <= 0 &&
            masterWindowEnd.compareTo(newBidWindowEnd) >= 0) {
            // The master window already covers the new window.
            return true;
        }

        Time unionStart = newBidWindowStart.compareTo(masterWindowStart) < 0 ?
                newBidWindowStart : masterWindowStart;
        Time unionEnd = newBidWindowEnd.compareTo(masterWindowEnd) > 0 ?
                newBidWindowEnd : masterWindowEnd;
        TimePeriod unionPeriod = new TimePeriodImpl(unionStart, unionEnd);
        Commitment unionWindow = new CommitmentImpl(unionPeriod, m_masterBidWindow.getLocation());

        if (m_scheduleManager.replace(m_masterBidWindow, unionWindow)) {
            m_masterBidWindow = unionWindow;
            return true;
        }
        return false;
    }

    /*
     * Removes oldBidWindow from consideration. This shrinks the master
     * if the oldBidWindow was extreme in either direction.
     */
    private void removeBidWindow(Commitment oldBidWindow) {
        Time oldBidWindowStart = oldBidWindow.getTimePeriod().getStartTime();
        Time oldBidWindowEnd = oldBidWindow.getTimePeriod().getEndTime();
        Require.eqTrue(m_bidWindowStarts.remove(oldBidWindowStart), "m_bidWindowStarts.remove(oldBidWindowStart)");
        Require.eqTrue(m_bidWindowEnds.remove(oldBidWindowEnd), "m_bidWindowEnds.remove(oldBidWindowEnd)");

        if (m_bidWindowStarts.isEmpty()) {
            m_scheduleManager.removeFromSchedule(m_masterBidWindow);
            m_masterBidWindow = null;
        } else {
            TimePeriod newWindowPeriod = new TimePeriodImpl(m_bidWindowStarts.peek(), m_bidWindowEnds.peek());
            Commitment newWindow = new CommitmentImpl(newWindowPeriod, m_masterBidWindow.getLocation());
            if (!newWindow.equals(m_masterBidWindow)) {
                // This cannot fail because the new window must be smaller than
                // the old master window.
                Require.eqTrue(m_scheduleManager.replace(m_masterBidWindow, newWindow),
                        "m_scheduleManager.replace(m_masterBidWindow, newWindow)");
                m_masterBidWindow = newWindow;
            }
        }
    }

    private class TaskSolicitationDetail implements CommitmentListener {

        final Task m_task;
        final Listener m_listener;
        TaskBid m_outstandingBid;
        boolean m_acceptingBids=true;
        boolean m_bidConfirmed;
        boolean m_bidRejected;
        int m_nRetries;
        ScheduleManager.Commitment m_individualBidWindow;

        public TaskSolicitationDetail(Task task, Listener listener) {
            m_task=task;
            m_listener=listener;
            m_scheduleManager.subscribeCommitments(this);
        }

        private void bidOnTask() {
            if (m_nRetries++>MAX_REBID_TRIES || !m_acceptingBids
                || m_outstandingBid!=null || m_bidRejected)
            {
                return;
            }
            ServiceDescriptor serviceDescriptor=m_serviceManager.getImplementingService(m_task);
            if (null==serviceDescriptor) {
                m_logger.debug("AuctionParticipationManagerImpl: No implementing service available for "+m_task+".");
                return;
            }
            ServiceInvocationDescriptor serviceInvocationDescriptor=new ServiceInvocationDescriptor(serviceDescriptor, m_task);

            final Commitment serviceCommitment=m_scheduleManager.addToSchedule(serviceInvocationDescriptor);
            if (null==serviceCommitment) {
                // we cannot fit this task into the schedule
                m_logger.info("AuctionParticipationManagerImpl: Cannot bid on task "+m_task+" due to conflict.");
                return;
            }

            double capabilityFraction=m_serviceManager.getCapabilityFraction(m_task);
            Time currentTime=m_systemTimeProvider.getSystemTime();
            Location currentLocation=m_scheduleManager.getLocation(currentTime);

            Time deadline=getDeadlineForBidding(currentTime, currentLocation);
            if (deadline==null || deadline.minus(currentTime).compareTo(MAX_BID_WAIT_TIME)>0) {
                deadline=currentTime.plus(MAX_BID_WAIT_TIME);
            }
            m_individualBidWindow=new CommitmentImpl(currentTime.periodToTime(deadline), currentLocation);
            if (!attemptToExpandBidWindow(m_individualBidWindow)) {
                m_individualBidWindow=null;
                m_scheduleManager.removeFromSchedule(serviceCommitment);
                m_logger.info("AuctionParticipationManagerImpl: Cannot bid on task "+m_task+" due to bid window conflict.");
                return;
            }

            TaskBid bid=new TaskBidImpl(m_task, m_hostId, capabilityFraction, deadline);

            //submit the bid
            m_outstandingBid=bid;
            m_listener.handleBid(new Bidder() {

                public void confirmBid(TaskBid taskBid) {
                    Require.eqFalse(m_bidRejected, "m_bidRejected");
                    Require.eqFalse(m_bidConfirmed, "m_bidConfirmed");
                    m_logger.info("AuctionParticipationManagerImpl: Bid confirmed for "+taskBid+".");
                    m_bidConfirmed=true;
                    removeBidWindow(m_individualBidWindow);
                    m_individualBidWindow=null;
                    m_scheduleManager.unsubscribeCommitments(TaskSolicitationDetail.this);
                }

                public void rejectBid(TaskBid taskBid) {
                    Require.eqFalse(m_bidRejected, "m_bidRejected");
                    Require.eqFalse(m_bidConfirmed, "m_bidConfirmed");
                    m_logger.info("AuctionParticipationManagerImpl: Bid rejected for "+taskBid+".");

                    m_outstandingBid=null;
                    m_bidRejected=true;

                    // as we made a firm commitment by adding this bid to our schedule, we must now remove it from our schedule
                    m_scheduleManager.removeFromSchedule(serviceCommitment);
                    removeBidWindow(m_individualBidWindow);
                    m_individualBidWindow=null;
                    m_scheduleManager.unsubscribeCommitments(TaskSolicitationDetail.this);
                }
            }, bid);

            m_logger.info("AuctionParticipationManagerImpl: Bidding on task "+m_task+".");
        }

        /*
         * Hackery to make sure that we don't conflict with our own bid window.
         * We remove the bid window and then add it back again. This is dirty
         * but seems better than making the ScheduleManager aware of different
         * Commitment types. If we end up with many different types of Commitments
         * with rich interaction, consider revising this method to let the
         * loitering window computation ignore the masterBidWindow.
         */
        private Time getDeadlineForBidding(Time currentTime, Location currentLocation) {
            if (null != m_masterBidWindow) {
                m_scheduleManager.removeFromSchedule(m_masterBidWindow);
            }
            Time deadline = m_scheduleManager.getEndOfLoiteringWindow(currentTime, currentLocation);
            if (null != m_masterBidWindow) {
                // Revert the schedule to include the bid window again. This cannot fail.
                Require.eqTrue(m_scheduleManager.addToSchedule(m_masterBidWindow),"m_scheduleManager.addToSchedule(m_masterBidWindow)");
            }
            return deadline;
        }

        public void updateCommitment(Commitment oldCommitment, Commitment newCommitment) {
            if (newCommitment != null) {
                return;
            }
            bidOnTask();
        }
    }
}
