package edu.wustl.cse.mobilab.openworkflow.comm;

import edu.wustl.cse.mobilab.openworkflow.verify.Assert;
import edu.wustl.cse.mobilab.openworkflow.verify.Require;
import edu.wustl.cse.mobilab.openworkflow.workflow.ConfigurationNode;
import lights.adapters.Tuple;
import lights.interfaces.ITuple;
import lime.AgentID;
import lime.AgentLocation;
import lime.LimeException;
import lime.LimeTupleSpace;
import lime.NoSuchReactionException;
import lime.Reaction;
import lime.ReactionEvent;
import lime.ReactionListener;
import lime.RegisteredReaction;
import lime.StationaryAgent;
import lime.TupleSpaceEngineException;
import lime.UbiquitousReaction;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Executor;
import java.util.concurrent.LinkedBlockingQueue;

public class FragmentQueryableAgent extends StationaryAgent implements ReactionListener {

    private Executor m_executor;

    enum State {

        UNINITIALIZED, READY, DISPOSED
    }
    private static final long serialVersionUID=1;

    private ConfigurationNode m_configurationNode;
    private FragmentQueryable.Listener m_listener;
    private State m_state=State.UNINITIALIZED;
    private LimeTupleSpace m_lts;
    private RegisteredReaction[] m_reactions;

    private final BlockingQueue<Runnable> m_runnables=new LinkedBlockingQueue<Runnable>();

    private static final Runnable SHUTDOWN=new Runnable() {
        public void run() {
            Assert.statementNeverExecuted();
        }
    };

    //----------------------------------------------------------------
    public void run() {

        try {
            m_lts=new LimeTupleSpace("OpenWorkflow");
            m_lts.setShared(true);
        } catch (TupleSpaceEngineException e) {
            Assert.exceptionNeverCaught(e);
        }

        while (true) {
            Runnable runnable=null;
            try {
                runnable=m_runnables.take();
            } catch (InterruptedException e) {
                Assert.exceptionNeverCaught(e);
            }
            if (SHUTDOWN==runnable) {
                break;
            }
            runnable.run();
        }

        // Unregister reaction.
        try {
            m_lts.removeWeakReaction(m_reactions);
        } catch (TupleSpaceEngineException e) {
            Assert.exceptionNeverCaught(e);
        } catch (NoSuchReactionException e) {
            Assert.exceptionNeverCaught(e);
        }

    }

    //----------------------------------------------------------------
    public void reactsTo(final ReactionEvent event) {
        m_runnables.add(new Runnable() {
            public void run() {
                ITuple tuple=event.getEventTuple();
                AgentID agentID=event.getSourceAgent();
                Long idx=(Long)tuple.get(1).getValue();
                ITuple template=new Tuple().addActual(idx).addFormal(CompositeWorkflowWrapper.class);
                try {
                    ITuple workflowTuple=m_lts.rdp(new AgentLocation(agentID), AgentLocation.UNSPECIFIED, template);
                    final CompositeWorkflowWrapper wrapper=(CompositeWorkflowWrapper)workflowTuple.get(1).getValue();
                    m_executor.execute(new Runnable() {
                        public void run() {
                            m_listener.handleFragment(wrapper.getCompositeWorkflow());
                        }
                    });
                } catch (LimeException le) {
                    System.err.println("Couldn't react.");
                    System.err.println(le.getMessage());
                }
            }
        });
    }

    //----------------------------------------------------------------
    public void subscribe(FragmentQueryable.Listener listener, ConfigurationNode configurationNode, Executor executor) {
        Require.neqNull(listener, "listener");
        Require.neqNull(configurationNode, "configurationNode");
        Require.neqNull(executor, "executor");
        Require.eq(m_state, "m_state", State.UNINITIALIZED, "State.UNINITIALIZED");
        m_state=State.READY;
        m_listener=listener;
        m_executor=executor;
        m_configurationNode=configurationNode;
        m_runnables.add(new Runnable() {
            public void run() {
                try {
                    // Register a reaction.
                    ITuple template=new Tuple().addActual(new ConfigurationNodeWrapper(m_configurationNode)).addFormal(Long.class);
                    m_reactions=m_lts.addWeakReaction(new Reaction[] {new UbiquitousReaction(template, FragmentQueryableAgent.this, Reaction.ONCEPERTUPLE)});
                } catch (TupleSpaceEngineException e) {
                    Assert.exceptionNeverCaught(e);
                }
            }
        });
    }

    //----------------------------------------------------------------
    public void unsubscribe() {
        Require.eq(m_state, "m_state", State.READY, "State.READY");
        m_state=State.DISPOSED;
        m_runnables.add(SHUTDOWN);
    }
}
