package edu.wustl.cse.mobilab.openworkflow.comm;

import edu.wustl.cse.mobilab.openworkflow.auction.TaskBid;
import edu.wustl.cse.mobilab.openworkflow.shell.Disposable;
import edu.wustl.cse.mobilab.openworkflow.verify.Assert;
import edu.wustl.cse.mobilab.openworkflow.verify.Require;
import edu.wustl.cse.mobilab.openworkflow.verify.RequirementFailure;
import edu.wustl.cse.mobilab.openworkflow.workflow.CompositeWorkflow;
import edu.wustl.cse.mobilab.openworkflow.workflow.ConfigurationNode;
import edu.wustl.cse.mobilab.openworkflow.workflow.ExecutionDataName;
import edu.wustl.cse.mobilab.openworkflow.workflow.Task;
import edu.wustl.cse.mobilab.openworkflow.workflow.TaskNode;
import lime.LimeException;
import lime.LimeServer;
import org.apache.log4j.Logger;

import java.io.Serializable;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Executor;

//--------------------------------------------------------------------
/** (description of class)
* @author Created by Louis Thomas on Feb 16, 2009. */
public class SingleProcessCommProxy implements CommProxy, Disposable {

    private final boolean m_bUseLime;

    private final Object m_stateLock=new Object();
    private final Executor m_executor;
    private FragmentQueryableSupport m_fragmentQueryableSupport;
    private ExecutionDataQueryableSupport m_executionDataQueryableSupport;
    private Logger m_logger;

    private final TaskFeasibilityQueryable m_taskFeasibilityQueryable=new TaskFeasibilityQueryable() {
        public void subscribeTaskFeasibility(TaskNode taskNode, Listener listener) {
            Assert.statementNeverExecuted();
        }

        public void unsubscribeTaskFeasibility(TaskNode taskNode, Listener listener) {
            Assert.statementNeverExecuted();
        }
    };

    private final AuctionQueryable m_auctionQueryable=new AuctionQueryable() {
        public void startSolicitingBids(Task task, Listener listener) {
            try {
                SingleProcessCommProxy.this.startSolicitingBids(task, listener);
            } catch (RequirementFailure requirementFailure) {
                throw requirementFailure.adjustForDelegatingMethodAndSyntheticAccessor();
            }
        }

        public void stopSolicitingBids(Task task, Listener listener) {
            try {
                SingleProcessCommProxy.this.stopSolicitingBids(task, listener);
            } catch (RequirementFailure requirementFailure) {
                throw requirementFailure.adjustForDelegatingMethodAndSyntheticAccessor();
            }
        }
    };

    private final Set<TaskFeasibilityQueryable> m_taskFeasibilityQueryables=new HashSet<TaskFeasibilityQueryable>();
    private final Map<AuctionQueryableSubscription, AuctionQueryableSubscription> m_auctionQueryableSubscriptions=new LinkedHashMap<AuctionQueryableSubscription, AuctionQueryableSubscription>();

    private final Set<AuctionQueryable> m_auctionQueryables=new HashSet<AuctionQueryable>();

    //################################################################

    //----------------------------------------------------------------
    public SingleProcessCommProxy(Executor executor, boolean bUseLime) {
        Require.neqNull(executor, "executor");
        m_executor=executor;
        m_logger=Logger.getLogger("CommProxy");
        m_bUseLime=bUseLime;

        if (m_bUseLime) {
            // Fire up Lime for communication.
            new lime.util.Launcher().launch(new String[]{"-engage", "on", /*"-messages", "on", "-errors", "on", "-debug", "on"*/}, 0);
        }

        if (m_bUseLime) {
            m_fragmentQueryableSupport=new LimeFragmentQueryableSupport(m_logger, m_executor);
        } else {
            m_fragmentQueryableSupport=new SimpleFragmentQueryableSupport(m_executor);
        }

        m_executionDataQueryableSupport=new SimpleExecutionDataQueryableSupport(m_executor);
    }

    //----------------------------------------------------------------
    public void dispose() {
        if (m_bUseLime) {
            // Shut down Lime gracefully.
            LimeServer.getServer().shutdown(true);
        }
    }

    //----------------------------------------------------------------
    // from CommProxy
    public FragmentQueryable getFragmentQueryable() {
        return m_fragmentQueryableSupport.getFragmentQueryable();
    }

    //----------------------------------------------------------------
    // from CommProxy
    public TaskFeasibilityQueryable getTaskFeasibilityQueryable() {
        return m_taskFeasibilityQueryable;
    }

    //----------------------------------------------------------------
    // from CommProxy
    public AuctionQueryable getAuctionQueryable() {
        return m_auctionQueryable;
    }

    //----------------------------------------------------------------
    // from CommProxy
    public ExecutionDataQueryable getExecutionDataQueryable() {
        return m_executionDataQueryableSupport.getExecutionDataQueryable();
    }

    //################################################################

    //----------------------------------------------------------------
    // from CommProxy
    public void addFragmentQueryable(FragmentQueryable fragmentQueryable) {
        Require.neqNull(fragmentQueryable, "fragmentQueryable");
        try {
            m_fragmentQueryableSupport.addFragmentQueryable(fragmentQueryable);
        } catch (RequirementFailure requirementFailure) {
            throw requirementFailure.adjustForDelegatingMethod();
        }
    }

    //----------------------------------------------------------------
    // from CommProxy
    public void addTaskFeasibilityQueryable(TaskFeasibilityQueryable taskFeasibilityQueryable) {
        Require.neqNull(taskFeasibilityQueryable, "taskFeasibilityQueryable");
        synchronized (m_stateLock) {
            Require.eqFalse(m_taskFeasibilityQueryables.contains(taskFeasibilityQueryable), "m_taskFeasibilityQueryables.contains(taskFeasibilityQueryable)");
            m_taskFeasibilityQueryables.add(taskFeasibilityQueryable);
        }
    }

    //----------------------------------------------------------------
    // from CommProxy
    public void addAuctionQueryable(AuctionQueryable auctionQueryable) {
        Require.neqNull(auctionQueryable, "auctionQueryable");
        synchronized (m_stateLock) {
            Require.eqFalse(m_auctionQueryables.contains(auctionQueryable), "m_auctionQueryables.contains(auctionQueryable)");
            m_auctionQueryables.add(auctionQueryable);
            for (AuctionQueryableSubscription auctionQueryableSubscription : m_auctionQueryableSubscriptions.values()) {
                auctionQueryableSubscription.subscribe(auctionQueryable);
            }
        }
    }

    //----------------------------------------------------------------
    // from CommProxy
    public void addExecutionDataQueryable(ExecutionDataQueryable executionDataQueryable) {
        Require.neqNull(executionDataQueryable, "executionDataQueryable");
        try {
            m_executionDataQueryableSupport.addExecutionDataQueryable(executionDataQueryable);
        } catch (RequirementFailure requirementFailure) {
            throw requirementFailure.adjustForDelegatingMethod();
        }
    }

    //################################################################
    // FragmentQueryable

    //----------------------------------------------------------------
    private interface FragmentQueryableSupport {
        FragmentQueryable getFragmentQueryable();
        void addFragmentQueryable(FragmentQueryable fragmentQueryable);
    }

    //----------------------------------------------------------------
    private static class SimpleFragmentQueryableSupport implements FragmentQueryableSupport {

        private final Object m_stateLock=new Object();
        private final Executor m_executor;

        private final FragmentQueryable m_fragmentQueryable=new FragmentQueryable() {
            public void subscribeFragmentsWithConfiguration(ConfigurationNode configurationNode, Listener listener) {
                try {
                    SimpleFragmentQueryableSupport.this.subscribeFragmentsWithConfiguration(configurationNode, listener);
                } catch (RequirementFailure requirementFailure) {
                    throw requirementFailure.adjustForDelegatingMethodAndSyntheticAccessor();
                }
            }

            public void unsubscribeFragmentsWithConfiguration(ConfigurationNode configurationNode, Listener listener) {
                try {
                    SimpleFragmentQueryableSupport.this.unsubscribeFragmentsWithConfiguration(configurationNode, listener);
                } catch (RequirementFailure requirementFailure) {
                    throw requirementFailure.adjustForDelegatingMethodAndSyntheticAccessor();
                }
            }
        };

        private final Set<FragmentQueryable> m_fragmentQueryables=new HashSet<FragmentQueryable>();
        private final Map<FragmentQueryableSubscription, FragmentQueryableSubscription> m_fragmentQueryableSubscriptions=new LinkedHashMap<FragmentQueryableSubscription, FragmentQueryableSubscription>();

        //------------------------------------------------------------
        public SimpleFragmentQueryableSupport(Executor executor) {
            m_executor=executor;
        }

        //------------------------------------------------------------
        // from FragmentQueryableSupport
        public FragmentQueryable getFragmentQueryable() {
            return m_fragmentQueryable;
        }

        //------------------------------------------------------------
        // from FragmentQueryableSupport
        public void addFragmentQueryable(FragmentQueryable fragmentQueryable) {
            Require.neqNull(fragmentQueryable, "fragmentQueryable");
            synchronized (m_stateLock) {
                Require.eqFalse(m_fragmentQueryables.contains(fragmentQueryable), "m_fragmentQueryables.contains(fragmentQueryable)");
                m_fragmentQueryables.add(fragmentQueryable);
                for (FragmentQueryableSubscription fragmentQueryableSubscription : m_fragmentQueryableSubscriptions.values()) {
                    fragmentQueryableSubscription.subscribe(fragmentQueryable);
                }
            }
        }

        //------------------------------------------------------------
        // from FragmentQueryable
        private void subscribeFragmentsWithConfiguration(ConfigurationNode configurationNode, FragmentQueryable.Listener listener) {
            Require.neqNull(configurationNode, "configurationNode");
            Require.neqNull(listener, "listener");

            synchronized (m_stateLock) {
                FragmentQueryableSubscription fragmentQueryableSubscription=new FragmentQueryableSubscription(configurationNode, listener);
                Require.eqFalse(m_fragmentQueryableSubscriptions.containsKey(fragmentQueryableSubscription), "m_fragmentQueryableSubscriptions.containsKey(fragmentQueryableSubscription)");
                m_fragmentQueryableSubscriptions.put(fragmentQueryableSubscription, fragmentQueryableSubscription);
                for (FragmentQueryable fragmentQueryable : m_fragmentQueryables) {
                    fragmentQueryableSubscription.subscribe(fragmentQueryable);
                }
            }
        }

        //------------------------------------------------------------
        // from FragmentQueryable
        private void unsubscribeFragmentsWithConfiguration(ConfigurationNode configurationNode, FragmentQueryable.Listener listener) {
            Require.neqNull(configurationNode, "configurationNode");
            Require.neqNull(listener, "listener");

            synchronized (m_stateLock) {
                FragmentQueryableSubscription fragmentQueryableSubscription=m_fragmentQueryableSubscriptions.remove(new FragmentQueryableSubscription(configurationNode, listener));
                Require.neqNull(fragmentQueryableSubscription, "fragmentQueryableSubscription");
                for (FragmentQueryable fragmentQueryable : m_fragmentQueryables) {
                    fragmentQueryableSubscription.unsubscribe(fragmentQueryable);
                }
            }
        }

        //------------------------------------------------------------
        private class FragmentQueryableSubscription implements FragmentQueryable.Listener {
            private final ConfigurationNode m_configurationNode;
            private final FragmentQueryable.Listener m_listener;

            public FragmentQueryableSubscription(ConfigurationNode configurationNode, FragmentQueryable.Listener listener) {
                m_configurationNode=configurationNode;
                m_listener=listener;
            }

            @Override
            public boolean equals(Object that) {
                if (this==that) { return true; }
                if (that==null || getClass()!=that.getClass()) { return false; }
                FragmentQueryableSubscription fragmentQueryableSubscription=(FragmentQueryableSubscription)that;
                if (!m_configurationNode.equals(fragmentQueryableSubscription.m_configurationNode)) { return false; }
                if (!m_listener.equals(fragmentQueryableSubscription.m_listener)) { return false; }
                return true;
            }

            @Override
            public int hashCode() {
                int result=m_configurationNode.hashCode();
                result=31*result+m_listener.hashCode();
                return result;
            }

            public void handleFragment(final CompositeWorkflow compositeWorkflow) {
                m_executor.execute(new Runnable() {
                    public void run() {
                        m_listener.handleFragment(compositeWorkflow);
                    }
                });
            }

            public void subscribe(final FragmentQueryable fragmentQueryable) {
                m_executor.execute(new Runnable() {
                    public void run() {
                        fragmentQueryable.subscribeFragmentsWithConfiguration(m_configurationNode, FragmentQueryableSubscription.this);
                    }
                });
            }

            public void unsubscribe(final FragmentQueryable fragmentQueryable) {
                m_executor.execute(new Runnable() {
                    public void run() {
                        fragmentQueryable.unsubscribeFragmentsWithConfiguration(m_configurationNode, FragmentQueryableSubscription.this);
                    }
                });
            }
        }


    }

    //----------------------------------------------------------------
    private static class LimeFragmentQueryableSupport implements FragmentQueryableSupport {

        private FragmentQueryableImpl m_fragmentQueryable;
        private final Logger m_logger;

        //------------------------------------------------------------
        public LimeFragmentQueryableSupport(Logger logger, Executor executor) {
            Require.neqNull(logger, "logger");
            m_logger=logger;
            m_fragmentQueryable=new FragmentQueryableImpl(logger, executor);
        }

        //------------------------------------------------------------
        public FragmentQueryable getFragmentQueryable() {
            return m_fragmentQueryable;
        }

        //------------------------------------------------------------
        public void addFragmentQueryable(FragmentQueryable fragmentQueryable) {
            Require.neqNull(fragmentQueryable, "fragmentQueryable");
            FragmentObservable fragmentObservable=(FragmentObservable)fragmentQueryable;

            /* Create a new agent to handle the observable. */
            FragmentObserver observer = null;
            try {
                observer = (FragmentObserver) LimeServer.getServer().loadAgent("edu.wustl.cse.mobilab.openworkflow.comm.FragmentObserver", new Serializable[0]);
            } catch (LimeException le) {
                m_logger.error("Could not load FragmentObserver");
                m_logger.error(le.getMessage());
            }

            /* Observe the observable. */
            fragmentObservable.subscribe(observer);
        }
    }

    //################################################################
    // TaskFeasibilityQueryable

    //----------------------------------------------------------------
    // from TaskFeasibilityQueryable
    public void subscribeTaskFeasibility(TaskNode taskNode, TaskFeasibilityQueryable.Listener listener) {
        Assert.statementNeverExecuted();
    }

    //----------------------------------------------------------------
    // from TaskFeasibilityQueryable
    public void unsubscribeTaskFeasibility(TaskNode taskNode, TaskFeasibilityQueryable.Listener listener) {
        Assert.statementNeverExecuted();
    }

    //################################################################
    // AuctionQueryable

    //----------------------------------------------------------------
    // from AuctionQueryable
    private void startSolicitingBids(Task task, AuctionQueryable.Listener listener) {
        Require.neqNull(task, "task");
        Require.neqNull(listener, "listener");

        synchronized (m_stateLock) {
            AuctionQueryableSubscription auctionQueryableSubscription=new AuctionQueryableSubscription(task, listener);
            Require.eqFalse(m_auctionQueryableSubscriptions.containsKey(auctionQueryableSubscription), "m_auctionQueryableSubscriptions.containsKey(auctionQueryableSubscription)");
            m_auctionQueryableSubscriptions.put(auctionQueryableSubscription, auctionQueryableSubscription);
            for (AuctionQueryable auctionQueryable : m_auctionQueryables) {
                auctionQueryableSubscription.subscribe(auctionQueryable);
            }
        }
    }

    //----------------------------------------------------------------
    // from AuctionQueryable
    private void stopSolicitingBids(Task task, AuctionQueryable.Listener listener) {
        Require.neqNull(task, "task");
        Require.neqNull(listener, "listener");

        synchronized (m_stateLock) {
            AuctionQueryableSubscription auctionQueryableSubscription=m_auctionQueryableSubscriptions.remove(new AuctionQueryableSubscription(task, listener));
            Require.neqNull(auctionQueryableSubscription, "auctionQueryableSubscription");
            for (AuctionQueryable auctionQueryable : m_auctionQueryables) {
                auctionQueryableSubscription.unsubscribe(auctionQueryable);
            }
        }
    }

    //----------------------------------------------------------------
    private class AuctionQueryableSubscription implements AuctionQueryable.Listener {
        private final Task m_task;
        private final AuctionQueryable.Listener m_listener;

        //------------------------------------------------------------
        public AuctionQueryableSubscription(Task task, AuctionQueryable.Listener listener) {
            m_task=task;
            m_listener=listener;
        }

        //------------------------------------------------------------
        @Override
        public boolean equals(Object that) {
            if (this==that) { return true; }
            if (that==null || getClass()!=that.getClass()) { return false; }
            AuctionQueryableSubscription fragmentQueryableSubscription=(AuctionQueryableSubscription)that;
            if (!m_task.equals(fragmentQueryableSubscription.m_task)) { return false; }
            if (!m_listener.equals(fragmentQueryableSubscription.m_listener)) { return false; }
            return true;
        }

        //------------------------------------------------------------
        @Override
        public int hashCode() {
            int result=m_task.hashCode();
            result=31*result+m_listener.hashCode();
            return result;
        }

        //------------------------------------------------------------
        // from AuctionQueryable.Listener
        public void handleBid(final AuctionQueryable.Bidder bidder, final TaskBid taskBid) {
            m_executor.execute(new Runnable() {
                public void run() {
                    m_listener.handleBid(new AuctionQueryable.Bidder() {
                        public void confirmBid(final TaskBid confirmedTaskBid) {
                            m_executor.execute(new Runnable() {
                                public void run() {
                                    bidder.confirmBid(confirmedTaskBid);
                                }
                            });
                        }

                        public void rejectBid(final TaskBid rejectedTaskBid) {
                            m_executor.execute(new Runnable() {
                                public void run() {
                                    bidder.rejectBid(rejectedTaskBid);
                                }
                            });
                        }
                    }, taskBid);
                }
            });
        }

        //------------------------------------------------------------
        public void subscribe(final AuctionQueryable auctionQueryable) {
            m_executor.execute(new Runnable() {
                public void run() {
                    auctionQueryable.startSolicitingBids(m_task, AuctionQueryableSubscription.this);
                }
            });
        }

        //------------------------------------------------------------
        public void unsubscribe(final AuctionQueryable auctionQueryable) {
            m_executor.execute(new Runnable() {
                public void run() {
                    auctionQueryable.stopSolicitingBids(m_task, AuctionQueryableSubscription.this);
                }
            });
        }
    }

    //################################################################
    // ExecutionDataQueryable

    //----------------------------------------------------------------
    private interface ExecutionDataQueryableSupport {
        ExecutionDataQueryable getExecutionDataQueryable();
        void addExecutionDataQueryable(ExecutionDataQueryable executionDataQueryable);
    }

    //----------------------------------------------------------------
    private static class SimpleExecutionDataQueryableSupport implements ExecutionDataQueryableSupport {

        private final Object m_stateLock=new Object();
        private final Executor m_executor;

        private final ExecutionDataQueryable m_executiondataQueryable=new ExecutionDataQueryable() {
            public void subscribeData(ExecutionDataName executionDataName, Listener listener) {
                try {
                    SimpleExecutionDataQueryableSupport.this.subscribeData(executionDataName, listener);
                } catch (RequirementFailure requirementFailure) {
                    throw requirementFailure.adjustForDelegatingMethodAndSyntheticAccessor();
                }
            }

            public void unsubscribeData(ExecutionDataName executionDataName, Listener listener) {
                try {
                    SimpleExecutionDataQueryableSupport.this.unsubscribeData(executionDataName, listener);
                } catch (RequirementFailure requirementFailure) {
                    throw requirementFailure.adjustForDelegatingMethodAndSyntheticAccessor();
                }
            }

            public void declareData(ExecutionDataName executionDataName) {
                try {
                    SimpleExecutionDataQueryableSupport.this.declareData(executionDataName);
                } catch (RequirementFailure requirementFailure) {
                    throw requirementFailure.adjustForDelegatingMethodAndSyntheticAccessor();
                }
            }
        };

        private final Set<ExecutionDataQueryable> m_executiondataQueryables=new HashSet<ExecutionDataQueryable>();
        private final Map<ExecutionDataQueryableSubscription, ExecutionDataQueryableSubscription> m_executiondataQueryableSubscriptions=new LinkedHashMap<ExecutionDataQueryableSubscription, ExecutionDataQueryableSubscription>();

        //------------------------------------------------------------
        public SimpleExecutionDataQueryableSupport(Executor executor) {
            m_executor=executor;
        }

        //------------------------------------------------------------
        // from ExecutionDataQueryableSupport
        public ExecutionDataQueryable getExecutionDataQueryable() {
            return m_executiondataQueryable;
        }

        //------------------------------------------------------------
        // from ExecutionDataQueryableSupport
        public void addExecutionDataQueryable(ExecutionDataQueryable executiondataQueryable) {
            Require.neqNull(executiondataQueryable, "executiondataQueryable");
            synchronized (m_stateLock) {
                Require.eqFalse(m_executiondataQueryables.contains(executiondataQueryable), "m_executiondataQueryables.contains(executiondataQueryable)");
                m_executiondataQueryables.add(executiondataQueryable);
                for (ExecutionDataQueryableSubscription executiondataQueryableSubscription : m_executiondataQueryableSubscriptions.values()) {
                    executiondataQueryableSubscription.subscribe(executiondataQueryable);
                }
            }
        }

        //------------------------------------------------------------
        // from ExecutionDataQueryable
        public void subscribeData(ExecutionDataName executionDataName, ExecutionDataQueryable.Listener listener) {
            Require.neqNull(executionDataName, "executionDataName");
            Require.neqNull(listener, "listener");

            synchronized (m_stateLock) {
                ExecutionDataQueryableSubscription executiondataQueryableSubscription=new ExecutionDataQueryableSubscription(executionDataName, listener);
                Require.eqFalse(m_executiondataQueryableSubscriptions.containsKey(executiondataQueryableSubscription), "m_executiondataQueryableSubscriptions.containsKey(executiondataQueryableSubscription)");
                m_executiondataQueryableSubscriptions.put(executiondataQueryableSubscription, executiondataQueryableSubscription);
                for (ExecutionDataQueryable executiondataQueryable : m_executiondataQueryables) {
                    executiondataQueryableSubscription.subscribe(executiondataQueryable);
                }
            }
        }

        //------------------------------------------------------------
        // from ExecutionDataQueryable
        public void unsubscribeData(ExecutionDataName executionDataName, ExecutionDataQueryable.Listener listener) {
            Require.neqNull(executionDataName, "executionDataName");
            Require.neqNull(listener, "listener");

            synchronized (m_stateLock) {
                ExecutionDataQueryableSubscription executiondataQueryableSubscription=m_executiondataQueryableSubscriptions.remove(new ExecutionDataQueryableSubscription(executionDataName, listener));
                Require.neqNull(executiondataQueryableSubscription, "executiondataQueryableSubscription");
                for (ExecutionDataQueryable executiondataQueryable : m_executiondataQueryables) {
                    executiondataQueryableSubscription.unsubscribe(executiondataQueryable);
                }
            }
        }

        //------------------------------------------------------------
        // from ExecutionDataQueryable
        public void declareData(final ExecutionDataName executionDataName) {
            Require.neqNull(executionDataName, "executionDataName");
            synchronized (m_stateLock) {
                for (final ExecutionDataQueryable executiondataQueryable : m_executiondataQueryables) {
                    m_executor.execute(new Runnable() {
                        public void run() {
                            executiondataQueryable.declareData(executionDataName);
                        }
                    });
                }
            }
        }

        //------------------------------------------------------------
        private class ExecutionDataQueryableSubscription implements ExecutionDataQueryable.Listener {
            private final ExecutionDataName m_executionDataName;
            private final ExecutionDataQueryable.Listener m_listener;

            public ExecutionDataQueryableSubscription(ExecutionDataName executionDataName, ExecutionDataQueryable.Listener listener) {
                m_executionDataName=executionDataName;
                m_listener=listener;
            }

            @Override
            public boolean equals(Object that) {
                if (this==that) { return true; }
                if (that==null || getClass()!=that.getClass()) { return false; }
                ExecutionDataQueryableSubscription executiondataQueryableSubscription=(ExecutionDataQueryableSubscription)that;
                if (!m_executionDataName.equals(executiondataQueryableSubscription.m_executionDataName)) { return false; }
                if (!m_listener.equals(executiondataQueryableSubscription.m_listener)) { return false; }
                return true;
            }

            @Override
            public int hashCode() {
                int result=m_executionDataName.hashCode();
                result=31*result+m_listener.hashCode();
                return result;
            }

            public void handleData(final ExecutionDataName executionDataName, final Object data) {
                m_executor.execute(new Runnable() {
                    public void run() {
                        m_listener.handleData(executionDataName, data);
                    }
                });
            }

            public void subscribe(final ExecutionDataQueryable executiondataQueryable) {
                m_executor.execute(new Runnable() {
                    public void run() {
                        executiondataQueryable.subscribeData(m_executionDataName, ExecutionDataQueryableSubscription.this);
                    }
                });
            }

            public void unsubscribe(final ExecutionDataQueryable executiondataQueryable) {
                m_executor.execute(new Runnable() {
                    public void run() {
                        executiondataQueryable.unsubscribeData(m_executionDataName, ExecutionDataQueryableSubscription.this);
                    }
                });
            }
        }
    }

    //################################################################

    public void addFragmentObservable(FragmentObservable fragmentObservable) {
    }
}
