package edu.wustl.cse.mobilab.openworkflow.schedule;

import edu.wustl.cse.mobilab.openworkflow.util.Tuple2;
import edu.wustl.cse.mobilab.openworkflow.verify.Require;

import java.util.LinkedHashSet;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

//--------------------------------------------------------------------
/** (description of class)
 * @author Created by Louis Thomas on Nov 6, 2008. */
public class ScheduleManagerImpl implements ScheduleManager {

    private static final Location IN_TRANSIT = new LocationImpl("In transit");
    
    private final LocationManager m_locationManager;
    private final Location m_startLocation;

    private final SortedSet<Commitment> m_commitments = new TreeSet<Commitment>();
    private final Set<CommitmentListener> m_commitmentListeners = new LinkedHashSet<CommitmentListener>();
    
    private Location m_currentLocation;
    private final Set<CurrentLocationListener> m_currentLocationListeners = new LinkedHashSet<CurrentLocationListener>();

    //private final Map<TimePeriod, Location> m_durationToLocationObligations=new LinkedHashMap<TimePeriod, Location>();

    //----------------------------------------------------------------
    public ScheduleManagerImpl(LocationManager locationManager, Location startLocation) {
        Require.neqNull(locationManager, "locationManager");
        Require.neqNull(startLocation, "startLocation");
        m_locationManager = locationManager;
        m_currentLocation = m_startLocation = startLocation;
    }

    //----------------------------------------------------------------
    public void dispose() {
        // nothing to do yet
    }

    //----------------------------------------------------------------
    public ServiceCommitment addToSchedule(ServiceInvocationDescriptor serviceInvocationDescriptor) {
        ServiceCommitment serviceCommitment = new ServiceCommitmentImpl(serviceInvocationDescriptor);
        if (!addToSchedule(serviceCommitment)) {
            return null;
        }
        return serviceCommitment;
    }
    
    /*
     * Attempts to add commitment to schedule. Returns false on failure.
     */
    public boolean addToSchedule(Commitment commitment) {
        if (!doAddToSchedule(commitment)) {
            return false;
        }
        for (CommitmentListener commitmentListener : m_commitmentListeners) {
            commitmentListener.updateCommitment(null, commitment);
        }
        return true;
    }
    
    private boolean doAddToSchedule(Commitment commitment) {
        Require.neqNull(commitment, "commitment");
        TimePeriod taskTimePeriod = commitment.getTimePeriod();
        Tuple2<Commitment, Commitment> neightbors = findNeighborCommitments(taskTimePeriod.getStartTime());
        Commitment previousCommitment = neightbors.getA();
        Commitment nextCommitment = neightbors.getB();

        if (null != previousCommitment) {
            Duration interCommitmentDuration = taskTimePeriod.getStartTime().minus(previousCommitment.getTimePeriod().getEndTime());
            Duration travelDuration = m_locationManager.getTravelTime(previousCommitment.getLocation(), commitment.getLocation());
            if (travelDuration.compareTo(interCommitmentDuration) >= 0) {
                return false;
            }
        }

        if (null != nextCommitment) {
            Duration interCommitmentDuration = nextCommitment.getTimePeriod().getStartTime().minus(taskTimePeriod.getEndTime());
            Duration travelDuration = m_locationManager.getTravelTime(commitment.getLocation(), nextCommitment.getLocation());
            if (travelDuration.compareTo(interCommitmentDuration) >= 0) {
                return false;
            }
        }
 
        m_commitments.add(commitment);
        return true;
    }
    
    //----------------------------------------------------------------
    public void removeFromSchedule(Commitment commitment) {
        doRemoveFromSchedule(commitment);
        for (CommitmentListener commitmentListener : m_commitmentListeners) {
            commitmentListener.updateCommitment(commitment, null);
        }
    }
    
    private void doRemoveFromSchedule(Commitment commitment) {
        Require.neqNull(commitment, "commitment");
        boolean bWasInSet = m_commitments.remove(commitment);
        Require.eqTrue(bWasInSet, "bWasInSet");
    }

    /*
     * Attempts to update oldCommitment by atomically deleting oldCommitment
     * and adding newCommitment. That is, if the only commitment which conflicts
     * with newCommitment is oldCommitment, then the replace succeeds.
     */
    public boolean replace(Commitment oldCommitment, Commitment newCommitment) {
        Require.neqNull(oldCommitment, "oldCommitment");
        Require.neqNull(newCommitment, "newCommitment");
        
        doRemoveFromSchedule(oldCommitment);
        if (doAddToSchedule(newCommitment)) {
            // Let everyone know we succeeded.
            for (CommitmentListener commitmentListener : m_commitmentListeners) {
                commitmentListener.updateCommitment(oldCommitment, newCommitment);
            }
            return true;
        }
        // Revert the schedule.
        Require.eqTrue(doAddToSchedule(oldCommitment), "doAddToSchedule(oldCommitment)");
        return false;
    }

    //----------------------------------------------------------------
    private Tuple2<Commitment, Commitment> findNeighborCommitments(Time time) {
        Commitment previousCommitment = null;
        Commitment nextCommitment = null;
        for (Commitment commitment : m_commitments) {
            if (commitment.getTimePeriod().getStartTime().compareTo(time) < 0) {
                previousCommitment = commitment;
            } else if (commitment.getTimePeriod().getStartTime().compareTo(time) >= 0) {
                nextCommitment = commitment;
                break;
            }
        }
        return new Tuple2<Commitment, Commitment>(previousCommitment, nextCommitment);
    }

    /*
     * This method returns the latest possible time at which we can still be
     * found at the specified location. Our schedule dicatates the constraints.
     * 
     * If this method returns null, the window is infinite and we never need
     * to leave.
     */
    public Time getEndOfLoiteringWindow(Time time, Location location) {
        Require.neqNull(time, "time");
        Require.neqNull(location, "location");
        Require.neq(IN_TRANSIT, "IN_TRANSIT", location);
        Tuple2<Commitment, Commitment> neightbors = findNeighborCommitments(time);
        Commitment nextCommitment = neightbors.getB();
        return getTimeToLeaveForCommitment(location, nextCommitment);
    }
    
    /*
     * As getEndOfLoiteringWindow(Time, Location) but using the location at
     * the specified time.
     */
    public Time getEndOfLoiteringWindow(Time time) {
        return getEndOfLoiteringWindow(time, getLocation(time));
    }

    //----------------------------------------------------------------
    public Location getLocation(Time time) {
        Tuple2<Commitment, Commitment> neightbors = findNeighborCommitments(time);
        Commitment previousCommitment = neightbors.getA();
        Commitment nextCommitment = neightbors.getB();
        
        if (previousCommitment == null) {
            return m_startLocation;
        }
        
        Location previousLocation = previousCommitment.getLocation();
        if (previousCommitment.getTimePeriod().getEndTime().compareTo(time) < 0) {
            // We are actually in the middle of the commitment.
            return previousLocation;
        }
        Time leaveTime = getTimeToLeaveForCommitment(previousLocation, nextCommitment);
        if (leaveTime.compareTo(time) < 0) {
            // We're moving right now (or had better be!)
            return IN_TRANSIT;
        }
        // We're done with the commitment but are loitering.
        return previousLocation;
    }
    
    /*
     * Returns the time at which we must leave the given location to arrive at
     * the next commitment.
     */
    private Time getTimeToLeaveForCommitment(Location location, Commitment nextCommitment) {
        Require.neq(IN_TRANSIT, "IN_TRANSIT", location);
        if (nextCommitment == null) {
            return null;
        }
        Duration travelDuration = m_locationManager.getTravelTime(location, nextCommitment.getLocation());
        Time windowEnd = nextCommitment.getTimePeriod().getStartTime().minus(travelDuration);
        return windowEnd;
    }

    //----------------------------------------------------------------
    public Location getNextLocation(Time time) {
        return null;
        //TimePeriod period=getNextIntervalAfter(time);
        //return period==null?null:m_durationToLocationObligations.get(period);
    }

    //----------------------------------------------------------------
    public Time getNextCommitmentTimeAfter(Time time) {
        return null;
        //TimePeriod period=getNextIntervalAfter(time);
        //return null==period?null:period.getStartTime();
    }

    //----------------------------------------------------------------
    private TimePeriod getNextIntervalAfter(Time time) {
        return null;
        ////returns the time of the next obligation after time time.
        //TimePeriod defaultPeriod=new TimeImpl(1000000000, 0);
        //TimePeriod closestPeriod=defaultPeriod;
        //for (TimePeriod period : m_durationToLocationObligations.keySet()) {
        //    if (period.getStartTime().isAfter(time)) {
        //        if (period.getStartTime().isBefore(closestPeriod.getStartTime())) {
        //            closestPeriod=period;
        //        }
        //    }
        //}
        //if (closestPeriod!=defaultPeriod) {
        //    return closestPeriod;
        //} else {
        //    return null;
        //}
    }

    //----------------------------------------------------------------
    public Duration getTravelTime(Location startLocation, Location endLocation) {
        return m_locationManager.getTravelTime(startLocation, endLocation);
    }

    //################################################################

    //----------------------------------------------------------------
    public void subscribeCommitments(CommitmentListener commitmentListener) {
        Require.neqNull(commitmentListener, "commitmentListener");
        Require.eqFalse(m_commitmentListeners.contains(commitmentListener), "m_commitmentListeners.contains(commitmentListener)");
        m_commitmentListeners.add(commitmentListener);
        for (Commitment commitment : m_commitments) {
            commitmentListener.updateCommitment(null, commitment);
        }
    }

    //----------------------------------------------------------------
    public void unsubscribeCommitments(CommitmentListener commitmentListener) {
        Require.neqNull(commitmentListener, "commitmentListener");
        Require.eqTrue(m_commitmentListeners.contains(commitmentListener), "m_commitmentListeners.contains(commitmentListener)");
        m_commitmentListeners.remove(commitmentListener);
    }

    //################################################################
    // current location tracking

    //----------------------------------------------------------------
    public void subscribeCurrentLocation(CurrentLocationListener currentLocationListener) {
        Require.neqNull(currentLocationListener, "currentLocationListener");
        Require.eqFalse(m_currentLocationListeners.contains(currentLocationListener), "m_currentLocationListeners.contains(currentLocationListener)");
        m_currentLocationListeners.add(currentLocationListener);
        currentLocationListener.updateCurrentLocation(m_currentLocation);
    }

    //----------------------------------------------------------------
    public void unsubscribeCurrentLocation(CurrentLocationListener currentLocationListener) {
        Require.neqNull(currentLocationListener, "currentLocationListener");
        Require.eqTrue(m_currentLocationListeners.contains(currentLocationListener), "m_currentLocationListeners.contains(currentLocationListener)");
        m_currentLocationListeners.remove(currentLocationListener);
    }

    //----------------------------------------------------------------
    public void setCurrentLocation(Location location) {
        m_currentLocation = location;
        for (CurrentLocationListener currentLocationListener : m_currentLocationListeners) {
            currentLocationListener.updateCurrentLocation(m_currentLocation);
        }
    }

    //################################################################

    //----------------------------------------------------------------
    private static class ServiceCommitmentImpl extends CommitmentImpl implements ServiceCommitment {

        private final ServiceInvocationDescriptor m_serviceInvocationDescriptor;

        //------------------------------------------------------------
        public ServiceCommitmentImpl(ServiceInvocationDescriptor serviceInvocationDescriptor) {
            super(Require.neqNull(serviceInvocationDescriptor, "serviceInvocationDescriptor").getTimePeriod(), serviceInvocationDescriptor.getLocation());
            m_serviceInvocationDescriptor = serviceInvocationDescriptor;
        }

        //------------------------------------------------------------
        @Override
        public boolean equals(Object that) {
            if (this == that) {
                return true;
            }
            if (!(that instanceof ServiceCommitmentImpl)) {
                return false;
            }
            if (!super.equals(that)) {
                return false;
            }
            ServiceCommitmentImpl serviceCommitment = (ServiceCommitmentImpl) that;
            if (!m_serviceInvocationDescriptor.equals(serviceCommitment.m_serviceInvocationDescriptor)) {
                return false;
            }
            return true;
        }

        //------------------------------------------------------------
        @Override
        public int hashCode() {
            int result = super.hashCode();
            result = 31 * result + m_serviceInvocationDescriptor.hashCode();
            return result;
        }

        //------------------------------------------------------------
        public ServiceInvocationDescriptor getServiceInvocationDescriptor() {
            return m_serviceInvocationDescriptor;
        }
    }
}
