package edu.wustl.cse.mobilab.openworkflow.service;

import edu.wustl.cse.mobilab.openworkflow.comm.ExecutionDataQueryable;
import edu.wustl.cse.mobilab.openworkflow.verify.Assert;
import edu.wustl.cse.mobilab.openworkflow.verify.Require;
import edu.wustl.cse.mobilab.openworkflow.workflow.ExecutionDataName;

import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

//--------------------------------------------------------------------
/** (description of class)
* @author Created by Louis Thomas on Mar 5, 2009. */
public class ExecutionDataRepository implements ExecutionDataQueryable {

    private final Set<ExecutionDataSubscription> m_executionDataSubscriptions=new LinkedHashSet<ExecutionDataSubscription>();
    private final Map<ExecutionDataName, Object> m_executionDataNameToDataMap=new LinkedHashMap<ExecutionDataName, Object>();

    //----------------------------------------------------------------
    // from ExecutionDataQueryable
    public void subscribeData(ExecutionDataName executionDataName, Listener listener) {
        Require.neqNull(executionDataName, "executionDataName");
        Require.neqNull(listener, "replyListener");
        ExecutionDataSubscription executionDataSubscription=new ExecutionDataSubscription(executionDataName, listener);
        Require.eqFalse(m_executionDataSubscriptions.contains(executionDataSubscription), "m_executionDataSubscriptions.contains(executionDataSubscription)");
        m_executionDataSubscriptions.add(executionDataSubscription);

        if (m_executionDataNameToDataMap.containsKey(executionDataName)) {
            Object data=m_executionDataNameToDataMap.remove(executionDataName);
            executionDataSubscription.handleData(data);
        }
    }

    //----------------------------------------------------------------
    // from ExecutionDataQueryable
    public void unsubscribeData(ExecutionDataName executionDataName, Listener listener) {
        Require.neqNull(executionDataName, "executionDataName");
        Require.neqNull(listener, "listener");

        ExecutionDataSubscription executionDataSubscription=new ExecutionDataSubscription(executionDataName, listener);
        Require.eqTrue(m_executionDataSubscriptions.contains(executionDataSubscription), "m_executionDataSubscriptions.contains(executionDataSubscription)");
        m_executionDataSubscriptions.remove(executionDataSubscription);
    }

    //----------------------------------------------------------------
    // from ExecutionDataQueryable
    public void declareData(ExecutionDataName executionDataName) {
        Require.neqNull(executionDataName, "executionDataName");
        // nothing to do
    }

    //----------------------------------------------------------------
    public void handleOutputData(ExecutionDataName executionDataName, Object data) {
        Require.neqNull(executionDataName, "executionDataName");
        Assert.eqFalse(m_executionDataNameToDataMap.containsKey(executionDataName), "m_executionDataNameToDataMap.containsKey(executionDataName)");
        m_executionDataNameToDataMap.put(executionDataName, data);

        for (ExecutionDataSubscription executionDataSubscription : m_executionDataSubscriptions) {
            if (executionDataSubscription.getExecutionDataName().equals(executionDataName)) {
                executionDataSubscription.handleData(data);
                m_executionDataNameToDataMap.remove(executionDataName);
                break;
            }
        }
    }

    //----------------------------------------------------------------
    private static class ExecutionDataSubscription {

        private final ExecutionDataName m_executionDataName;
        private final Listener m_listener;
        private boolean m_bNotified;

        //------------------------------------------------------------
        private ExecutionDataSubscription(ExecutionDataName executionDataName, Listener listener) {
            Require.neqNull(executionDataName, "executionDataName");
            Require.neqNull(listener, "listener");
            m_executionDataName=executionDataName;
            m_listener=listener;
        }

        //------------------------------------------------------------
        public ExecutionDataName getExecutionDataName() {
            return m_executionDataName;
        }

        //------------------------------------------------------------
        public void handleData(Object data) {
            Assert.eqFalse(m_bNotified, "m_bNotified");
            m_bNotified=true;
            m_listener.handleData(m_executionDataName, data);
        }

        //------------------------------------------------------------
        @Override
        public boolean equals(Object that) {
            if (this==that) { return true; }
            if (that==null || getClass()!=that.getClass()) { return false; }
            ExecutionDataSubscription executionDataSubscription=(ExecutionDataSubscription)that;
            if (!m_executionDataName.equals(executionDataSubscription.m_executionDataName)) { return false; }
            if (!m_listener.equals(executionDataSubscription.m_listener)) { return false; }
            return true;
        }

        //----------------------------------------------------------------
        @Override
        public int hashCode() {
            int result=m_executionDataName.hashCode();
            result=31*result+m_listener.hashCode();
            return result;
        }
    }
}
