package edu.wustl.cse.mobilab.openworkflow.service;

import edu.wustl.cse.mobilab.openworkflow.comm.ExecutionDataQueryable;
import edu.wustl.cse.mobilab.openworkflow.schedule.Location;
import edu.wustl.cse.mobilab.openworkflow.schedule.ScheduleManager;
import edu.wustl.cse.mobilab.openworkflow.schedule.ServiceInvocationDescriptor;
import edu.wustl.cse.mobilab.openworkflow.schedule.Time;
import edu.wustl.cse.mobilab.openworkflow.util.TimeNotifier;
import edu.wustl.cse.mobilab.openworkflow.verify.Assert;
import edu.wustl.cse.mobilab.openworkflow.verify.Require;
import edu.wustl.cse.mobilab.openworkflow.workflow.ExecutionDataName;
import edu.wustl.cse.mobilab.openworkflow.workflow.ParameterName;
import org.apache.log4j.Logger;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.concurrent.Executor;

//--------------------------------------------------------------------
/** (description of class)
 * @author Created by Louis Thomas on Nov 6, 2008. */
public class ExecutionManagerImpl implements ExecutionManager {

    private final ScheduleManager m_scheduleManager;
    private final ServiceManager m_serviceManager;
    private final ExecutionDataQueryable m_executionDataQueryable;
    private final TimeNotifier m_timeNotifier;
    private final Executor m_executor;
    private final Logger m_logger;

    private final Map<ScheduleManager.Commitment, TriggerConditions> m_commitmentToTriggerConditionsMap=new LinkedHashMap<ScheduleManager.Commitment, TriggerConditions>();
    private final ExecutionDataRepository m_executionDataRepository=new ExecutionDataRepository();

    //----------------------------------------------------------------
    public ExecutionManagerImpl(ScheduleManager scheduleManager, ServiceManager serviceManager, ExecutionDataQueryable executionDataQueryable, TimeNotifier timeNotifier, Executor executor, Logger logger) {
        Require.neqNull(scheduleManager, "scheduleManager");
        Require.neqNull(serviceManager, "serviceManager");
        Require.neqNull(executionDataQueryable, "executionDataQueryable");
        Require.neqNull(timeNotifier, "timeNotifier");
        Require.neqNull(executor, "executor");
        Require.neqNull(logger, "logger");
        m_scheduleManager=scheduleManager;
        m_serviceManager=serviceManager;
        m_executionDataQueryable=executionDataQueryable;
        m_timeNotifier=timeNotifier;
        m_executor=executor;
        m_logger=logger;

        m_scheduleManager.subscribeCommitments(new ScheduleManager.CommitmentListener() {
            public void updateCommitment(ScheduleManager.Commitment oldCommitment, ScheduleManager.Commitment newCommitment) {
                ExecutionManagerImpl.this.updateCommitment(oldCommitment, newCommitment);
            }
        });
    }

    //----------------------------------------------------------------
    public void dispose() {
        // nothing to do yet
    }


    //----------------------------------------------------------------
    private void updateCommitment(ScheduleManager.Commitment oldCommitment, ScheduleManager.Commitment newCommitment) {
        if (null!=oldCommitment && m_commitmentToTriggerConditionsMap.containsKey(oldCommitment)) {
            m_logger.debug("ExecutionManagerImpl: Cancelled service commitment.");
            m_commitmentToTriggerConditionsMap.get(oldCommitment).dispose();
        }

        if (null!=newCommitment && newCommitment instanceof ScheduleManager.ServiceCommitment) {
            ScheduleManager.ServiceCommitment serviceCommitment=(ScheduleManager.ServiceCommitment)newCommitment;
            m_logger.debug("ExecutionManagerImpl: New service commitment.");

            TriggerConditions triggerConditions=new TriggerConditions(serviceCommitment);
            m_commitmentToTriggerConditionsMap.put(serviceCommitment, triggerConditions);
            triggerConditions.start();
        }
    }

    //################################################################

    //----------------------------------------------------------------
    // from ExecutionDataQueryable
    public void subscribeData(ExecutionDataName executionDataName, Listener listener) {
        m_executionDataRepository.subscribeData(executionDataName, listener);
    }

    //----------------------------------------------------------------
    // from ExecutionDataQueryable
    public void unsubscribeData(ExecutionDataName executionDataName, Listener listener) {
        m_executionDataRepository.unsubscribeData(executionDataName, listener);
    }

    //----------------------------------------------------------------
    // from ExecutionDataQueryable
    public void declareData(ExecutionDataName executionDataName) {
        m_executionDataRepository.declareData(executionDataName);
    }

    //----------------------------------------------------------------
    private void handleOutputData(ExecutionDataName executionDataName, Object data) {
        m_logger.debug("ExecutionManagerImpl: Declaring output execution data "+executionDataName+":\""+data+"\".");
        m_executionDataRepository.handleOutputData(executionDataName, data);
        m_executionDataQueryable.declareData(executionDataName);
    }

    //################################################################

    //----------------------------------------------------------------
    private class TriggerConditions implements TimeNotifier.Listener, ScheduleManager.CurrentLocationListener {

        private final ScheduleManager.ServiceCommitment m_serviceCommitment;

        private boolean m_bStarted;
        private boolean m_bDisposed;
        private boolean m_bStartTimeReached;
        private boolean m_bAtCorrectLocation;
        private final InputDetail[] m_inputDetails;

        //------------------------------------------------------------
        public TriggerConditions(ScheduleManager.ServiceCommitment serviceCommitment) {
            m_serviceCommitment=serviceCommitment;

            // add time listener
            m_timeNotifier.subscribeTimeNotification(m_serviceCommitment.getTimePeriod().getStartTime(), this);

            // add location listener
            m_scheduleManager.subscribeCurrentLocation(this);

            // add input listeners
            Map<ExecutionDataName, ParameterName> inputSubscriptionNameToParameterNameMap=m_serviceCommitment.getServiceInvocationDescriptor().getTask().getInputSubscriptionNameToParameterNameMap();
            m_inputDetails=new InputDetail[inputSubscriptionNameToParameterNameMap.size()];
            int nIndex=0;
            for (Map.Entry<ExecutionDataName, ParameterName> entry : inputSubscriptionNameToParameterNameMap.entrySet()) {
                m_inputDetails[nIndex]=new InputDetail(entry.getKey(), entry.getValue());
                nIndex++;
            }
            // do this second so that m_inputDetails is completely initialized first
            for (InputDetail inputDetail : m_inputDetails) {
                inputDetail.subscribeData(m_executionDataQueryable);
            }
        }

        //------------------------------------------------------------
        public void start() {
            m_bStarted=true;
            checkConditions();
        }

        //------------------------------------------------------------
        // from TimeNotifier.Listener
        public void notifyTimeReached(Time time) {
            Assert.equals(time, "time", m_serviceCommitment.getTimePeriod().getStartTime(), "m_serviceCommitment.getTimePeriod().getStartTime()");
            m_bStartTimeReached=true;
            checkConditions();
        }

        //------------------------------------------------------------
        // from ScheduleManager.CurrentLocationListener
        public void updateCurrentLocation(Location location) {
            m_bAtCorrectLocation=location.equals(m_serviceCommitment.getLocation());
            checkConditions();
        }

        //------------------------------------------------------------
        private void checkConditions() {
            if (m_bDisposed) { return; }
            if (!m_bStarted) { return; }
            if (!m_bStartTimeReached) { return; }
            if (!m_bAtCorrectLocation) { return; }
            for (InputDetail inputDetail : m_inputDetails) {
                if (!inputDetail.isInputReceived()) { return; }
            }
            trigger();
        }

        //------------------------------------------------------------
        private void trigger() {
            m_logger.debug("ExecutionManagerImpl: Service commitment conditions met. Triggering service invocation");
            dispose();
            m_executor.execute(new Runnable() {
                public void run() {
                    ServiceInvocationDescriptor serviceInvocationDescriptor=m_serviceCommitment.getServiceInvocationDescriptor();
                    ServiceDescriptor serviceDescriptor=serviceInvocationDescriptor.getServiceDescriptor();
                    m_logger.debug("ExecutionManagerImpl: Invoking "+serviceInvocationDescriptor+".");

                    // marshal input parameters
                    LinkedHashMap<String, Object> inputParameterMap=new LinkedHashMap<String, Object>();
                    for (InputDetail inputDetail : m_inputDetails) {
                        inputParameterMap.put(inputDetail.getParameterName().getName(), inputDetail.getData());
                    }

                    // invoke service
                    Map<String, Object> outputParameterMap=m_serviceManager.invokeService(serviceDescriptor, inputParameterMap);

                    // dispatch output parameters
                    Assert.assertion(null!=outputParameterMap, "null!=outputParameterMap", serviceDescriptor, "serviceDescriptor");
                    Map<ExecutionDataName, ParameterName> outputSubscriptionNameToParameterNameMap=serviceInvocationDescriptor.getTask().getOutputSubscriptionNameToParameterNameMap();
                    for (Map.Entry<ExecutionDataName, ParameterName> entry : outputSubscriptionNameToParameterNameMap.entrySet()) {
                        String sParameterName=entry.getValue().getName();
                        Assert.assertion(outputParameterMap.containsKey(sParameterName), "outputParameterMap.containsKey(sParameterName)", outputParameterMap, "outputParameterMap", sParameterName, "sParameterName", serviceDescriptor, "serviceDescriptor");
                        handleOutputData(entry.getKey(), outputParameterMap.get(sParameterName));
                    }
                }
            });
        }

        //------------------------------------------------------------
        public void dispose() {
            if (m_bDisposed) { return; }
            m_bDisposed=true;
            m_commitmentToTriggerConditionsMap.remove(m_serviceCommitment);

            // remove time listener
            m_timeNotifier.unsubscribeTimeNotification(m_serviceCommitment.getTimePeriod().getStartTime(), this);

            // remove location listener
            m_scheduleManager.unsubscribeCurrentLocation(this);

            // remove input listeners
            for (InputDetail inputDetail : m_inputDetails) {
                inputDetail.unsubscribeData(m_executionDataQueryable);
            }
        }

        //----------------------------------------------------------------
        private class InputDetail implements Listener {

            private final ExecutionDataName m_executionDataName;
            private final ParameterName m_parameterName;

            private boolean m_bInputReceived;
            private Object m_data;

            //----------------------------------------------------------------
            public InputDetail(ExecutionDataName executionDataName, ParameterName parameterName) {
                Require.neqNull(executionDataName, "executionDataName");
                Require.neqNull(parameterName, "parameterName");
                m_executionDataName=executionDataName;
                m_parameterName=parameterName;
            }

            //--------------------------------------------------------
            public ExecutionDataName getExecutionDataName() {
                return m_executionDataName;
            }

            //--------------------------------------------------------
            public ParameterName getParameterName() {
                return m_parameterName;
            }

            //--------------------------------------------------------
            public void subscribeData(ExecutionDataQueryable executionDataQueryable) {
                executionDataQueryable.subscribeData(m_executionDataName, this);
            }

            //--------------------------------------------------------
            public void unsubscribeData(ExecutionDataQueryable executionDataQueryable) {
                executionDataQueryable.unsubscribeData(m_executionDataName, this);
            }

            //--------------------------------------------------------
            public boolean isInputReceived() {
                return m_bInputReceived;
            }

            //--------------------------------------------------------
            public Object getData() {
                Require.eqTrue(m_bInputReceived, "m_bInputReceived");
                return m_data;
            }

            //--------------------------------------------------------
            // from ExecutionDataQueryable.Listener
            public void handleData(ExecutionDataName executionDataName, Object data) {
                Require.equals(m_executionDataName, "m_executionDataName", executionDataName, "executionDataName");
                Require.eqFalse(m_bInputReceived, "m_bInputReceived");
                m_data=data;
                m_bInputReceived=true;
                m_logger.debug("ExecutionManagerImpl: Received execution data \""+executionDataName+"\".");
                checkConditions();
            }
        }
    }
}
