package edu.wustl.cse.mobilab.openworkflow.service;

import edu.wustl.cse.mobilab.openworkflow.comm.TaskFeasibilityQueryable;
import edu.wustl.cse.mobilab.openworkflow.schedule.TimePeriod;
import edu.wustl.cse.mobilab.openworkflow.shell.PersistenceProvider;
import edu.wustl.cse.mobilab.openworkflow.shell.ui.TabManager;
import edu.wustl.cse.mobilab.openworkflow.util.SimpleXmlSerializer;
import edu.wustl.cse.mobilab.openworkflow.verify.Assert;
import edu.wustl.cse.mobilab.openworkflow.verify.Require;
import edu.wustl.cse.mobilab.openworkflow.workflow.Task;
import edu.wustl.cse.mobilab.openworkflow.workflow.TaskNode;
import edu.wustl.cse.mobilab.openworkflow.workflow.XmlDeserializationException;
import org.apache.log4j.Logger;
import org.jdom.Element;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

//--------------------------------------------------------------------
/** (description of class)
 * @author Created by Louis Thomas on Nov 6, 2008. */
public class ServiceManagerImpl implements ServiceManager {

    private final PersistenceProvider m_persistenceProvider;
    private final Logger m_logger;
    private final List<ServiceDescriptor> m_serviceDescriptors=new ArrayList<ServiceDescriptor>();
    private final Set<ServiceDescriptorListener> m_serviceDescriptorListeners=new LinkedHashSet<ServiceDescriptorListener>();
    private final TabManager m_tabManager;

    //----------------------------------------------------------------
    public ServiceManagerImpl(PersistenceProvider persistenceProvider, Logger logger, TabManager tabManager) {
        Require.neqNull(persistenceProvider, "persistenceProvider");
        Require.neqNull(logger, "logger");
        m_persistenceProvider=persistenceProvider;
        m_logger=logger;
        m_tabManager=tabManager; // may be null

        // Read configuration info from disk.
        Element element;
        try {
            element = m_persistenceProvider.loadXmlByName(SERVICESTORE_NAME);
        } catch (PersistenceProvider.NameNotFoundException nameNotFoundException) {
            throw Assert.exceptionNeverCaught(nameNotFoundException);
        }

        // Convert XML code into actual service objects
        SimpleXmlSerializer simpleXmlSerializer = new SimpleXmlSerializer();
        for (Object o : element.getContent()) {
            try {
                m_serviceDescriptors.add(simpleXmlSerializer.deserializeServiceNode((Element) o));
            } catch (XmlDeserializationException e) {
                throw Assert.exceptionNeverCaught(e);
            }
        }
    }

    //----------------------------------------------------------------
    public void dispose() {
        // nothing to do yet
    }

    //----------------------------------------------------------------
    public void subscribeTaskFeasibility(TaskNode taskNode, TaskFeasibilityQueryable.Listener listener) {
        // Keep track of this listener and if a new service is added whose contract matches this taskNode, then notify
        // this listener of that service.

        // ISSUE: WHO/WHAT ADDS SERVICES DYNAMICALLY TO THE LIST AS THEY BECOME AVAILABLE??  IS SERVICEMANAGER NOTIFIED OF THIS??

        // Look through all the services & return those whose contract currently matches the param taskNode
        for (ServiceDescriptor serviceDescriptor : m_serviceDescriptors) {
            if(serviceDescriptor.getContract().equals(taskNode.getContractLabel())){
                listener.handleTaskIsFeasible();
            }
        }
    }

    //----------------------------------------------------------------
    public void unsubscribeTaskFeasibility(TaskNode taskNode, TaskFeasibilityQueryable.Listener listener) {
        // Stop contacting the param listener when new services match its taskNode
        Assert.statementNeverExecuted();
    }

    //----------------------------------------------------------------
    public ServiceDescriptor getImplementingService(TaskNode taskNode) {
        for (ServiceDescriptor serviceDescriptor : m_serviceDescriptors) {
            if (serviceDescriptor.getContract().equals(taskNode.getContractLabel())) {
                return serviceDescriptor;
            }
        }
        return null;
    }

    //----------------------------------------------------------------
    public TimePeriod getTimeEstimate(Task task) {
        // Currently takes time from task--ideally, this value would depend on the implementing service.
        return task.getTimePeriod();
    }

    //----------------------------------------------------------------
    public double getCapabilityFraction(Task task) {
        // sf7: Strictly to make the project compile. This assumes that every task
        // corresponds to exactly one service and that every service is equally
        // valuable.
        return 1.0/m_serviceDescriptors.size();
    }

    //----------------------------------------------------------------
    public Map<String, Object> invokeService(ServiceDescriptor serviceDescriptor, Map<String, Object> parameters) {
        Require.neqNull(serviceDescriptor, "serviceDescriptor");
        Require.neqNull(parameters, "parameters");
        m_logger.debug("ServiceManagerImpl: Invoking service \""+serviceDescriptor.getServiceName()+"\" ("+serviceDescriptor.getJavaClass()+") with parameters "+parameters+".");

        // set up a few extra parameters
        Map<String, Object> extendedParameters=new LinkedHashMap<String, Object>(parameters);
        extendedParameters.put(Parameters.LOGGER, m_logger);
        extendedParameters.put(Parameters.CONFIGURATION_DATA, serviceDescriptor.getConfigurationData());
        if (null!=m_tabManager) {
            extendedParameters.put(Parameters.TAB_MANAGER, m_tabManager);
        }

        Map<String, Object> outputParameters;
        try {
            SimpleService simpleService=(SimpleService)Class.forName(serviceDescriptor.getJavaClass()).newInstance();
            outputParameters=simpleService.invoke(extendedParameters);
        } catch (InstantiationException e) {
            throw Assert.exceptionNeverCaught(e);
        } catch (IllegalAccessException e) {
            throw Assert.exceptionNeverCaught(e);
        } catch (ClassNotFoundException e) {
            throw Assert.exceptionNeverCaught(e);
        }
        m_logger.debug("ServiceManagerImpl: Service \""+serviceDescriptor.getServiceName()+"\" ("+serviceDescriptor.getJavaClass()+") completed with output parameters "+outputParameters+".");
        return outputParameters;
    }

    //----------------------------------------------------------------
    public void subscribeServiceDescriptors(ServiceDescriptorListener serviceDescriptorListener) {
        Require.neqNull(serviceDescriptorListener, "serviceDescriptorListener");
        Require.eqFalse(m_serviceDescriptorListeners.contains(serviceDescriptorListener), "m_serviceDescriptorListeners.contains(serviceDescriptorListener)");
        m_serviceDescriptorListeners.add(serviceDescriptorListener);
        for (ServiceDescriptor serviceDescriptor : m_serviceDescriptors) {
            serviceDescriptorListener.updateServiceDescriptor(null, serviceDescriptor);
        }
    }

    //----------------------------------------------------------------
    public void unsubscribeServiceDescriptors(ServiceDescriptorListener serviceDescriptorListener) {
        Require.neqNull(serviceDescriptorListener, "serviceDescriptorListener");
        Require.eqTrue(m_serviceDescriptorListeners.contains(serviceDescriptorListener), "m_serviceDescriptorListeners.contains(serviceDescriptorListener)");
        m_serviceDescriptorListeners.remove(serviceDescriptorListener);
    }
}
