package edu.wustl.cse.mobilab.openworkflow.shell.ui;

import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.LayoutManager;
import java.awt.Rectangle;
import java.awt.geom.AffineTransform;
import javax.swing.Icon;
import javax.swing.SwingUtilities;
import javax.swing.plaf.metal.MetalTabbedPaneUI;
import javax.swing.text.View;

//--------------------------------------------------------------------
/** Custom UI class that has the following fixes:
 * <UL>
 * <LI> Lays out lables and draws the text sideways.
 * <LI> Uses a custom TabPaneLayout that calculates the preferred size
 * correctly even when the content size is 0x0.
 * <LI> Has a border only on one side.
 * </UL>
 * @author Created by Louis Thomas on Feb 5, 2009. */
public class CustomTabbedPaneUI extends MetalTabbedPaneUI {

    public static final int BORDER_SIZE=1;

    //----------------------------------------------------------------
    // always use our own custom (wrapped) layout
    @Override
    protected LayoutManager createLayoutManager() {
        return new CustomTabbedPaneLayout();
    }

    //----------------------------------------------------------------
    // set the content border thickness to zero.
    @Override
    protected void installDefaults() {
        super.installDefaults();
        if (TOP==tabPane.getTabPlacement()) {
            contentBorderInsets=new Insets(BORDER_SIZE, 0, 0, 0);
        } else {
            contentBorderInsets=new Insets(0, BORDER_SIZE, 0, 0);
        }
    }

    //----------------------------------------------------------------
    @Override
    protected int calculateTabHeight(int tabPlacement, int tabIndex, int fontHeight) {
        if (TOP==tabPlacement) {
            return super.calculateTabHeight(tabPlacement, tabIndex, fontHeight);
        } else {
            return super.calculateTabWidth(tabPlacement, tabIndex, getFontMetrics());
        }
    }

    //----------------------------------------------------------------
    @Override
    protected int calculateTabWidth(int tabPlacement, int tabIndex, FontMetrics metrics) {
        if (TOP==tabPlacement) {
            return super.calculateTabWidth(tabPlacement, tabIndex, metrics);
        } else {
            return super.calculateTabHeight(tabPlacement, tabIndex, metrics.getHeight());
        }
    }

    //----------------------------------------------------------------
    @Override
    protected void layoutLabel(int tabPlacement, FontMetrics metrics, int tabIndex, String title, Icon icon, Rectangle tabRect, Rectangle iconRect, Rectangle textRect, boolean isSelected) {
        textRect.x=textRect.y=iconRect.x=iconRect.y=0;

        View v=getTextViewForTab(tabIndex);
        if (v!=null) {
            tabPane.putClientProperty("html", v);
        }

        if (TOP!=tabPlacement) {
            transposeRect(tabRect);
        }

        SwingUtilities.layoutCompoundLabel(tabPane,
            metrics, title, icon,
            SwingUtilities.CENTER,
            SwingUtilities.CENTER,
            SwingUtilities.CENTER,
            SwingUtilities.TRAILING,
            tabRect,
            iconRect,
            textRect,
            textIconGap);

        if (TOP!=tabPlacement) {
            transposeRect(tabRect);
            transposeRect(iconRect);
            transposeRect(textRect);
        }

        tabPane.putClientProperty("html", null);

        int xNudge=getTabLabelShiftX(tabPlacement, tabIndex, isSelected);
        int yNudge=getTabLabelShiftY(tabPlacement, tabIndex, isSelected);
        iconRect.x+=xNudge;
        iconRect.y+=yNudge;
        textRect.x+=xNudge;
        textRect.y+=yNudge;
    }

    //----------------------------------------------------------------
    private void transposeRect(Rectangle rectangle) {
        int temp=rectangle.width;
        rectangle.width=rectangle.height;
        rectangle.height=temp;

        temp=rectangle.x;
        rectangle.x=rectangle.y;
        rectangle.y=temp;
    }

    //----------------------------------------------------------------
    @Override
    protected void paintText(Graphics g, int tabPlacement, Font font, FontMetrics metrics, int tabIndex, String title, Rectangle textRect, boolean isSelected) {
        if (TOP!=tabPlacement) {
            Graphics2D g2=(Graphics2D)g;

            // save old values
            AffineTransform transform=g2.getTransform();
            int oldX=textRect.x;
            int oldY=textRect.y;
            int oldWidth=textRect.width;
            int oldHeight=textRect.height;

            // rotate 90 degrees
            //   positive rotation goes from x+ to y+; since y+ is down, positive is clockwise
            g2.rotate(-Math.PI/2);

            // rearrange the rectange dimensions
            textRect.x=-oldY-oldHeight;
            textRect.y=oldX;
            textRect.width=oldHeight;
            textRect.height=oldWidth;

            super.paintText(g, tabPlacement, font, metrics, tabIndex, title, textRect, isSelected);

            // restore old values
            textRect.x=oldX;
            textRect.y=oldY;
            textRect.width=oldWidth;
            textRect.height=oldHeight;
            g2.setTransform(transform);
        } else {
            super.paintText(g, tabPlacement, font, metrics, tabIndex, title, textRect, isSelected);
        }

    }

    //----------------------------------------------------------------
    @Override
    protected void paintContentBorderTopEdge(Graphics g, int tabPlacement, int selectedIndex, int x, int y, int w, int h) {
        if (TOP==tabPlacement) {
            super.paintContentBorderTopEdge(g, tabPlacement, selectedIndex, x, y, w, h);
        }
    }

    //----------------------------------------------------------------
    @Override
    protected void paintContentBorderBottomEdge(Graphics g, int tabPlacement, int selectedIndex, int x, int y, int w, int h) {
        if (BOTTOM==tabPlacement) {
            super.paintContentBorderBottomEdge(g, tabPlacement, selectedIndex, x, y, w, h);
        }
    }

    //----------------------------------------------------------------
    @Override
    protected void paintContentBorderLeftEdge(Graphics g, int tabPlacement, int selectedIndex, int x, int y, int w, int h) {
        if (LEFT==tabPlacement) {
            super.paintContentBorderLeftEdge(g, tabPlacement, selectedIndex, x, y, w, h);
        }
    }

    //----------------------------------------------------------------
    @Override
    protected void paintContentBorderRightEdge(Graphics g, int tabPlacement, int selectedIndex, int x, int y, int w, int h) {
        if (RIGHT==tabPlacement) {
            super.paintContentBorderRightEdge(g, tabPlacement, selectedIndex, x, y, w, h);
        }
    }

    //################################################################

    //----------------------------------------------------------------
    public class CustomTabbedPaneLayout extends TabbedPaneLayout {

        //------------------------------------------------------------
        // always assume 1 row of tabs is preferred
        @Override
        protected int preferredTabAreaHeight(int tabPlacement, int width) {
            int tabCount=tabPane.getTabCount();
            int total=0;
            if (tabCount>0) {
                int maxTabHeight=calculateMaxTabHeight(tabPlacement);
                total=calculateTabAreaHeight(tabPlacement, 1, maxTabHeight);
            }
            return total;
        }

        //------------------------------------------------------------
        // always assume 1 column of tabs is preferred
        @Override
        protected int preferredTabAreaWidth(int tabPlacement, int height) {
            int tabCount=tabPane.getTabCount();
            int total=0;
            if (tabCount>0) {
                maxTabWidth=calculateMaxTabWidth(tabPlacement);
                total=calculateTabAreaWidth(tabPlacement, 1, maxTabWidth);
            }
            return total;
        }
    }
}