package edu.wustl.cse.mobilab.openworkflow.shell.ui;

import edu.wustl.cse.mobilab.openworkflow.verify.Require;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.util.LinkedHashMap;
import java.util.Map;
import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.UIManager;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

//--------------------------------------------------------------------
/** Primary implementation for {@link TabManager}.
 * @author Created by Louis Thomas on Jan 22, 2009. */
public class TabManagerImpl implements TabManager {

    private static final String TAB_NAME_PROPERTY="tabName";

    private final JTabbedPane m_staticTabbedPane;
    private final JTabbedPane m_dynamicTabbedPane;
    private final JPanel m_contentPanel;
    private final CardLayout m_contentPanelCardLayout;
    private final JPanel m_outerPanel;
    private final Map<TabModel, TabModelInfo> m_tabModelToTabModelInfoMap=new LinkedHashMap<TabModel, TabModelInfo>();

    //----------------------------------------------------------------
    public TabManagerImpl() {
        m_staticTabbedPane=new JTabbedPane(JTabbedPane.LEFT, JTabbedPane.WRAP_TAB_LAYOUT);
        m_staticTabbedPane.setUI(new CustomTabbedPaneUI());
        m_dynamicTabbedPane=new JTabbedPane(JTabbedPane.TOP, JTabbedPane.WRAP_TAB_LAYOUT);
        m_dynamicTabbedPane.setUI(new CustomTabbedPaneUI());

        // add space before the top tabs so they don't hang over the edge of the content
        m_staticTabbedPane.addComponentListener(new ComponentAdapter() {
            @Override public void componentResized(ComponentEvent e) {
                m_dynamicTabbedPane.setBorder(BorderFactory.createEmptyBorder(0, m_staticTabbedPane.getWidth()-CustomTabbedPaneUI.BORDER_SIZE, 0, 0));
            }
        });

        m_contentPanelCardLayout=new CardLayout();
        m_contentPanel=new JPanel(m_contentPanelCardLayout);
        m_contentPanel.setBorder(BorderFactory.createEmptyBorder(3, 3, 3, 3));
        m_contentPanel.setBackground(UIManager.getColor("TabbedPane.selected"));

        m_outerPanel=new JPanel(new BorderLayout());
        m_outerPanel.add(m_staticTabbedPane, BorderLayout.WEST);
        m_outerPanel.add(m_dynamicTabbedPane, BorderLayout.NORTH);
        m_outerPanel.add(m_contentPanel, BorderLayout.CENTER);

        // add model changed listeners to the tabbed panes to track which tab is selected.
        m_staticTabbedPane.getModel().addChangeListener(makeTabChangeListener(m_staticTabbedPane, m_dynamicTabbedPane));
        m_dynamicTabbedPane.getModel().addChangeListener(makeTabChangeListener(m_dynamicTabbedPane, m_staticTabbedPane));
    }

    //----------------------------------------------------------------
    private ChangeListener makeTabChangeListener(final JTabbedPane thisTabbedPane, final JTabbedPane otherTabbedPane) {
        return new ChangeListener() {
            public void stateChanged(ChangeEvent e) {
                Component component=thisTabbedPane.getSelectedComponent();
                if (null!=component) {
                    String name=(String)((JComponent)component).getClientProperty(TAB_NAME_PROPERTY);
                    m_contentPanelCardLayout.show(m_contentPanel, name);
                    otherTabbedPane.setSelectedIndex(-1);
                }
            }
        };
    }

    //----------------------------------------------------------------
    public void removeTab(TabModel tabModel) {
        Require.neqNull(tabModel, "tabModel");
        Require.eqTrue(m_tabModelToTabModelInfoMap.containsKey(tabModel), "m_tabModelToTabModelInfoMap.containsKey(tabModel)");

        TabModelInfo tabModelInfo=m_tabModelToTabModelInfoMap.remove(tabModel);
        m_contentPanel.remove(tabModelInfo.getComponent());
        if (tabModelInfo.getCategory()==Category.DYNAMIC) {
            m_dynamicTabbedPane.remove(tabModelInfo.getPanel());
        } else {
            m_staticTabbedPane.remove(tabModelInfo.getPanel());
        }
    }

    //----------------------------------------------------------------
    public void addTab(TabModel tabModel, Category category) {
        Require.neqNull(tabModel, "tabModel");
        Require.neqNull(category, "category");
        Require.eqFalse(m_tabModelToTabModelInfoMap.containsKey(tabModel), "m_tabModelToTabModelInfoMap.containsKey(tabModel)");
        JPanel panel=new JPanel();
        Dimension dimension=new Dimension(0, 0);
        panel.setPreferredSize(dimension);
        panel.setMinimumSize(dimension);
        panel.setMaximumSize(dimension);

        TabModelInfo tabModelInfo = new TabModelInfo(panel, category, tabModel.getComponent());
        m_tabModelToTabModelInfoMap.put(tabModel,tabModelInfo);

        panel.putClientProperty(TAB_NAME_PROPERTY, tabModel.getName());
        m_contentPanel.add(tabModel.getComponent(), tabModel.getName());
        if (category==Category.DYNAMIC) {
            m_dynamicTabbedPane.addTab(tabModel.getName(), panel);
        } else {
            m_staticTabbedPane.addTab(tabModel.getName(), panel);
        }
    }

    //----------------------------------------------------------------
    public JComponent getComponent() {
        return m_outerPanel;
    }

    //----------------------------------------------------------------
    private static class TabModelInfo {

        private final JPanel m_panel;
        private final Category m_category;
        private final Component m_component;

        //------------------------------------------------------------
        public TabModelInfo(JPanel panel, Category category, Component component) {
            m_panel=panel;
            m_category=category;
            m_component=component;
        }

        //------------------------------------------------------------
        public Component getComponent() {
            return m_component;
        }

        //------------------------------------------------------------
        public Category getCategory() {
            return m_category;
        }

        //------------------------------------------------------------
        public Component getPanel() {
            return m_panel;
        }
    }
}
