package edu.wustl.cse.mobilab.openworkflow.util;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

//--------------------------------------------------------------------
/** (description of class)
 * @author Created by Louis Thomas on Jun 20, 2006 */
public class ArgsHelper {
    private String[] m_args;
    private int m_nNextArg;

    //----------------------------------------------------------------
    public ArgsHelper(String[] args) {
        m_args=args;
        m_nNextArg=0;
    }

    //----------------------------------------------------------------
    public boolean moreArgsAvailable() {
        return  m_args.length!=m_nNextArg;
    }

    //----------------------------------------------------------------
    public String peekNextArg() {
        if (!moreArgsAvailable()) {
            return null;
        }
        return m_args[m_nNextArg];
    }

    //----------------------------------------------------------------
    public String getNextArg() {
        if (!moreArgsAvailable()) {
            throw new IllegalArgumentException("Missing parameter after \""+m_args[m_nNextArg-1]+"\".");
        }
        String sArg=m_args[m_nNextArg];
        m_nNextArg++;
        return sArg;
    }

    //----------------------------------------------------------------
    public String getNextSwitch() {
        String sArg=getNextArg();
        if ('-'!=sArg.charAt(0) && '/'!=sArg.charAt(0)) {
            throw new IllegalArgumentException("Expected a switch, but got \""+sArg+"\".");
        }
        return sArg.substring(1);
    }

    //----------------------------------------------------------------
    public long getNextNumber() {
        String sArg=getNextArg().toLowerCase();
        try {
            if (sArg.startsWith("0x")) {
                return Long.parseLong(sArg.substring(2), 16);
            } else {
                return Long.parseLong(sArg);
            }
        } catch (NumberFormatException e) {
            throw new IllegalArgumentException("Expected a number but got \""+sArg+"\"");
        }
    }

    private static final Pattern DATE_WITH_DASHES_PATTERN=Pattern.compile("(\\d{4})-(\\d{2})-(\\d{2})");
    private static final Pattern DATE_NO_DASHES_PATTERN=Pattern.compile("(\\d{4})(\\d{2})(\\d{2})");
    private static final Pattern DATE_WITH_SLASHES_PATTERN=Pattern.compile("(\\d{2})/(\\d{2})/(\\d{2}|\\d{4})");

    //----------------------------------------------------------------
    public int getNextDateYyyymmdd() {
        return parseAndCheckDateYyyymmdd(getNextArg());
    }

    //----------------------------------------------------------------
    public static int parseAndCheckDateYyyymmdd(String sArg) {
        Matcher matcher=DATE_WITH_DASHES_PATTERN.matcher(sArg);
        if (matcher.matches()) {
            return validateDate(Integer.parseInt(matcher.group(1))*10000+Integer.parseInt(matcher.group(2))*100+Integer.parseInt(matcher.group(3)), sArg);
        }
        matcher=DATE_NO_DASHES_PATTERN.matcher(sArg);
        if (matcher.matches()) {
            return validateDate(Integer.parseInt(matcher.group(1))*10000+Integer.parseInt(matcher.group(2))*100+Integer.parseInt(matcher.group(3)), sArg);
        }
        matcher=DATE_WITH_SLASHES_PATTERN.matcher(sArg);
        if (matcher.matches()) {
            String sYear=matcher.group(3);
            int nYear=Integer.parseInt(sYear);
            if (2==sYear.length()) {
                nYear+=(nYear>=80?1900:2000); // like SimpleDateFormatter, but relative to 2000 instead of current year.
            }
            return validateDate(nYear*10000+Integer.parseInt(matcher.group(1))*100+Integer.parseInt(matcher.group(2)), sArg);
        }
        throw new IllegalArgumentException("Date \""+sArg+"\" is not in YYYYMMDD, YYYY-MM-DD, MM/DD/YYYY, or MM/DD/YY format.");
    }

    //----------------------------------------------------------------
    private static int validateDate(int nDateYyyymmdd, String sArg) {
        int nYear=nDateYyyymmdd/10000;
        int nMonth=(nDateYyyymmdd/100)%100;
        int nDay=nDateYyyymmdd%100;
        if (nYear<1500 || nYear>2500 || nMonth<1 || nMonth>12 || nDay<1 || nDay>31) {
            throw new IllegalArgumentException("Expected a valid date, but got \""+sArg+"\"");
        }
        return nDateYyyymmdd;
    }

    //----------------------------------------------------------------
    public boolean getNextBoolean() {
        String sArg=getNextArg().toLowerCase();
        if ("y".equalsIgnoreCase(sArg) || "yes".equalsIgnoreCase(sArg) || "t".equalsIgnoreCase(sArg) || "true".equalsIgnoreCase(sArg) || "1".equalsIgnoreCase(sArg)) {
            return true;
        } else if ("n".equalsIgnoreCase(sArg) || "no".equalsIgnoreCase(sArg) || "f".equalsIgnoreCase(sArg) || "false".equalsIgnoreCase(sArg) || "0".equalsIgnoreCase(sArg)) {
            return false;
        }
        throw new IllegalArgumentException("Expected Y or N but got \""+sArg+"\".");
    }

    //----------------------------------------------------------------
    public String[] getAllRemainingArgs() {
        int nRemainingArgs=m_args.length-m_nNextArg;
        String[] remainingArgs=new String[nRemainingArgs];
        if (nRemainingArgs>0) {
            System.arraycopy(m_args, m_nNextArg, remainingArgs, 0, nRemainingArgs);
        }
        m_nNextArg+=nRemainingArgs;
        return remainingArgs;
    }

    //----------------------------------------------------------------
    public void confirmNoMoreArgs() {
        if (moreArgsAvailable()) {
            throw new IllegalArgumentException("Expected end of arguments but got \""+getNextArg()+"\".");
        }
    }
}
