package edu.wustl.cse.mobilab.openworkflow.util;

import edu.wustl.cse.mobilab.openworkflow.verify.Require;
import edu.wustl.cse.mobilab.openworkflow.verify.RequirementFailure;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

//--------------------------------------------------------------------
/** Utility methods for creating collections.
 * @author Created by Louis Thomas on Mar 9, 2006 */
public class CollectionUtil {

    public static final byte[] ZERO_LENGTH_BYTE_ARRAY=new byte[0];
    public static final short[] ZERO_LENGTH_SHORT_ARRAY=new short[0];
    public static final int[] ZERO_LENGTH_INT_ARRAY=new int[0];
    public static final int[][] ZERO_LENGTH_INT_ARRAY_ARRAY=new int[0][];
    public static final long[] ZERO_LENGTH_LONG_ARRAY=new long[0];
    public static final float[] ZERO_LENGTH_FLOAT_ARRAY=new float[0];
    public static final double[] ZERO_LENGTH_DOUBLE_ARRAY=new double[0];
    public static final double[][] ZERO_LENGTH_DOUBLE_ARRAY_ARRAY=new double[0][];
    public static final Object[] ZERO_LENGTH_OBJECT_ARRAY=new Object[0];
    public static final String[] ZERO_LENGTH_STRING_ARRAY=new String[0];
    public static final String[][] ZERO_LENGTH_STRING_ARRAY_ARRAY=new String[0][];

    //----------------------------------------------------------------
    public static <K, V> Map<K, V> unmodifiableMapFromArray(Class<K> typeK, Class<V> typeV, Object... data) {
        try {
            return Collections.unmodifiableMap(mapFromArray(typeK, typeV, data));
        } catch (RequirementFailure e) {
            throw e.adjustForDelegatingMethod();
        }
    }

    //----------------------------------------------------------------
    public static <K, V> Map<K, V> unmodifiableInvertMap(Map<V, K> sourceMap) {
        try {
            return Collections.unmodifiableMap(invertMap(sourceMap));
        } catch (RequirementFailure e) {
            throw e.adjustForDelegatingMethod();
        }
    }

    //----------------------------------------------------------------
    public static <E> Set<E> unmodifiableSetFromArray(E... data) {
        try {
            return Collections.unmodifiableSet(setFromArray(data));
        } catch (RequirementFailure e) {
            throw e.adjustForDelegatingMethod();
        }
    }

    //----------------------------------------------------------------
    @SuppressWarnings({"unchecked"})
    public static <K, V> Map<K, V> mapFromArray(Class<K> typeK, Class<V> typeV, Object... data) {
        Require.neqNull(data, "data");
        Require.requirement(0==data.length%2, "0==data.length%2");
        Map<K, V> map=newSizedLinkedHashMap(data.length/2);
        for (int nIndex=0; nIndex<data.length; nIndex+=2) {
            Object key=data[nIndex];
            if (null!=key) {
                Require.instanceOf(key, "key", typeK);
            }
            Require.requirement(false==map.containsKey(key), "false==map.containsKey(data[nIndex])", key, "data[nIndex]");
            Object value=data[nIndex+1];
            if (null!=value) {
                Require.instanceOf(value, "value", typeV);
            }
            map.put((K)key, (V)value);
        }
        return map;
    }

    //----------------------------------------------------------------
    public static <K, V> Map<K, V> invertMap(Map<V, K> sourceMap) {
        Require.neqNull(sourceMap, "sourceMap");
        Map<K, V> targetMap=newSizedLinkedHashMap(sourceMap.size());
        for (Map.Entry<V, K> entry : sourceMap.entrySet()) {
            K key=entry.getValue();
            V value=entry.getKey();
            Require.requirement(false==targetMap.containsKey(key), "false==targetMap.containsKey(key)", key, "key");
            targetMap.put(key, value);
        }
        return targetMap;
    }

    //----------------------------------------------------------------
    public static <E> Set<E> setFromArray(E... data) {
        Require.neqNull(data, "data");
        Set<E> set=newSizedLinkedHashSet(data.length);
        for (E elem : data) {
            Require.requirement(false==set.contains(elem), "false==set.contains(data[nIndex])", elem, "data[nIndex]");
            set.add(elem);
        }
        return set;
    }

    //----------------------------------------------------------------
    public static <E> List<E> listFromArray(E ... data) {
        Require.neqNull(data, "data");
        List<E> list=new ArrayList<E>(data.length);
        for (E elem : data) {
            list.add(elem);
        }
        return list;
    }

    //----------------------------------------------------------------
    /** Returns an empty {@link HashMap} with a big enough capacity
     * such that the given number of entries can be added without
     * resizing. */
    public static <K, V> Map<K, V> newSizedHashMap(int nEntries) {
        return new HashMap<K, V>((nEntries+1)*4/3);
    }

    //----------------------------------------------------------------
    /** Returns an empty {@link LinkedHashMap} with a big enough
     * capacity such that the given number of entries can be added
     * without resizing. */
    public static <K, V> Map<K, V> newSizedLinkedHashMap(int nEntries) {
        return new LinkedHashMap<K, V>((nEntries+1)*4/3);
    }

    //----------------------------------------------------------------
    /** Returns an empty {@link HashSet} with a big enough capacity
     * such that the given number of entries can be added without
     * resizing. */
    public static <E> Set<E> newSizedHashSet(int nEntries) {
        return new HashSet<E>((nEntries+1)*4/3);
    }

    //----------------------------------------------------------------
    /** Returns an empty {@link LinkedHashSet} with a big enough capacity
     * such that the given number of entries can be added without
     * resizing. */
    public static <E> Set<E> newSizedLinkedHashSet(int nEntries) {
        return new LinkedHashSet<E>((nEntries+1)*4/3);
    }

    //----------------------------------------------------------------
    /** Converts a List of Doubles to an array of doubles. */
    public static double[] convertDoublesToPrimitiveArray(Collection<Double> collection) {
        double[] array=new double[collection.size()];
        Iterator<Double> iter=collection.iterator();
        for (int nIndex=0; nIndex<array.length; nIndex++) {
            array[nIndex]=iter.next();
        }
        return array;
    }

    //----------------------------------------------------------------
    /** Converts a Collection of Integers to an array of ints. */
    public static int[] convertIntegersToPrimitiveArray(Collection<Integer> collection) {
        int[] array=new int[collection.size()];
        Iterator<Integer> iter=collection.iterator();
        for (int nIndex=0; nIndex<array.length; nIndex++) {
            array[nIndex]=iter.next();
        }
        return array;
    }

    //----------------------------------------------------------------
    /** Converts a Collection of Integers to an array of ints. */
    public static long[] convertLongsToPrimitiveArray(Collection<Long> collection) {
        long[] array=new long[collection.size()];
        Iterator<Long> iter=collection.iterator();
        for (int nIndex=0; nIndex<array.length; nIndex++) {
            array[nIndex]=iter.next();
        }
        return array;
    }

    //----------------------------------------------------------------
    /** Shuffles the elements in an array. */
    public static <E> void shuffle(E[] elems) {
        for (int nIndex=elems.length; nIndex>1; nIndex--) {
            int nChoice=(int)(Math.random()*nIndex);
            int nSwapWith=nIndex-1;
            if (nChoice!=nSwapWith) {
                E temp=elems[nSwapWith];
                elems[nSwapWith]=elems[nChoice];
                elems[nChoice]=temp;
            }
        }
    }

    //----------------------------------------------------------------
    /** Formats a set as a sorted list of strings. */
    public static <E> String formatSetStringSorted(Set<E> set) {
        if (null==set) {
            return "(null)";
        }
        String[] strings=new String[set.size()];
        int nOffset=0;
        for (E e : set) {
            String s=null==e?"(null)":e.toString();
            strings[nOffset++]=s;
        }
        Arrays.sort(strings);
        return Arrays.toString(strings);
    }
}
