package edu.wustl.cse.mobilab.openworkflow.util;

import edu.wustl.cse.mobilab.openworkflow.verify.Assert;
import edu.wustl.cse.mobilab.openworkflow.verify.Require;

//--------------------------------------------------------------------
/** An object that handles reply callbacks and allows a thread to
 * wait for the callback to occur.
 * @author Created by Louis Thomas on Apr 28, 2006 */
public class ReplyCatcher<SuccessType> implements ReplyListener<SuccessType> {

    /** The lock that protects most of the state. */
    private final Object m_oStateLock=this;

    /** The reply status from the notification. */
    private ReplyStatus m_replyStatus;

    /** The reply detail from the notification. */
    private Object m_replyDetail;

    /** The extra info from the notification. */
    private Object m_extra;

    /** Indicates whether a notification has been recieved yet or
     * not. */
    private boolean m_bDone;

    //----------------------------------------------------------------
    /** Saves the reply and notifies anyone waiting for the reply. */
    public void replyNotification(ReplyStatus replyStatus, Object replyDetail, Object extra) {
        synchronized (m_oStateLock) {
            Require.eqFalse(m_bDone, "m_bDone");
            m_replyStatus=replyStatus;
            m_replyDetail=replyDetail;
            m_extra=extra;
            m_bDone=true;
            m_oStateLock.notifyAll();
        }
    }

    //----------------------------------------------------------------
    /** Waits for the reply to come back. */
    public void waitForReply() {
        synchronized (m_oStateLock) {
            while (!m_bDone) {
                try {
                    m_oStateLock.wait();
                } catch (InterruptedException e) {
                    Assert.exceptionNeverCaught(e);
                }
            }
        }
    }

    //----------------------------------------------------------------
    /** Gets the reply status of the request. */
    public ReplyStatus getReplyStatus() {
        synchronized (m_oStateLock) {
            Require.eqTrue(isDone(), "isDone()");
            return m_replyStatus;
        }
    }

    //----------------------------------------------------------------
    /** Gets the reply detail of the request. */
    public Object getReplyDetail() {
        synchronized (m_oStateLock) {
            Require.eqTrue(isDone(), "isDone()");
            return m_replyDetail;
        }
    }

    //----------------------------------------------------------------
    /** Gets the extra info of the request. */
    public Object getExtra() {
        synchronized (m_oStateLock) {
            Require.eqTrue(isDone(), "isDone()");
            return m_extra;
        }
    }

    //----------------------------------------------------------------
    /** Gets whether or not a reply has been received let. */
    public boolean isDone() {
        synchronized (m_oStateLock) {
            return m_bDone;
        }
    }

    //----------------------------------------------------------------
    /** Returns a short description of the reply in "STATUS (detail)"
     * format. */
    @Override
    public String toString() {
        synchronized (m_oStateLock) {
            if (!m_bDone) {
                return "(reply not yet received)";
            } else {
                return m_replyStatus.formatWithDetail(m_replyDetail);
            }
        }
    }
}
