package edu.wustl.cse.mobilab.openworkflow.util;

import edu.wustl.cse.mobilab.openworkflow.service.ServiceDescriptor;
import edu.wustl.cse.mobilab.openworkflow.service.ServiceDescriptorImpl;
import edu.wustl.cse.mobilab.openworkflow.verify.Require;
import edu.wustl.cse.mobilab.openworkflow.workflow.CompositeWorkflow;
import edu.wustl.cse.mobilab.openworkflow.workflow.CompositeWorkflowBuilder;
import edu.wustl.cse.mobilab.openworkflow.workflow.CompositeWorkflowBuilderImpl;
import edu.wustl.cse.mobilab.openworkflow.workflow.ConfigurationNode;
import edu.wustl.cse.mobilab.openworkflow.workflow.ConfigurationNodeImpl;
import edu.wustl.cse.mobilab.openworkflow.workflow.Junction;
import edu.wustl.cse.mobilab.openworkflow.workflow.TaskNode;
import edu.wustl.cse.mobilab.openworkflow.workflow.TaskNodeImpl;
import edu.wustl.cse.mobilab.openworkflow.workflow.XmlDeserializationException;
import org.jdom.Element;

import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

//--------------------------------------------------------------------
/** Serializes system objects to XML and deserializes system objects
 * from XML.
 * @author Created by Louis Thomas on Dec 3, 2008. */
public class SimpleXmlSerializer {

    public static final String COMPOSITE_WORKFLOW_ELEMENT="CompositeWorkflow";
    public static final String TASK_NODE_ELEMENT="Task";
    public static final String CONTRACT_LABEL_ATTRIBUTE="ContractLabel";
    public static final String JUNCTION_ATTRIBUTE="Junction";
    public static final String CONJUNCT_VALUE="*";
    public static final String DISJUNCT_VALUE="+";
    public static final String PRECONDITION_ELEMENT="Precondition";
    public static final String POSTCONDITION_ELEMENT="Postcondition";
    public static final String CONFIGURATION_NODE_ELEMENT="Configuration";
    public static final String PREDICATE_LABEL_ATTRIBUTE="PredicateLabel";
    public static final String SERVICE_NODE_ELEMENT="Service";
    public static final String SERVICE_NAME_ATTRIBUTE="Name";
    public static final String SERVICE_CONTRACT_ATTRIBUTE="Contract";
    public static final String SERVICE_JAVACLASS_ATTRIBUTE="JavaClass";

    //################################################################

    //----------------------------------------------------------------
    public CompositeWorkflow deserializeCompositeWorkflow(Element compositeWorkflowElement) throws XmlDeserializationException {
        Require.neqNull(compositeWorkflowElement, "compositeWorkflowElement");

        if (!COMPOSITE_WORKFLOW_ELEMENT.equals(compositeWorkflowElement.getName())) {
            throw new XmlDeserializationException("Expected \""+COMPOSITE_WORKFLOW_ELEMENT+"\" element, but got \""+compositeWorkflowElement.getName()+"\" element.");
        }
        CompositeWorkflowBuilder compositeWorkflowBuilder=new CompositeWorkflowBuilderImpl();

        //noinspection unchecked
        for (Element childElement : (List<Element>)compositeWorkflowElement.getChildren()) {
            if (TASK_NODE_ELEMENT.equals(childElement.getName())) {
                TaskNode taskNode=deserializeTaskNode(childElement);
                compositeWorkflowBuilder.addTaskNode(taskNode);
            } else {
                throw new XmlDeserializationException("Unknown child element \""+childElement.getName()+"\" in \""+COMPOSITE_WORKFLOW_ELEMENT+"\" element.");
            }
        }
        return compositeWorkflowBuilder;
    }

    //----------------------------------------------------------------
    public TaskNode deserializeTaskNode(Element taskNodeElement) throws XmlDeserializationException {

        Require.neqNull(taskNodeElement, "taskNodeElement");

        if (!TASK_NODE_ELEMENT.equals(taskNodeElement.getName())) {
            throw new XmlDeserializationException("Expected \""+TASK_NODE_ELEMENT+"\" element, but got \""+taskNodeElement.getName()+"\" element.");
        }

        String sContractLabel=taskNodeElement.getAttributeValue(CONTRACT_LABEL_ATTRIBUTE);
        if (null==sContractLabel || 0==sContractLabel.length()) {
            throw new XmlDeserializationException("\""+TASK_NODE_ELEMENT+"\" element is missing \""+CONTRACT_LABEL_ATTRIBUTE+"\" attribute.");
        }

        String sJunction=taskNodeElement.getAttributeValue(JUNCTION_ATTRIBUTE);
        if (null==sJunction || 0==sJunction.length()) {
            throw new XmlDeserializationException("\""+TASK_NODE_ELEMENT+"\" element is missing \""+JUNCTION_ATTRIBUTE+"\" attribute.");
        }
        if (!CONJUNCT_VALUE.equals(sJunction) && !DISJUNCT_VALUE.equals(sJunction)) {
            throw new XmlDeserializationException("Expected \""+CONJUNCT_VALUE+"\" or \""+DISJUNCT_VALUE+"\" for \""+JUNCTION_ATTRIBUTE+"\" attribute, but got \""+sJunction+"\".");
        }
        Junction junction=CONJUNCT_VALUE.equals(sJunction)?Junction.CONJUNCT:Junction.DISJUNCT;

        Set<ConfigurationNode> preconditionConfigurationNodes=Collections.emptySet();
        Set<ConfigurationNode> postconditionConfigurationNodes=Collections.emptySet();

        //noinspection unchecked
        for (Element childElement : (List<Element>)taskNodeElement.getChildren()) {
            if (PRECONDITION_ELEMENT.equals(childElement.getName())) {
                preconditionConfigurationNodes=deserializeConfigurationNodeSet(childElement);
            } else if (POSTCONDITION_ELEMENT.equals(childElement.getName())) {
                postconditionConfigurationNodes=deserializeConfigurationNodeSet(childElement);
            } else {
                throw new XmlDeserializationException("Unknown child element \""+childElement.getName()+"\" in \""+TASK_NODE_ELEMENT+"\" element.");
            }
        }

        if (preconditionConfigurationNodes.isEmpty()) {
            throw new XmlDeserializationException("No precondition configuration nodes specified.");
        }
        if (postconditionConfigurationNodes.isEmpty()) {
            throw new XmlDeserializationException("No postcondition configuration nodes specified.");
        }

        return new TaskNodeImpl(sContractLabel, junction, preconditionConfigurationNodes, postconditionConfigurationNodes);
    }

    //----------------------------------------------------------------
    public Set<ConfigurationNode> deserializeConfigurationNodeSet(Element element) throws XmlDeserializationException {
        Require.neqNull(element, "element");

        Set<ConfigurationNode> configurationNodes=new LinkedHashSet<ConfigurationNode>();

        //noinspection unchecked
        for (Element childElement : (List<Element>)element.getChildren()) {
            if (CONFIGURATION_NODE_ELEMENT.equals(childElement.getName())) {
                configurationNodes.add(deserializeConfigurationNode(childElement));
            } else {
                throw new XmlDeserializationException("Unknown child element \""+childElement.getName()+"\" in \""+element.getName()+"\" element.");
            }
        }
        return configurationNodes;
    }

    //----------------------------------------------------------------
    public ConfigurationNode deserializeConfigurationNode(Element configurationNodeElement) throws XmlDeserializationException {
        Require.neqNull(configurationNodeElement, "configurationNodeElement");

        if (!CONFIGURATION_NODE_ELEMENT.equals(configurationNodeElement.getName())) {
            throw new XmlDeserializationException("Expected \""+CONFIGURATION_NODE_ELEMENT+"\" element, but got \""+configurationNodeElement.getName()+"\" element.");
        }

        String sPredicateLabel=configurationNodeElement.getAttributeValue(PREDICATE_LABEL_ATTRIBUTE);
        if (null==sPredicateLabel || 0==sPredicateLabel.length()) {
            throw new XmlDeserializationException("Missing \""+PREDICATE_LABEL_ATTRIBUTE+"\" attribute for \""+CONFIGURATION_NODE_ELEMENT+"\" element.");
        }

        return new ConfigurationNodeImpl(sPredicateLabel);
    }

    //----------------------------------------------------------------
    public ServiceDescriptor deserializeServiceNode(Element serviceNodeElement) throws XmlDeserializationException {
        Require.neqNull(serviceNodeElement, "serviceNodeElement");

        if (!SERVICE_NODE_ELEMENT.equals(serviceNodeElement.getName())) {
            throw new XmlDeserializationException("Expected \""+SERVICE_NODE_ELEMENT+"\" element, but got \""+serviceNodeElement.getName()+"\" element.");
        }

        String sName=serviceNodeElement.getAttributeValue(SERVICE_NAME_ATTRIBUTE);
        if (null==sName || 0==sName.length()) {
            throw new XmlDeserializationException("Missing \""+SERVICE_NAME_ATTRIBUTE+"\" attribute for \""+SERVICE_NODE_ELEMENT+"\" element.");
        }

        String sContract=serviceNodeElement.getAttributeValue(SERVICE_CONTRACT_ATTRIBUTE);
        if (null==sContract || 0==sContract.length()) {
            throw new XmlDeserializationException("Missing \""+SERVICE_CONTRACT_ATTRIBUTE+"\" attribute for \""+SERVICE_NODE_ELEMENT+"\" element.");
        }

        String sJavaClass=serviceNodeElement.getAttributeValue(SERVICE_JAVACLASS_ATTRIBUTE);
        if (null==sJavaClass || 0==sJavaClass.length()) {
            throw new XmlDeserializationException("Missing \""+SERVICE_JAVACLASS_ATTRIBUTE+"\" attribute for \""+SERVICE_NODE_ELEMENT+"\" element.");
        }

        return new ServiceDescriptorImpl(sName, sContract, sJavaClass, serviceNodeElement);
    }

    //################################################################

    //----------------------------------------------------------------
    public Element serializeServiceNode(ServiceDescriptor serviceDescriptor){
        Require.neqNull(serviceDescriptor, "serviceDescriptor");
        Element serviceNodeElement = new Element(SERVICE_NODE_ELEMENT);
        serviceNodeElement.setAttribute(SERVICE_NAME_ATTRIBUTE, serviceDescriptor.getServiceName());
        serviceNodeElement.setAttribute(SERVICE_CONTRACT_ATTRIBUTE, serviceDescriptor.getContract());
        serviceNodeElement.setAttribute(SERVICE_JAVACLASS_ATTRIBUTE, serviceDescriptor.getJavaClass());
        return serviceNodeElement;
    }

    //----------------------------------------------------------------
    public Element serializeConfigurationNode(ConfigurationNode configurationNode) {
        Require.neqNull(configurationNode, "configurationNode");
        Element configurationNodeElement = new Element(CONFIGURATION_NODE_ELEMENT);
        configurationNodeElement.setAttribute(PREDICATE_LABEL_ATTRIBUTE, configurationNode.getPredicateLabel());
        return configurationNodeElement;

    }

    //----------------------------------------------------------------
    public Element serializeTaskNode(TaskNode taskNode) {
        Require.neqNull(taskNode, "taskNode");
        Element taskNodeElement = new Element(TASK_NODE_ELEMENT);
        taskNodeElement.setAttribute(CONTRACT_LABEL_ATTRIBUTE, taskNode.getContractLabel());
        taskNodeElement.setAttribute(JUNCTION_ATTRIBUTE, Junction.CONJUNCT==taskNode.getJunction()?CONJUNCT_VALUE:DISJUNCT_VALUE);
        Element preconditionElement = new Element(PRECONDITION_ELEMENT);
        for (ConfigurationNode configurationNode : taskNode.getPreconditions()) {
            preconditionElement.addContent(serializeConfigurationNode(configurationNode));
        }
        taskNodeElement.addContent(preconditionElement);

        Element postconditionElement = new Element(POSTCONDITION_ELEMENT);
        for (ConfigurationNode configurationNode : taskNode.getPostconditions()) {
            postconditionElement.addContent(serializeConfigurationNode(configurationNode));
        }
        taskNodeElement.addContent(postconditionElement);
        return taskNodeElement;

    }

    //----------------------------------------------------------------
    public Element serializeCompositeWorkflow(CompositeWorkflow compositeWorkflow) {
        Require.neqNull(compositeWorkflow, "compositeWorkflow");
        Element compositeWorkflowElement = new Element(COMPOSITE_WORKFLOW_ELEMENT);
        for (TaskNode taskNode : compositeWorkflow.getTaskNodes()) {
            compositeWorkflowElement.addContent(serializeTaskNode(taskNode));
        }
        return compositeWorkflowElement;
    }
}
