package edu.wustl.cse.mobilab.openworkflow.util;

import edu.wustl.cse.mobilab.openworkflow.schedule.Time;
import edu.wustl.cse.mobilab.openworkflow.shell.SystemTimeProvider;
import edu.wustl.cse.mobilab.openworkflow.verify.Require;

import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

//--------------------------------------------------------------------
public class SimulationTimeNotifier implements TimeNotifier, SystemTimeProvider {

    private final SortedMap<Time, Set<Listener>> m_timeToFutureListenersMap=new TreeMap<Time, Set<Listener>>();
    private final SortedMap<Time, Set<Listener>> m_timeToPastListenersMap=new TreeMap<Time, Set<Listener>>();

    private Time m_systemTime;

    //----------------------------------------------------------------
    public SimulationTimeNotifier(Time systemTime) {
        m_systemTime=systemTime;
        Require.neqNull(systemTime, "systemTime");
    }

    //----------------------------------------------------------------
    public void setSystemTime(Time systemTime) {
        Require.neqNull(systemTime, "systemTime");
        Require.ltZero(m_systemTime.compareTo(systemTime), "m_systemTime.compareTo(systemTime)");
        m_systemTime=systemTime;
        notifyReadyTimeListeners();
    }

    //----------------------------------------------------------------
    // from SystemTimeProvider
    public Time getSystemTime() {
        return m_systemTime;
    }

    //----------------------------------------------------------------
    // from TimeNotifier
    public void subscribeTimeNotification(Time time, Listener listener) {
        Require.neqNull(time, "time");
        Require.neqNull(listener, "timeNotifierListener");

        Set<Listener> timeNotifierListeners=m_timeToFutureListenersMap.get(time);
        if (null==timeNotifierListeners) {
            timeNotifierListeners=new LinkedHashSet<Listener>();
            m_timeToFutureListenersMap.put(time, timeNotifierListeners);
        }
        Require.eqFalse(timeNotifierListeners.contains(listener), "timeNotifierListeners.contains(timeNotifierListener)");
        timeNotifierListeners.add(listener);
        notifyReadyTimeListeners();
    }

    //----------------------------------------------------------------
    // from TimeNotifier
    public void unsubscribeTimeNotification(Time time, Listener listener) {
        Require.neqNull(time, "time");
        Require.neqNull(listener, "timeNotifierListener");

        SortedMap<Time, Set<Listener>> timeToTimeListenerMap;
        if (m_timeToFutureListenersMap.containsKey(time)) {
            timeToTimeListenerMap=m_timeToFutureListenersMap;
        } else {
            Require.eqTrue(m_timeToPastListenersMap.containsKey(time), "m_timeToPastListenersMap.containsKey(time)");
            timeToTimeListenerMap=m_timeToPastListenersMap;
        }
        Set<Listener> timeNotifierListeners=timeToTimeListenerMap.get(time);
        Require.eqTrue(timeNotifierListeners.contains(listener), "timeNotifierListeners.contains(timeNotifierListener)");
        timeNotifierListeners.remove(listener);
        if (timeNotifierListeners.isEmpty()) {
            timeToTimeListenerMap.remove(time);
        }
    }

    //----------------------------------------------------------------
    private void notifyReadyTimeListeners() {

        // do all list manipulation before notifying listeners to prevent concurrent modification
        List<Tuple2<Listener, Time>> pendingNotifications=new LinkedList<Tuple2<Listener, Time>>();

        for (Iterator<Map.Entry<Time, Set<Listener>>> itr=m_timeToFutureListenersMap.entrySet().iterator(); itr.hasNext();) {
            Map.Entry<Time, Set<Listener>> entry=itr.next();
            Time desiredTime=entry.getKey();
            if (desiredTime.compareTo(m_systemTime)<=0) {
                itr.remove();
                Set<Listener> futureTimeNotifierListeners=entry.getValue();
                for (Listener timeNotifierListener : futureTimeNotifierListeners) {
                    pendingNotifications.add(new Tuple2<Listener, Time>(timeNotifierListener, desiredTime));
                }
                if (m_timeToPastListenersMap.containsKey(desiredTime)) {
                    Set<Listener> pastTimeNotifierListeners=m_timeToPastListenersMap.get(desiredTime);
                    pastTimeNotifierListeners.addAll(futureTimeNotifierListeners);
                } else {
                    m_timeToPastListenersMap.put(desiredTime, futureTimeNotifierListeners);
                }
            } else {
                break;
            }
        }

        for (Tuple2<Listener, Time> pendingNotification : pendingNotifications) {
            pendingNotification.getA().notifyTimeReached(pendingNotification.getB());
        }
    }
}
