package edu.wustl.cse.mobilab.openworkflow.util;

import edu.wustl.cse.mobilab.openworkflow.verify.Assert;
import edu.wustl.cse.mobilab.openworkflow.verify.Require;

import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.Executor;

//--------------------------------------------------------------------
/** Trivial implementation of a lightweight executable framework.
 * From <U>Concurrent Programming In Java, 2nd Ed.</U>
 *   (&sect; 4.1.3.1) by Doug Lea.
 * @author Created by Louis Thomas on Oct 10, 2002 */
public class SingleThreadQueuedExecutor implements Executor {

    //################################################################
    // fields

    /** The lock that protects the internal state. */
    private final Object m_oStateLock=new Object();

    /** The list of tasks to execute.
     * Protected by {@link #m_oStateLock}.*/
    private final List<Runnable> m_taskList=new LinkedList<Runnable>();

    /** When this object is encountered on the task list, exit. */
    private Runnable m_terminationIndicator;

    /** The thread that is executing tasks from the list. */
    private Thread m_executionThread;

    //################################################################
    // methods

    //----------------------------------------------------------------
    /** Creates a new task queue and a thread to execute tasks placed
     * on that queue. */
    public SingleThreadQueuedExecutor(String sName) {
        m_executionThread=new Thread(new Runnable() {
            public void run() {
                dispatchTasks();
            }
        }, "SingleThreadQueuedExecutor for "+sName);
        //m_executionThread.setDaemon(true);
        m_executionThread.start();
    }

    //----------------------------------------------------------------
    /** Places a task on the queue for deferred execution. */
    public void execute(Runnable runnable) {
        Require.neqNull(runnable, "runnable");
        synchronized (m_oStateLock) {
            m_taskList.add(runnable);
            m_oStateLock.notify();
        }
    }

    //----------------------------------------------------------------
    /** Executed tasks off the queue. */
    private void dispatchTasks() {
        while (true) {
            Runnable task;

            // get a task off the task list
            synchronized (m_oStateLock) {
                while (m_taskList.isEmpty()) {
                    try {
                        m_oStateLock.wait();
                    } catch (InterruptedException e) {
                        Assert.exceptionNeverCaught(e);
                    }
                }
                task=m_taskList.remove(0);

                // see if we need to exit
                if (task==m_terminationIndicator) {
                    break;
                }
            }

            // execute it
            task.run();

            task=null; // encourage GC
        }
    }

    //----------------------------------------------------------------
    /** Stops the execution thread and waits for it to exit, first
     * executing all currenly queued tasks. */
    public void waitForQueueEmptyAndShutDown() {
        shutdownInternal(false);
    }

    //----------------------------------------------------------------
    /** Stops the execution thread and waits for it to exit,
     * discarding all currently queued tasks. */
    public void shutDownNow() {
        shutdownInternal(true);
    }

    private void shutdownInternal(boolean shutdownNow) {
        synchronized (m_oStateLock) {
            Require.eqNull(m_terminationIndicator, "m_terminationIndicator");
            m_terminationIndicator=m_executionThread;
            if (shutdownNow) {
                m_taskList.clear();
            }
            execute(m_terminationIndicator);
        }
        try {
            m_executionThread.join();
        } catch (InterruptedException e) {
            Assert.exceptionNeverCaught(e);
        }
    }

    public void setName(String name) {
        Require.neqNull(name, "name");
        m_executionThread.setName(name);
    }
}
