package edu.wustl.cse.mobilab.openworkflow.verify;

import java.awt.EventQueue;

//--------------------------------------------------------------------
/** Requirement methods for simple runtime program verification.
 * Failed requirements throw {@link RequirementFailure}.
 * <p> Methods:
 * <ul>
 * <li> void requirement(boolean condition, String conditionText[, String detailMessage][, int numCallsBelowRequirer])
 * <li> void requirement(boolean condition, String conditionText, value0, String name0, value1, String name1, ... [, int numCallsBelowRequirer])
 * </ul>
 * <ul>
 * <li> void statementNeverExecuted([int numCallsBelowRequirer])
 * <li> void statementNeverExecuted(String statementDescription[, int numCallsBelowRequirer])
 * <li> void execptionNeverCaught(Exception caughtException[, int numCallsBelowRequirer])
 * <li> void exceptionNeverCaught(String tryStatementDescription, Exception caughtException[, int numCallsBelowRequirer])
 * <li> void valueNeverOccurs(value, String name[, int numCallsBelowRequirer])
 * <li> void valuesNeverOccur(value0, name0, value1, name1, ... [, int numCallsBelowRequirer])
 * </ul>
 * <ul>
 * <li> void holdsLock/notHoldsLock(Object, String name[, int numCallsBelowRequirer])
 * </ul>
 * <ul>
 * <li> void instanceOf/notInstanceOf(Object, String name, Class type[, int numCallsBelowRequirer])
 * </ul>
 * <ul>
 * <li> void eq/neq(boolean/char/byte/short/int/long/float/double, String name0, boolean/char/byte/short/int/long/float/double[, String name1][, int numCallsBelowRequirer])
 * <li> void lt/leq/gt/geq(char/byte/short/int/long/float/double, String name0, char/byte/short/int/long/float/double[, String name1][, int numCallsBelowRequirer])
 * </ul>
 * <ul>
 * <li> void eqFalse/neqFalse/eqTrue/neqTrue(boolean, String name[, int numCallsBelowRequirer])
 * <li> void eqZero/neqZero(char/byte/short/int/long/float/double, String name[, int numCallsBelowRequirer])
 * <li> void ltZero/leqZero/gtZero/geqZero(byte/short/int/long/float/double, String name[, int numCallsBelowRequirer])
 * </ul>
 * <ul>
 * <li> void eq/neq(Object, name0, Object[, name1][, int numCallsBelowRequirer])
 * <li> void eqNull/neqNull(Object, String name[, int numCallsBelowRequirer])
 * </ul>
 * <ul>
 * <li> void equals(Object, String name0, Object, String name1[, int numCallsBelowRequirer])
 * <li> void nonempty(String, String name[, int numCallsBelowRequirer])
 * </ul>
 * <p> Naming Rationale:
 * <ul>
 * <li> eq, neq, lt, leq, gt, get correspond to ==, !=, <, <=, >, >=, e.g.,
 *      <ul>
 *      <li> For Object a and b,
 *           Require.eq(a, "a", b, "b") corresponds to require (a == b)
 *      <li> For Object o,
 *           Require.neqNull(o, "o") corresponds to require (o != null)
 *      <li> for int x,
 *           Require.eqZero(x, "x") corresponds to require (x == 0)
 *      </ul>
 * <li> equals corresponds to Object.equals (preceded by necessary null checks), e.g.,
 *      <ul>
 *      <li> For Object a and b,
 *           Require.equals(a, "a", b, "b") corresponds to require (a!= null && b != null && a.equals(b))
 *      <li> for String s,
 *           Require.nonempty(s, "s") corresponds to require (s != null && s.length() != 0)
 *      </ul>
 * </ul>
 * @author  Created by Louis Thomas on Mar 9, 2006
 */
public final class Require {

    //################################################################
    // Handle failed requirements

    //----------------------------------------------------------------
    private static void fail(String conditionText, int numCallsBelowRequirer) {
        throw new RequirementFailure(ExceptionMessageUtil.failureMessage("Requirement", "required", conditionText, null), numCallsBelowRequirer+1);
    }

    //----------------------------------------------------------------
    private static void fail(String conditionText, String detailMessage, int numCallsBelowRequirer) {
        throw new RequirementFailure(ExceptionMessageUtil.failureMessage("Requirement", "required", conditionText, detailMessage), numCallsBelowRequirer+1);
    }

    //################################################################
    // requirement

    //----------------------------------------------------------------
    /** require (condition, conditionText) */
    public static void requirement(boolean condition, String conditionText, int numCallsBelowRequirer) {
        if (!(condition)) {
            fail(conditionText, numCallsBelowRequirer + 1);
        }
    }

    public static void requirement(boolean condition, String conditionText) {
        requirement(condition, conditionText, 1);
    }

    //----------------------------------------------------------------
    /** require (condition, conditionText, detailMessage) */
    public static void requirement(boolean condition, String conditionText, String detailMessage, int numCallsBelowRequirer) {
        if (!(condition)) {
            fail(conditionText, detailMessage, numCallsBelowRequirer + 1);
        }
    }

    public static void requirement(boolean condition, String conditionText, String detailMessage) {
        requirement(condition, conditionText, detailMessage, 1);
    }

    //----------------------------------------------------------------
    /** require (condition, Object o0, String name0, ... ) */
    public static void requirement(boolean condition, String conditionText, Object o0, String name0, int numCallsBelowRequirer) {
        if (!(condition)) {
            fail(conditionText, ExceptionMessageUtil.valueAndName(o0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void requirement(boolean condition, String conditionText, Object o0, String name0) {
        requirement(condition, conditionText, o0, name0, 1);
    }

    public static void requirement(boolean condition, String conditionText, Object o0, String name0, Object o1, String name1, int numCallsBelowRequirer) {
        if (!(condition)) {
            fail(conditionText, ExceptionMessageUtil.valueAndName(o0, name0, o1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void requirement(boolean condition, String conditionText, Object o0, String name0, Object o1, String name1) {
        requirement(condition, conditionText, o0, name0, o1, name1, 1);
    }

    public static void requirement(boolean condition, String conditionText, Object o0, String name0, Object o1, String name1, Object o2, String name2, int numCallsBelowRequirer) {
        if (!(condition)) {
            fail(conditionText, ExceptionMessageUtil.valueAndName(o0, name0, o1, name1, o2, name2), numCallsBelowRequirer + 1);
        }
    }

    public static void requirement(boolean condition, String conditionText, Object o0, String name0, Object o1, String name1, Object o2, String name2) {
        requirement(condition, conditionText, o0, name0, o1, name1, o2, name2, 1);
    }

    public static void requirement(boolean condition, String conditionText, Object o0, String name0, Object o1, String name1, Object o2, String name2, Object o3, String name3, int numCallsBelowRequirer) {
        if (!(condition)) {
            fail(conditionText, ExceptionMessageUtil.valueAndName(o0, name0, o1, name1, o2, name2, o3, name3), numCallsBelowRequirer + 1);
        }
    }

    public static void requirement(boolean condition, String conditionText, Object o0, String name0, Object o1, String name1, Object o2, String name2, Object o3, String name3) {
        requirement(condition, conditionText, o0, name0, o1, name1, o2, name2, o3, name3, 1);
    }

    //----------------------------------------------------------------
    /** require (condition, boolean b0, String name0, ... ) */
    public static void requirement(boolean condition, String conditionText, boolean b0, String name0, int numCallsBelowRequirer) {
        if (!(condition)) {
            fail(conditionText, ExceptionMessageUtil.valueAndName(b0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void requirement(boolean condition, String conditionText, boolean b0, String name0) {
        requirement(condition, conditionText, b0, name0, 1);
    }

    public static void requirement(boolean condition, String conditionText, boolean b0, String name0, boolean b1, String name1, int numCallsBelowRequirer) {
        if (!(condition)) {
            fail(conditionText, ExceptionMessageUtil.valueAndName(b0, name0, b1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void requirement(boolean condition, String conditionText, boolean b0, String name0, boolean b1, String name1) {
        requirement(condition, conditionText, b0, name0, b1, name1, 1);
    }

    public static void requirement(boolean condition, String conditionText, boolean b0, String name0, double d1, String name1, int numCallsBelowRequirer) {
        if (!(condition)) {
            fail(conditionText, ExceptionMessageUtil.concat(ExceptionMessageUtil.valueAndName(b0, name0), ExceptionMessageUtil.valueAndName(d1, name1)), numCallsBelowRequirer + 1);
        }
    }

    public static void requirement(boolean condition, String conditionText, boolean b0, String name0, double d1, String name1) {
        requirement(condition, conditionText, b0, name0, d1, name1, 1);
    }

    public static void requirement(boolean condition, String conditionText, boolean b0, String name0, boolean b1, String name1, boolean b2, String name2, int numCallsBelowRequirer) {
        if (!(condition)) {
            fail(conditionText, ExceptionMessageUtil.valueAndName(b0, name0, b1, name1, b2, name2), numCallsBelowRequirer + 1);
        }
    }

    public static void requirement(boolean condition, String conditionText, boolean b0, String name0, boolean b1, String name1, boolean b2, String name2) {
        requirement(condition, conditionText, b0, name0, b1, name1, b2, name2, 1);
    }

    public static void requirement(boolean condition, String conditionText, boolean b0, String name0, boolean b1, String name1, boolean b2, String name2, boolean b3, String name3, int numCallsBelowRequirer) {
        if (!(condition)) {
            fail(conditionText, ExceptionMessageUtil.valueAndName(b0, name0, b1, name1, b2, name2, b3, name3), numCallsBelowRequirer + 1);
        }
    }

    public static void requirement(boolean condition, String conditionText, boolean b0, String name0, boolean b1, String name1, boolean b2, String name2, boolean b3, String name3) {
        requirement(condition, conditionText, b0, name0, b1, name1, b2, name2, b3, name3, 1);
    }

    //----------------------------------------------------------------
    /** require (condition, int i0, String name0, ... ) */
    public static void requirement(boolean condition, String conditionText, int i0, String name0, int numCallsBelowRequirer) {
        if (!(condition)) {
            fail(conditionText, ExceptionMessageUtil.valueAndName(i0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void requirement(boolean condition, String conditionText, int i0, String name0) {
        requirement(condition, conditionText, i0, name0, 1);
    }

    public static void requirement(boolean condition, String conditionText, int i0, String name0, int i1, String name1, int numCallsBelowRequirer) {
        if (!(condition)) {
            fail(conditionText, ExceptionMessageUtil.valueAndName(i0, name0, i1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void requirement(boolean condition, String conditionText, int i0, String name0, int i1, String name1) {
        requirement(condition, conditionText, i0, name0, i1, name1, 1);
    }

    //----------------------------------------------------------------
    /** require (condition, long l0, String name0, ... ) */
    public static void requirement(boolean condition, String conditionText, long l0, String name0, int numCallsBelowRequirer) {
        if (!(condition)) {
            fail(conditionText, ExceptionMessageUtil.valueAndName(l0, name0), numCallsBelowRequirer+1);
        }
    }

    public static void requirement(boolean condition, String conditionText, long l0, String name0) {
        requirement(condition, conditionText, l0, name0, 1);
    }

    //################################################################
    // statementNeverExceuted

    //----------------------------------------------------------------
    /** require (this statement is never executed) */
    public static void statementNeverExecuted(int numCallsBelowRequirer) {
        fail("statement is never executed", numCallsBelowRequirer + 1);
    }

    public static void statementNeverExecuted() {
        statementNeverExecuted(1);
    }

    //----------------------------------------------------------------
    /** require (statementDescription is never executed) */
    public static void statementNeverExecuted(String statementDescription, int numCallsBelowRequirer) {
        fail(statementDescription + " is never executed", numCallsBelowRequirer + 1);
    }

    public static void statementNeverExecuted(String statementDescription) {
        statementNeverExecuted(statementDescription, 1);
    }

    //################################################################
    // exceptionNeverCaught

    //----------------------------------------------------------------
    /** require (this exception is never caught, Exception e) */
    public static void exceptionNeverCaught(Exception e, int numCallsBelowRequirer) {
        try {
            fail(e.getClass().getName() + " is never caught", e.getClass().getName() + "(" + e.getMessage() + ") caught", numCallsBelowRequirer + 1);
        } catch (RequirementFailure requirementFailure) {
            requirementFailure.initCause(e);
            throw requirementFailure;
        }
    }

    public static void exceptionNeverCaught(Exception e) {
        exceptionNeverCaught(e, 1);
    }

    //----------------------------------------------------------------
    /** require (tryStatementDescription succeeds, Exception e) */
    public static void exceptionNeverCaught(String tryStatementDescription, Exception e, int numCallsBelowRequirer) {
        try {
            fail(tryStatementDescription + " succeeds", e.getClass().getName() + "(" + e.getMessage() + ") caught", numCallsBelowRequirer + 1);
        } catch (RequirementFailure requirementFailure) {
            requirementFailure.initCause(e);
            throw requirementFailure;
        }
    }

    public static void exceptionNeverCaught(String tryStatementDescription, Exception e) {
        exceptionNeverCaught(tryStatementDescription, e, 1);
    }

    //################################################################
    // valueNeverOccurs

    //----------------------------------------------------------------
    /** require (this value never occurs, Object o, name) */
    public static void valueNeverOccurs(Object o, String name, int numCallsBelowRequirer) {
        fail(ExceptionMessageUtil.valueAndName(o, name) + " never occurs", numCallsBelowRequirer + 1);
    }

    public static void valueNeverOccurs(Object o, String name) {
        valueNeverOccurs(o, name, 1);
    }

    //----------------------------------------------------------------
    /** require (this value never occurs, boolean b, name) */
    public static void valueNeverOccurs(boolean b, String name, int numCallsBelowRequirer) {
        fail(ExceptionMessageUtil.valueAndName(b, name) + " never occurs", numCallsBelowRequirer + 1);
    }

    public static void valueNeverOccurs(boolean b, String name) {
        valueNeverOccurs(b, name, 1);
    }

    //----------------------------------------------------------------
    /** require (this value never occurs, char c, name) */
    public static void valueNeverOccurs(char c, String name, int numCallsBelowRequirer) {
        fail(ExceptionMessageUtil.valueAndName(c, name) + " never occurs", numCallsBelowRequirer + 1);
    }

    public static void valueNeverOccurs(char c, String name) {
        valueNeverOccurs(c, name, 1);
    }

    //----------------------------------------------------------------
    /** require (this value never occurs, byte b, name) */
    public static void valueNeverOccurs(byte b, String name, int numCallsBelowRequirer) {
        fail(ExceptionMessageUtil.valueAndName(b, name) + " never occurs", numCallsBelowRequirer + 1);
    }

    public static void valueNeverOccurs(byte b, String name) {
        valueNeverOccurs(b, name, 1);
    }

    //----------------------------------------------------------------
    /** require (this value never occurs, short s, name) */
    public static void valueNeverOccurs(short s, String name, int numCallsBelowRequirer) {
        fail(ExceptionMessageUtil.valueAndName(s, name) + " never occurs", numCallsBelowRequirer + 1);
    }

    public static void valueNeverOccurs(short s, String name) {
        valueNeverOccurs(s, name, 1);
    }

    //----------------------------------------------------------------
    /** require (this value never occurs, int i, name) */
    public static void valueNeverOccurs(int i, String name, int numCallsBelowRequirer) {
        fail(ExceptionMessageUtil.valueAndName(i, name) + " never occurs", numCallsBelowRequirer + 1);
    }

    public static void valueNeverOccurs(int i, String name) {
        valueNeverOccurs(i, name, 1);
    }

    //----------------------------------------------------------------
    /** require (this value never occurs, long l, name) */
    public static void valueNeverOccurs(long l, String name, int numCallsBelowRequirer) {
        fail(ExceptionMessageUtil.valueAndName(l, name) + " never occurs", numCallsBelowRequirer + 1);
    }

    public static void valueNeverOccurs(long l, String name) {
        valueNeverOccurs(l, name, 1);
    }

    //----------------------------------------------------------------
    /** require (this value never occurs, float f, name) */
    public static void valueNeverOccurs(float f, String name, int numCallsBelowRequirer) {
        fail(ExceptionMessageUtil.valueAndName(f, name) + " never occurs", numCallsBelowRequirer + 1);
    }

    public static void valueNeverOccurs(float f, String name) {
        valueNeverOccurs(f, name, 1);
    }

    //----------------------------------------------------------------
    /** require (this value never occurs, double d, name) */
    public static void valueNeverOccurs(double d, String name, int numCallsBelowRequirer) {
        fail(ExceptionMessageUtil.valueAndName(d, name) + " never occurs", numCallsBelowRequirer + 1);
    }

    public static void valueNeverOccurs(double d, String name) {
        valueNeverOccurs(d, name, 1);
    }

    //################################################################
    // holdsLock, notHoldsLock

    //----------------------------------------------------------------
    /** require (o != null && (current thread holds o's lock)) */
    public static void holdsLock(Object o, String name, int numCallsBelowRequirer) {
        neqNull(o, "o");
        if (!Thread.holdsLock(o)) {
            fail("\"" + Thread.currentThread().getName() + "\".holdsLock(" + name + ")", numCallsBelowRequirer + 1);
        }
    }

    public static void holdsLock(Object o, String name) {
        holdsLock(o, name, 1);
    }

    //----------------------------------------------------------------
    /** require (o != null && !(current thread holds o's lock)) */
    public static void notHoldsLock(Object o, String name, int numCallsBelowRequirer) {
        neqNull(o, "o");
        if (Thread.holdsLock(o)) {
            fail("!\"" + Thread.currentThread().getName() + "\".holdsLock(" + name + ")", numCallsBelowRequirer + 1);
        }
    }

    public static void notHoldsLock(Object o, String name) {
        notHoldsLock(o, name, 1);
    }


    //################################################################
    // instanceOf, notInstanceOf

    //----------------------------------------------------------------
    /** require (o instanceof type) */
    public static <T> void instanceOf(Object o, String name, Class<T> type, int numCallsBelowRequirer) {
        if (!type.isInstance(o)) {
            fail(name+" instanceof "+type, null==o?ExceptionMessageUtil.valueAndName(o, name):name+" instanceof "+o.getClass()+" ("+ExceptionMessageUtil.valueAndName(o, name)+")", numCallsBelowRequirer + 1);
        }
    }

    public static <T> void instanceOf(Object o, String name, Class<T> type) {
        instanceOf(o, name, type, 1);
    }

    //----------------------------------------------------------------
    /** require !(o instanceof type) */
    public static <T> void notInstanceOf(Object o, String name, Class<T> type, int numCallsBelowRequirer) {
        if (type.isInstance(o)) {
            fail("!("+name+" instanceof "+type+")", name+" instanceof "+o.getClass()+" ("+ExceptionMessageUtil.valueAndName(o, name)+")", numCallsBelowRequirer + 1);
        }
    }

    public static <T> void notInstanceOf(Object o, String name, Class<T> type) {
        notInstanceOf(o, name, type, 1);
    }

    //################################################################
    // isAWTThread, isNotAWTThread

    //----------------------------------------------------------------
    /** require (current thread is AWT Event Dispatch Thread) */
    public static void isAWTThread() {
        isAWTThread(1);
    }

    public static void isAWTThread(int numCallsBelowRequirer) {
        if (!EventQueue.isDispatchThread()) {
            fail("\""+Thread.currentThread().getName() + "\".isAWTThread()", numCallsBelowRequirer + 1);
        }
    }

    //----------------------------------------------------------------
    /** require (current thread is AWT Event Dispatch Thread) */
    public static void isNotAWTThread() {
        isNotAWTThread(1);
    }

    public static void isNotAWTThread(int numCallsBelowRequirer) {
        if (EventQueue.isDispatchThread()) {
            fail("!\"" + Thread.currentThread().getName() + "\".isAWTThread()", numCallsBelowRequirer + 1);
        }
    }

    //################################################################
    // eq (primitiveValue == primitiveValue)

    //----------------------------------------------------------------
    /** require (b0 == b1) */
    public static void eq(boolean b0, String name0, boolean b1, String name1, int numCallsBelowRequirer) {
        if (!(b0 == b1)) {
            fail(name0 + " == " + name1, ExceptionMessageUtil.valueAndName(b0, name0, b1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(boolean b0, String name0, boolean b1, String name1) {
        eq(b0, name0, b1, name1, 1);
    }

    public static void eq(boolean b0, String name0, boolean b1, int numCallsBelowRequirer) {
        if (!(b0 == b1)) {
            fail(name0 + " == " + b1, ExceptionMessageUtil.valueAndName(b0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(boolean b0, String name0, boolean b1) {
        eq(b0, name0, b1, 1);
    }

    //----------------------------------------------------------------
    /** require (c0 == c1) */
    public static void eq(char c0, String name0, char c1, String name1, int numCallsBelowRequirer) {
        if (!(c0 == c1)) {
            fail(name0 + " == " + name1, ExceptionMessageUtil.valueAndName(c0, name0, c1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(char c0, String name0, char c1, String name1) {
        eq(c0, name0, c1, name1, 1);
    }

    public static void eq(char c0, String name0, char c1, int numCallsBelowRequirer) {
        if (!(c0 == c1)) {
            fail(name0 + " == " + ExceptionMessageUtil.valueString(c1), ExceptionMessageUtil.valueAndName(c0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(char c0, String name0, char c1) {
        eq(c0, name0, c1, 1);
    }

    //----------------------------------------------------------------
    /** require (b0 == b1) */
    public static void eq(byte b0, String name0, byte b1, String name1, int numCallsBelowRequirer) {
        if (!(b0 == b1)) {
            fail(name0 + " == " + name1, ExceptionMessageUtil.valueAndName(b0, name0, b1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(byte b0, String name0, byte b1, String name1) {
        eq(b0, name0, b1, name1, 1);
    }

    public static void eq(byte b0, String name0, byte b1, int numCallsBelowRequirer) {
        if (!(b0 == b1)) {
            fail(name0 + " == " + b1, ExceptionMessageUtil.valueAndName(b0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(byte b0, String name0, byte b1) {
        eq(b0, name0, b1, 1);
    }

    //----------------------------------------------------------------
    /** require (s0 == s1) */
    public static void eq(short s0, String name0, short s1, String name1, int numCallsBelowRequirer) {
        if (!(s0 == s1)) {
            fail(name0 + " == " + name1, ExceptionMessageUtil.valueAndName(s0, name0, s1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(short s0, String name0, short s1, String name1) {
        eq(s0, name0, s1, name1, 1);
    }

    public static void eq(short s0, String name0, short s1, int numCallsBelowRequirer) {
        if (!(s0 == s1)) {
            fail(name0 + " == " + s1, ExceptionMessageUtil.valueAndName(s0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(short s0, String name0, short s1) {
        eq(s0, name0, s1, 1);
    }

    //----------------------------------------------------------------
    /** require (i0 == i1) */
    public static void eq(int i0, String name0, int i1, String name1, int numCallsBelowRequirer) {
        if (!(i0 == i1)) {
            fail(name0 + " == " + name1, ExceptionMessageUtil.valueAndName(i0, name0, i1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(int i0, String name0, int i1, String name1) {
        eq(i0, name0, i1, name1, 1);
    }

    public static void eq(int i0, String name0, int i1, int numCallsBelowRequirer) {
        if (!(i0 == i1)) {
            fail(name0 + " == " + i1, ExceptionMessageUtil.valueAndName(i0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(int i0, String name0, int i1) {
        eq(i0, name0, i1, 1);
    }

    //----------------------------------------------------------------
    /** require (l0 == l1) */
    public static void eq(long l0, String name0, long l1, String name1, int numCallsBelowRequirer) {
        if (!(l0 == l1)) {
            fail(name0 + " == " + name1, ExceptionMessageUtil.valueAndName(l0, name0, l1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(long l0, String name0, long l1, String name1) {
        eq(l0, name0, l1, name1, 1);
    }

    public static void eq(long l0, String name0, long l1, int numCallsBelowRequirer) {
        if (!(l0 == l1)) {
            fail(name0 + " == " + l1, ExceptionMessageUtil.valueAndName(l0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(long l0, String name0, long l1) {
        eq(l0, name0, l1, 1);
    }

    //----------------------------------------------------------------
    /** require (f0 == f1) */
    public static void eq(float f0, String name0, float f1, String name1, int numCallsBelowRequirer) {
        if (!(f0 == f1)) {
            fail(name0 + " == " + name1, ExceptionMessageUtil.valueAndName(f0, name0, f1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(float f0, String name0, float f1, String name1) {
        eq(f0, name0, f1, name1, 1);
    }

    public static void eq(float f0, String name0, float f1, int numCallsBelowRequirer) {
        if (!(f0 == f1)) {
            fail(name0 + " == " + f1, ExceptionMessageUtil.valueAndName(f0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(float f0, String name0, float f1) {
        eq(f0, name0, f1, 1);
    }

    //----------------------------------------------------------------
    /** require (d0 == d1) */
    public static void eq(double d0, String name0, double d1, String name1, int numCallsBelowRequirer) {
        if (!(d0 == d1)) {
            fail(name0 + " == " + name1, ExceptionMessageUtil.valueAndName(d0, name0, d1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(double d0, String name0, double d1, String name1) {
        eq(d0, name0, d1, name1, 1);
    }

    public static void eq(double d0, String name0, double d1, int numCallsBelowRequirer) {
        if (!(d0 == d1)) {
            fail(name0 + " == " + d1, ExceptionMessageUtil.valueAndName(d0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(double d0, String name0, double d1) {
        eq(d0, name0, d1, 1);
    }

    //################################################################
    // neq (primitiveValue != primitiveValue)

    //----------------------------------------------------------------
    /** require (b0 != b1) */
    public static void neq(boolean b0, String name0, boolean b1, String name1, int numCallsBelowRequirer) {
        if (!(b0 != b1)) {
            fail(name0 + " != " + name1, ExceptionMessageUtil.valueAndName(b0, name0, b1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(boolean b0, String name0, boolean b1, String name1) {
        neq(b0, name0, b1, name1, 1);
    }

    public static void neq(boolean b0, String name0, boolean b1, int numCallsBelowRequirer) {
        if (!(b0 != b1)) {
            fail(name0 + " != " + b1, ExceptionMessageUtil.valueAndName(b0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(boolean b0, String name0, boolean b1) {
        neq(b0, name0, b1, 1);
    }

    //----------------------------------------------------------------
    /** require (c0 != c1) */
    public static void neq(char c0, String name0, char c1, String name1, int numCallsBelowRequirer) {
        if (!(c0 != c1)) {
            fail(name0 + " != " + name1, ExceptionMessageUtil.valueAndName(c0, name0, c1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(char c0, String name0, char c1, String name1) {
        neq(c0, name0, c1, name1, 1);
    }

    public static void neq(char c0, String name0, char c1, int numCallsBelowRequirer) {
        if (!(c0 != c1)) {
            fail(name0 + " != " + ExceptionMessageUtil.valueString(c1), ExceptionMessageUtil.valueAndName(c0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(char c0, String name0, char c1) {
        neq(c0, name0, c1, 1);
    }

    //----------------------------------------------------------------
    /** require (b0 != b1) */
    public static void neq(byte b0, String name0, byte b1, String name1, int numCallsBelowRequirer) {
        if (!(b0 != b1)) {
            fail(name0 + " != " + name1, ExceptionMessageUtil.valueAndName(b0, name0, b1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(byte b0, String name0, byte b1, String name1) {
        neq(b0, name0, b1, name1, 1);
    }

    public static void neq(byte b0, String name0, byte b1, int numCallsBelowRequirer) {
        if (!(b0 != b1)) {
            fail(name0 + " != " + b1, ExceptionMessageUtil.valueAndName(b0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(byte b0, String name0, byte b1) {
        neq(b0, name0, b1, 1);
    }

    //----------------------------------------------------------------
    /** require (s0 != s1) */
    public static void neq(short s0, String name0, short s1, String name1, int numCallsBelowRequirer) {
        if (!(s0 != s1)) {
            fail(name0 + " != " + name1, ExceptionMessageUtil.valueAndName(s0, name0, s1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(short s0, String name0, short s1, String name1) {
        neq(s0, name0, s1, name1, 1);
    }

    public static void neq(short s0, String name0, short s1, int numCallsBelowRequirer) {
        if (!(s0 != s1)) {
            fail(name0 + " != " + s1, ExceptionMessageUtil.valueAndName(s0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(short s0, String name0, short s1) {
        neq(s0, name0, s1, 1);
    }

    //----------------------------------------------------------------
    /** require (i0 != i1) */
    public static void neq(int i0, String name0, int i1, String name1, int numCallsBelowRequirer) {
        if (!(i0 != i1)) {
            fail(name0 + " != " + name1, ExceptionMessageUtil.valueAndName(i0, name0, i1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(int i0, String name0, int i1, String name1) {
        neq(i0, name0, i1, name1, 1);
    }

    public static void neq(int i0, String name0, int i1, int numCallsBelowRequirer) {
        if (!(i0 != i1)) {
            fail(name0 + " != " + i1, ExceptionMessageUtil.valueAndName(i0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(int i0, String name0, int i1) {
        neq(i0, name0, i1, 1);
    }

    //----------------------------------------------------------------
    /** require (l0 != l1) */
    public static void neq(long l0, String name0, long l1, String name1, int numCallsBelowRequirer) {
        if (!(l0 != l1)) {
            fail(name0 + " != " + name1, ExceptionMessageUtil.valueAndName(l0, name0, l1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(long l0, String name0, long l1, String name1) {
        neq(l0, name0, l1, name1, 1);
    }

    public static void neq(long l0, String name0, long l1, int numCallsBelowRequirer) {
        if (!(l0 != l1)) {
            fail(name0 + " != " + l1, ExceptionMessageUtil.valueAndName(l0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(long l0, String name0, long l1) {
        neq(l0, name0, l1, 1);
    }

    //----------------------------------------------------------------
    /** require (f0 != f1) */
    public static void neq(float f0, String name0, float f1, String name1, int numCallsBelowRequirer) {
        if (!(f0 != f1)) {
            fail(name0 + " != " + name1, ExceptionMessageUtil.valueAndName(f0, name0, f1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(float f0, String name0, float f1, String name1) {
        neq(f0, name0, f1, name1, 1);
    }

    public static void neq(float f0, String name0, float f1, int numCallsBelowRequirer) {
        if (!(f0 != f1)) {
            fail(name0 + " != " + f1, ExceptionMessageUtil.valueAndName(f0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(float f0, String name0, float f1) {
        neq(f0, name0, f1, 1);
    }

    //----------------------------------------------------------------
    /** require (d0 != d1) */
    public static void neq(double d0, String name0, double d1, String name1, int numCallsBelowRequirer) {
        if (!(d0 != d1)) {
            fail(name0 + " != " + name1, ExceptionMessageUtil.valueAndName(d0, name0, d1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(double d0, String name0, double d1, String name1) {
        neq(d0, name0, d1, name1, 1);
    }

    public static void neq(double d0, String name0, double d1, int numCallsBelowRequirer) {
        if (!(d0 != d1)) {
            fail(name0 + " != " + d1, ExceptionMessageUtil.valueAndName(d0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(double d0, String name0, double d1) {
        neq(d0, name0, d1, 1);
    }

    //################################################################
    // lt (primitiveValue < primitiveValue)

    //----------------------------------------------------------------
    /** require (c0 < c1) */
    public static void lt(char c0, String name0, char c1, String name1, int numCallsBelowRequirer) {
        if (!(c0 < c1)) {
            fail(name0 + " < " + name1, ExceptionMessageUtil.valueAndName(c0, name0, c1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void lt(char c0, String name0, char c1, String name1) {
        lt(c0, name0, c1, name1, 1);
    }

    public static void lt(char c0, String name0, char c1, int numCallsBelowRequirer) {
        if (!(c0 < c1)) {
            fail(name0 + " < " + ExceptionMessageUtil.valueString(c1), ExceptionMessageUtil.valueAndName(c0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void lt(char c0, String name0, char c1) {
        lt(c0, name0, c1, 1);
    }

    //----------------------------------------------------------------
    /** require (b0 < b1) */
    public static void lt(byte b0, String name0, byte b1, String name1, int numCallsBelowRequirer) {
        if (!(b0 < b1)) {
            fail(name0 + " < " + name1, ExceptionMessageUtil.valueAndName(b0, name0, b1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void lt(byte b0, String name0, byte b1, String name1) {
        lt(b0, name0, b1, name1, 1);
    }

    public static void lt(byte b0, String name0, byte b1, int numCallsBelowRequirer) {
        if (!(b0 < b1)) {
            fail(name0 + " < " + b1, ExceptionMessageUtil.valueAndName(b0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void lt(byte b0, String name0, byte b1) {
        lt(b0, name0, b1, 1);
    }

    //----------------------------------------------------------------
    /** require (s0 < s1) */
    public static void lt(short s0, String name0, short s1, String name1, int numCallsBelowRequirer) {
        if (!(s0 < s1)) {
            fail(name0 + " < " + name1, ExceptionMessageUtil.valueAndName(s0, name0, s1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void lt(short s0, String name0, short s1, String name1) {
        lt(s0, name0, s1, name1, 1);
    }

    public static void lt(short s0, String name0, short s1, int numCallsBelowRequirer) {
        if (!(s0 < s1)) {
            fail(name0 + " < " + s1, ExceptionMessageUtil.valueAndName(s0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void lt(short s0, String name0, short s1) {
        lt(s0, name0, s1, 1);
    }

    //----------------------------------------------------------------
    /** require (i0 < i1) */
    public static void lt(int i0, String name0, int i1, String name1, int numCallsBelowRequirer) {
        if (!(i0 < i1)) {
            fail(name0 + " < " + name1, ExceptionMessageUtil.valueAndName(i0, name0, i1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void lt(int i0, String name0, int i1, String name1) {
        lt(i0, name0, i1, name1, 1);
    }

    public static void lt(int i0, String name0, int i1, int numCallsBelowRequirer) {
        if (!(i0 < i1)) {
            fail(name0 + " < " + i1, ExceptionMessageUtil.valueAndName(i0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void lt(int i0, String name0, int i1) {
        lt(i0, name0, i1, 1);
    }

    //----------------------------------------------------------------
    /** require (l0 < l1) */
    public static void lt(long l0, String name0, long l1, String name1, int numCallsBelowRequirer) {
        if (!(l0 < l1)) {
            fail(name0 + " < " + name1, ExceptionMessageUtil.valueAndName(l0, name0, l1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void lt(long l0, String name0, long l1, String name1) {
        lt(l0, name0, l1, name1, 1);
    }

    public static void lt(long l0, String name0, long l1, int numCallsBelowRequirer) {
        if (!(l0 < l1)) {
            fail(name0 + " < " + l1, ExceptionMessageUtil.valueAndName(l0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void lt(long l0, String name0, long l1) {
        lt(l0, name0, l1, 1);
    }

    //----------------------------------------------------------------
    /** require (f0 < f1) */
    public static void lt(float f0, String name0, float f1, String name1, int numCallsBelowRequirer) {
        if (!(f0 < f1)) {
            fail(name0 + " < " + name1, ExceptionMessageUtil.valueAndName(f0, name0, f1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void lt(float f0, String name0, float f1, String name1) {
        lt(f0, name0, f1, name1, 1);
    }

    public static void lt(float f0, String name0, float f1, int numCallsBelowRequirer) {
        if (!(f0 < f1)) {
            fail(name0 + " < " + f1, ExceptionMessageUtil.valueAndName(f0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void lt(float f0, String name0, float f1) {
        lt(f0, name0, f1, 1);
    }

    //----------------------------------------------------------------
    /** require (d0 < d1) */
    public static void lt(double d0, String name0, double d1, String name1, int numCallsBelowRequirer) {
        if (!(d0 < d1)) {
            fail(name0 + " < " + name1, ExceptionMessageUtil.valueAndName(d0, name0, d1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void lt(double d0, String name0, double d1, String name1) {
        lt(d0, name0, d1, name1, 1);
    }

    public static void lt(double d0, String name0, double d1, int numCallsBelowRequirer) {
        if (!(d0 < d1)) {
            fail(name0 + " < " + d1, ExceptionMessageUtil.valueAndName(d0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void lt(double d0, String name0, double d1) {
        lt(d0, name0, d1, 1);
    }

    //################################################################
    // leq (primitiveValue <= primitiveValue)

    //----------------------------------------------------------------
    /** require (c0 <= c1) */
    public static void leq(char c0, String name0, char c1, String name1, int numCallsBelowRequirer) {
        if (!(c0 <= c1)) {
            fail(name0 + " <= " + name1, ExceptionMessageUtil.valueAndName(c0, name0, c1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void leq(char c0, String name0, char c1, String name1) {
        leq(c0, name0, c1, name1, 1);
    }

    public static void leq(char c0, String name0, char c1, int numCallsBelowRequirer) {
        if (!(c0 <= c1)) {
            fail(name0 + " <= " + ExceptionMessageUtil.valueString(c1), ExceptionMessageUtil.valueAndName(c0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void leq(char c0, String name0, char c1) {
        leq(c0, name0, c1, 1);
    }

    //----------------------------------------------------------------
    /** require (b0 <= b1) */
    public static void leq(byte b0, String name0, byte b1, String name1, int numCallsBelowRequirer) {
        if (!(b0 <= b1)) {
            fail(name0 + " <= " + name1, ExceptionMessageUtil.valueAndName(b0, name0, b1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void leq(byte b0, String name0, byte b1, String name1) {
        leq(b0, name0, b1, name1, 1);
    }

    public static void leq(byte b0, String name0, byte b1, int numCallsBelowRequirer) {
        if (!(b0 <= b1)) {
            fail(name0 + " <= " + b1, ExceptionMessageUtil.valueAndName(b0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void leq(byte b0, String name0, byte b1) {
        leq(b0, name0, b1, 1);
    }

    //----------------------------------------------------------------
    /** require (s0 <= s1) */
    public static void leq(short s0, String name0, short s1, String name1, int numCallsBelowRequirer) {
        if (!(s0 <= s1)) {
            fail(name0 + " <= " + name1, ExceptionMessageUtil.valueAndName(s0, name0, s1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void leq(short s0, String name0, short s1, String name1) {
        leq(s0, name0, s1, name1, 1);
    }

    public static void leq(short s0, String name0, short s1, int numCallsBelowRequirer) {
        if (!(s0 <= s1)) {
            fail(name0 + " <= " + s1, ExceptionMessageUtil.valueAndName(s0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void leq(short s0, String name0, short s1) {
        leq(s0, name0, s1, 1);
    }

    //----------------------------------------------------------------
    /** require (i0 <= i1) */
    public static void leq(int i0, String name0, int i1, String name1, int numCallsBelowRequirer) {
        if (!(i0 <= i1)) {
            fail(name0 + " <= " + name1, ExceptionMessageUtil.valueAndName(i0, name0, i1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void leq(int i0, String name0, int i1, String name1) {
        leq(i0, name0, i1, name1, 1);
    }

    public static void leq(int i0, String name0, int i1, int numCallsBelowRequirer) {
        if (!(i0 <= i1)) {
            fail(name0 + " <= " + i1, ExceptionMessageUtil.valueAndName(i0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void leq(int i0, String name0, int i1) {
        leq(i0, name0, i1, 1);
    }

    //----------------------------------------------------------------
    /** require (l0 <= l1) */
    public static void leq(long l0, String name0, long l1, String name1, int numCallsBelowRequirer) {
        if (!(l0 <= l1)) {
            fail(name0 + " <= " + name1, ExceptionMessageUtil.valueAndName(l0, name0, l1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void leq(long l0, String name0, long l1, String name1) {
        leq(l0, name0, l1, name1, 1);
    }

    public static void leq(long l0, String name0, long l1, int numCallsBelowRequirer) {
        if (!(l0 <= l1)) {
            fail(name0 + " <= " + l1, ExceptionMessageUtil.valueAndName(l0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void leq(long l0, String name0, long l1) {
        leq(l0, name0, l1, 1);
    }

    //----------------------------------------------------------------
    /** require (f0 <= f1) */
    public static void leq(float f0, String name0, float f1, String name1, int numCallsBelowRequirer) {
        if (!(f0 <= f1)) {
            fail(name0 + " <= " + name1, ExceptionMessageUtil.valueAndName(f0, name0, f1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void leq(float f0, String name0, float f1, String name1) {
        leq(f0, name0, f1, name1, 1);
    }

    public static void leq(float f0, String name0, float f1, int numCallsBelowRequirer) {
        if (!(f0 <= f1)) {
            fail(name0 + " <= " + f1, ExceptionMessageUtil.valueAndName(f0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void leq(float f0, String name0, float f1) {
        leq(f0, name0, f1, 1);
    }

    //----------------------------------------------------------------
    /** require (d0 <= d1) */
    public static void leq(double d0, String name0, double d1, String name1, int numCallsBelowRequirer) {
        if (!(d0 <= d1)) {
            fail(name0 + " <= " + name1, ExceptionMessageUtil.valueAndName(d0, name0, d1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void leq(double d0, String name0, double d1, String name1) {
        leq(d0, name0, d1, name1, 1);
    }

    public static void leq(double d0, String name0, double d1, int numCallsBelowRequirer) {
        if (!(d0 <= d1)) {
            fail(name0 + " <= " + d1, ExceptionMessageUtil.valueAndName(d0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void leq(double d0, String name0, double d1) {
        leq(d0, name0, d1, 1);
    }

    //################################################################
    // gt (primitiveValue > primitiveValue)

    //----------------------------------------------------------------
    /** require (c0 > c1) */
    public static void gt(char c0, String name0, char c1, String name1, int numCallsBelowRequirer) {
        if (!(c0 > c1)) {
            fail(name0 + " > " + name1, ExceptionMessageUtil.valueAndName(c0, name0, c1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void gt(char c0, String name0, char c1, String name1) {
        gt(c0, name0, c1, name1, 1);
    }

    public static void gt(char c0, String name0, char c1, int numCallsBelowRequirer) {
        if (!(c0 > c1)) {
            fail(name0 + " > " + ExceptionMessageUtil.valueString(c1), ExceptionMessageUtil.valueAndName(c0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void gt(char c0, String name0, char c1) {
        gt(c0, name0, c1, 1);
    }

    //----------------------------------------------------------------
    /** require (b0 > b1) */
    public static void gt(byte b0, String name0, byte b1, String name1, int numCallsBelowRequirer) {
        if (!(b0 > b1)) {
            fail(name0 + " > " + name1, ExceptionMessageUtil.valueAndName(b0, name0, b1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void gt(byte b0, String name0, byte b1, String name1) {
        gt(b0, name0, b1, name1, 1);
    }

    public static void gt(byte b0, String name0, byte b1, int numCallsBelowRequirer) {
        if (!(b0 > b1)) {
            fail(name0 + " > " + b1, ExceptionMessageUtil.valueAndName(b0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void gt(byte b0, String name0, byte b1) {
        gt(b0, name0, b1, 1);
    }

    //----------------------------------------------------------------
    /** require (s0 > s1) */
    public static void gt(short s0, String name0, short s1, String name1, int numCallsBelowRequirer) {
        if (!(s0 > s1)) {
            fail(name0 + " > " + name1, ExceptionMessageUtil.valueAndName(s0, name0, s1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void gt(short s0, String name0, short s1, String name1) {
        gt(s0, name0, s1, name1, 1);
    }

    public static void gt(short s0, String name0, short s1, int numCallsBelowRequirer) {
        if (!(s0 > s1)) {
            fail(name0 + " > " + s1, ExceptionMessageUtil.valueAndName(s0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void gt(short s0, String name0, short s1) {
        gt(s0, name0, s1, 1);
    }

    //----------------------------------------------------------------
    /** require (i0 > i1) */
    public static void gt(int i0, String name0, int i1, String name1, int numCallsBelowRequirer) {
        if (!(i0 > i1)) {
            fail(name0 + " > " + name1, ExceptionMessageUtil.valueAndName(i0, name0, i1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void gt(int i0, String name0, int i1, String name1) {
        gt(i0, name0, i1, name1, 1);
    }

    public static void gt(int i0, String name0, int i1, int numCallsBelowRequirer) {
        if (!(i0 > i1)) {
            fail(name0 + " > " + i1, ExceptionMessageUtil.valueAndName(i0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void gt(int i0, String name0, int i1) {
        gt(i0, name0, i1, 1);
    }

    //----------------------------------------------------------------
    /** require (l0 > l1) */
    public static void gt(long l0, String name0, long l1, String name1, int numCallsBelowRequirer) {
        if (!(l0 > l1)) {
            fail(name0 + " > " + name1, ExceptionMessageUtil.valueAndName(l0, name0, l1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void gt(long l0, String name0, long l1, String name1) {
        gt(l0, name0, l1, name1, 1);
    }

    public static void gt(long l0, String name0, long l1, int numCallsBelowRequirer) {
        if (!(l0 > l1)) {
            fail(name0 + " > " + l1, ExceptionMessageUtil.valueAndName(l0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void gt(long l0, String name0, long l1) {
        gt(l0, name0, l1, 1);
    }

    //----------------------------------------------------------------
    /** require (f0 > f1) */
    public static void gt(float f0, String name0, float f1, String name1, int numCallsBelowRequirer) {
        if (!(f0 > f1)) {
            fail(name0 + " > " + name1, ExceptionMessageUtil.valueAndName(f0, name0, f1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void gt(float f0, String name0, float f1, String name1) {
        gt(f0, name0, f1, name1, 1);
    }

    public static void gt(float f0, String name0, float f1, int numCallsBelowRequirer) {
        if (!(f0 > f1)) {
            fail(name0 + " > " + f1, ExceptionMessageUtil.valueAndName(f0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void gt(float f0, String name0, float f1) {
        gt(f0, name0, f1, 1);
    }

    //----------------------------------------------------------------
    /** require (d0 > d1) */
    public static void gt(double d0, String name0, double d1, String name1, int numCallsBelowRequirer) {
        if (!(d0 > d1)) {
            fail(name0 + " > " + name1, ExceptionMessageUtil.valueAndName(d0, name0, d1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void gt(double d0, String name0, double d1, String name1) {
        gt(d0, name0, d1, name1, 1);
    }

    public static void gt(double d0, String name0, double d1, int numCallsBelowRequirer) {
        if (!(d0 > d1)) {
            fail(name0 + " > " + d1, ExceptionMessageUtil.valueAndName(d0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void gt(double d0, String name0, double d1) {
        gt(d0, name0, d1, 1);
    }

    //################################################################
    // geq (primitiveValue >= primitiveValue)

    //----------------------------------------------------------------
    /** require (c0 >= c1) */
    public static void geq(char c0, String name0, char c1, String name1, int numCallsBelowRequirer) {
        if (!(c0 >= c1)) {
            fail(name0 + " >= " + name1, ExceptionMessageUtil.valueAndName(c0, name0, c1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void geq(char c0, String name0, char c1, String name1) {
        geq(c0, name0, c1, name1, 1);
    }

    public static void geq(char c0, String name0, char c1, int numCallsBelowRequirer) {
        if (!(c0 >= c1)) {
            fail(name0 + " >= " + ExceptionMessageUtil.valueString(c1), ExceptionMessageUtil.valueAndName(c0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void geq(char c0, String name0, char c1) {
        geq(c0, name0, c1, 1);
    }

    //----------------------------------------------------------------
    /** require (b0 >= b1) */
    public static void geq(byte b0, String name0, byte b1, String name1, int numCallsBelowRequirer) {
        if (!(b0 >= b1)) {
            fail(name0 + " >= " + name1, ExceptionMessageUtil.valueAndName(b0, name0, b1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void geq(byte b0, String name0, byte b1, String name1) {
        geq(b0, name0, b1, name1, 1);
    }

    public static void geq(byte b0, String name0, byte b1, int numCallsBelowRequirer) {
        if (!(b0 >= b1)) {
            fail(name0 + " >= " + b1, ExceptionMessageUtil.valueAndName(b0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void geq(byte b0, String name0, byte b1) {
        geq(b0, name0, b1, 1);
    }

    //----------------------------------------------------------------
    /** require (s0 >= s1) */
    public static void geq(short s0, String name0, short s1, String name1, int numCallsBelowRequirer) {
        if (!(s0 >= s1)) {
            fail(name0 + " >= " + name1, ExceptionMessageUtil.valueAndName(s0, name0, s1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void geq(short s0, String name0, short s1, String name1) {
        geq(s0, name0, s1, name1, 1);
    }

    public static void geq(short s0, String name0, short s1, int numCallsBelowRequirer) {
        if (!(s0 >= s1)) {
            fail(name0 + " >= " + s1, ExceptionMessageUtil.valueAndName(s0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void geq(short s0, String name0, short s1) {
        geq(s0, name0, s1, 1);
    }

    //----------------------------------------------------------------
    /** require (i0 >= i1) */
    public static void geq(int i0, String name0, int i1, String name1, int numCallsBelowRequirer) {
        if (!(i0 >= i1)) {
            fail(name0 + " >= " + name1, ExceptionMessageUtil.valueAndName(i0, name0, i1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void geq(int i0, String name0, int i1, String name1) {
        geq(i0, name0, i1, name1, 1);
    }

    public static void geq(int i0, String name0, int i1, int numCallsBelowRequirer) {
        if (!(i0 >= i1)) {
            fail(name0 + " >= " + i1, ExceptionMessageUtil.valueAndName(i0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void geq(int i0, String name0, int i1) {
        geq(i0, name0, i1, 1);
    }

    //----------------------------------------------------------------
    /** require (l0 >= l1) */
    public static void geq(long l0, String name0, long l1, String name1, int numCallsBelowRequirer) {
        if (!(l0 >= l1)) {
            fail(name0 + " >= " + name1, ExceptionMessageUtil.valueAndName(l0, name0, l1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void geq(long l0, String name0, long l1, String name1) {
        geq(l0, name0, l1, name1, 1);
    }

    public static void geq(long l0, String name0, long l1, int numCallsBelowRequirer) {
        if (!(l0 >= l1)) {
            fail(name0 + " >= " + l1, ExceptionMessageUtil.valueAndName(l0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void geq(long l0, String name0, long l1) {
        geq(l0, name0, l1, 1);
    }

    //----------------------------------------------------------------
    /** require (f0 >= f1) */
    public static void geq(float f0, String name0, float f1, String name1, int numCallsBelowRequirer) {
        if (!(f0 >= f1)) {
            fail(name0 + " >= " + name1, ExceptionMessageUtil.valueAndName(f0, name0, f1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void geq(float f0, String name0, float f1, String name1) {
        geq(f0, name0, f1, name1, 1);
    }

    public static void geq(float f0, String name0, float f1, int numCallsBelowRequirer) {
        if (!(f0 >= f1)) {
            fail(name0 + " >= " + f1, ExceptionMessageUtil.valueAndName(f0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void geq(float f0, String name0, float f1) {
        geq(f0, name0, f1, 1);
    }

    //----------------------------------------------------------------
    /** require (d0 >= d1) */
    public static void geq(double d0, String name0, double d1, String name1, int numCallsBelowRequirer) {
        if (!(d0 >= d1)) {
            fail(name0 + " >= " + name1, ExceptionMessageUtil.valueAndName(d0, name0, d1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void geq(double d0, String name0, double d1, String name1) {
        geq(d0, name0, d1, name1, 1);
    }

    public static void geq(double d0, String name0, double d1, int numCallsBelowRequirer) {
        if (!(d0 >= d1)) {
            fail(name0 + " >= " + d1, ExceptionMessageUtil.valueAndName(d0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void geq(double d0, String name0, double d1) {
        geq(d0, name0, d1, 1);
    }

    //################################################################
    // eqFalse, neqFalse, eqTrue, neqTrue (boolean ==/!= false/true)

    //----------------------------------------------------------------
    /** require (b == false) */
    public static void eqFalse(boolean b, String name, int numCallsBelowRequirer) {
        if (!(false == b)) {
            fail(name + " == false", ExceptionMessageUtil.valueAndName(b, name), numCallsBelowRequirer + 1);
        }
    }

    public static void eqFalse(boolean b, String name) {
        eqFalse(b, name, 1);
    }

    //----------------------------------------------------------------
    /** require (b != false) */
    public static void neqFalse(boolean b, String name, int numCallsBelowRequirer) {
        if (!(false != b)) {
            fail(name + " != false", ExceptionMessageUtil.valueAndName(b, name), numCallsBelowRequirer + 1);
        }
    }

    public static void neqFalse(boolean b, String name) {
        neqFalse(b, name, 1);
    }

    //----------------------------------------------------------------
    /** require (b == true) */
    public static void eqTrue(boolean b, String name, int numCallsBelowRequirer) {
        if (!(true == b)) {
            fail(name + " == true", ExceptionMessageUtil.valueAndName(b, name), numCallsBelowRequirer + 1);
        }
    }

    public static void eqTrue(boolean b, String name) {
        eqTrue(b, name, 1);
    }

    //----------------------------------------------------------------
    /** require (b != true) */
    public static void neqTrue(boolean b, String name, int numCallsBelowRequirer) {
        if (!(true != b)) {
            fail(name + " != true", ExceptionMessageUtil.valueAndName(b, name), numCallsBelowRequirer + 1);
        }
    }

    public static void neqTrue(boolean b, String name) {
        neqTrue(b, name, 1);
    }

    //################################################################
    // eqZero (primitiveValue == 0)

    //----------------------------------------------------------------
    /** require (c == 0) */
    public static void eqZero(char c, String name, int numCallsBelowRequirer) {
        if (!(0 == c)) {
            fail(name + " == 0", ExceptionMessageUtil.valueAndName(c, name), numCallsBelowRequirer + 1);
        }
    }

    public static void eqZero(char c, String name) {
        eqZero(c, name, 1);
    }

    //----------------------------------------------------------------
    /** require (b == 0) */
    public static void eqZero(byte b, String name, int numCallsBelowRequirer) {
        if (!(0 == b)) {
            fail(name + " == 0", ExceptionMessageUtil.valueAndName(b, name), numCallsBelowRequirer + 1);
        }
    }

    public static void eqZero(byte b, String name) {
        eqZero(b, name, 1);
    }

    //----------------------------------------------------------------
    /** require (s == 0) */
    public static void eqZero(short s, String name, int numCallsBelowRequirer) {
        if (!(0 == s)) {
            fail(name + " == 0", ExceptionMessageUtil.valueAndName(s, name), numCallsBelowRequirer + 1);
        }
    }

    public static void eqZero(short s, String name) {
        eqZero(s, name, 1);
    }

    //----------------------------------------------------------------
    /** require (i == 0) */
    public static void eqZero(int i, String name, int numCallsBelowRequirer) {
        if (!(0 == i)) {
            fail(name + " == 0", ExceptionMessageUtil.valueAndName(i, name), numCallsBelowRequirer + 1);
        }
    }

    public static void eqZero(int i, String name) {
        eqZero(i, name, 1);
    }

    //----------------------------------------------------------------
    /** require (l == 0) */
    public static void eqZero(long l, String name, int numCallsBelowRequirer) {
        if (!(0 == l)) {
            fail(name + " == 0", ExceptionMessageUtil.valueAndName(l, name), numCallsBelowRequirer + 1);
        }
    }

    public static void eqZero(long l, String name) {
        eqZero(l, name, 1);
    }

    //----------------------------------------------------------------
    /** require (f == 0) */
    public static void eqZero(float f, String name, int numCallsBelowRequirer) {
        if (!(0 == f)) {
            fail(name + " == 0", ExceptionMessageUtil.valueAndName(f, name), numCallsBelowRequirer + 1);
        }
    }

    public static void eqZero(float f, String name) {
        eqZero(f, name, 1);
    }

    //----------------------------------------------------------------
    /** require (d == 0) */
    public static void eqZero(double d, String name, int numCallsBelowRequirer) {
        if (!(0 == d)) {
            fail(name + " == 0", ExceptionMessageUtil.valueAndName(d, name), numCallsBelowRequirer + 1);
        }
    }

    public static void eqZero(double d, String name) {
        eqZero(d, name, 1);
    }

    //################################################################
    // neqZero (primitiveValue != 0)

    //----------------------------------------------------------------
    /** require (c != 0) */
    public static void neqZero(char c, String name, int numCallsBelowRequirer) {
        if (!(0 != c)) {
            fail(name + " != 0", ExceptionMessageUtil.valueAndName(c, name), numCallsBelowRequirer + 1);
        }
    }

    public static void neqZero(char c, String name) {
        neqZero(c, name, 1);
    }

    //----------------------------------------------------------------
    /** require (b != 0) */
    public static void neqZero(byte b, String name, int numCallsBelowRequirer) {
        if (!(0 != b)) {
            fail(name + " != 0", ExceptionMessageUtil.valueAndName(b, name), numCallsBelowRequirer + 1);
        }
    }

    public static void neqZero(byte b, String name) {
        neqZero(b, name, 1);
    }

    //----------------------------------------------------------------
    /** require (s != 0) */
    public static void neqZero(short s, String name, int numCallsBelowRequirer) {
        if (!(0 != s)) {
            fail(name + " != 0", ExceptionMessageUtil.valueAndName(s, name), numCallsBelowRequirer + 1);
        }
    }

    public static void neqZero(short s, String name) {
        neqZero(s, name, 1);
    }

    //----------------------------------------------------------------
    /** require (i != 0) */
    public static void neqZero(int i, String name, int numCallsBelowRequirer) {
        if (!(0 != i)) {
            fail(name + " != 0", ExceptionMessageUtil.valueAndName(i, name), numCallsBelowRequirer + 1);
        }
    }

    public static void neqZero(int i, String name) {
        neqZero(i, name, 1);
    }

    //----------------------------------------------------------------
    /** require (l != 0) */
    public static void neqZero(long l, String name, int numCallsBelowRequirer) {
        if (!(0 != l)) {
            fail(name + " != 0", ExceptionMessageUtil.valueAndName(l, name), numCallsBelowRequirer + 1);
        }
    }

    public static void neqZero(long l, String name) {
        neqZero(l, name, 1);
    }

    //----------------------------------------------------------------
    /** require (f != 0) */
    public static void neqZero(float f, String name, int numCallsBelowRequirer) {
        if (!(0 != f)) {
            fail(name + " != 0", ExceptionMessageUtil.valueAndName(f, name), numCallsBelowRequirer + 1);
        }
    }

    public static void neqZero(float f, String name) {
        neqZero(f, name, 1);
    }

    //----------------------------------------------------------------
    /** require (d != 0) */
    public static void neqZero(double d, String name, int numCallsBelowRequirer) {
        if (!(0 != d)) {
            fail(name + " != 0", ExceptionMessageUtil.valueAndName(d, name), numCallsBelowRequirer + 1);
        }
    }

    public static void neqZero(double d, String name) {
        neqZero(d, name, 1);
    }

    //################################################################
    // ltZero (primitiveValue < 0)

    //----------------------------------------------------------------
    /** require (b < 0) */
    public static void ltZero(byte b, String name, int numCallsBelowRequirer) {
        if (!(b < 0)) {
            fail(name + " < 0", ExceptionMessageUtil.valueAndName(b, name), numCallsBelowRequirer + 1);
        }
    }

    public static void ltZero(byte b, String name) {
        ltZero(b, name, 1);
    }

    //----------------------------------------------------------------
    /** require (s < 0) */
    public static void ltZero(short s, String name, int numCallsBelowRequirer) {
        if (!(s < 0)) {
            fail(name + " < 0", ExceptionMessageUtil.valueAndName(s, name), numCallsBelowRequirer + 1);
        }
    }

    public static void ltZero(short s, String name) {
        ltZero(s, name, 1);
    }

    //----------------------------------------------------------------
    /** require (i < 0) */
    public static void ltZero(int i, String name, int numCallsBelowRequirer) {
        if (!(i < 0)) {
            fail(name + " < 0", ExceptionMessageUtil.valueAndName(i, name), numCallsBelowRequirer + 1);
        }
    }

    public static void ltZero(int i, String name) {
        ltZero(i, name, 1);
    }

    //----------------------------------------------------------------
    /** require (l < 0) */
    public static void ltZero(long l, String name, int numCallsBelowRequirer) {
        if (!(l < 0)) {
            fail(name + " < 0", ExceptionMessageUtil.valueAndName(l, name), numCallsBelowRequirer + 1);
        }
    }

    public static void ltZero(long l, String name) {
        ltZero(l, name, 1);
    }

    //----------------------------------------------------------------
    /** require (f < 0) */
    public static void ltZero(float f, String name, int numCallsBelowRequirer) {
        if (!(f < 0)) {
            fail(name + " < 0", ExceptionMessageUtil.valueAndName(f, name), numCallsBelowRequirer + 1);
        }
    }

    public static void ltZero(float f, String name) {
        ltZero(f, name, 1);
    }

    //----------------------------------------------------------------
    /** require (d < 0) */
    public static void ltZero(double d, String name, int numCallsBelowRequirer) {
        if (!(d < 0)) {
            fail(name + " < 0", ExceptionMessageUtil.valueAndName(d, name), numCallsBelowRequirer + 1);
        }
    }

    public static void ltZero(double d, String name) {
        ltZero(d, name, 1);
    }

    //################################################################
    // leqZero (primitiveValue <= 0)

    //----------------------------------------------------------------
    /** require (b <= 0) */
    public static void leqZero(byte b, String name, int numCallsBelowRequirer) {
        if (!(b <= 0)) {
            fail(name + " <= 0", ExceptionMessageUtil.valueAndName(b, name), numCallsBelowRequirer + 1);
        }
    }

    public static void leqZero(byte b, String name) {
        leqZero(b, name, 1);
    }

    //----------------------------------------------------------------
    /** require (s <= 0) */
    public static void leqZero(short s, String name, int numCallsBelowRequirer) {
        if (!(s <= 0)) {
            fail(name + " <= 0", ExceptionMessageUtil.valueAndName(s, name), numCallsBelowRequirer + 1);
        }
    }

    public static void leqZero(short s, String name) {
        leqZero(s, name, 1);
    }

    //----------------------------------------------------------------
    /** require (i <= 0) */
    public static void leqZero(int i, String name, int numCallsBelowRequirer) {
        if (!(i <= 0)) {
            fail(name + " <= 0", ExceptionMessageUtil.valueAndName(i, name), numCallsBelowRequirer + 1);
        }
    }

    public static void leqZero(int i, String name) {
        leqZero(i, name, 1);
    }

    //----------------------------------------------------------------
    /** require (l <= 0) */
    public static void leqZero(long l, String name, int numCallsBelowRequirer) {
        if (!(l <= 0)) {
            fail(name + " <= 0", ExceptionMessageUtil.valueAndName(l, name), numCallsBelowRequirer + 1);
        }
    }

    public static void leqZero(long l, String name) {
        leqZero(l, name, 1);
    }

    //----------------------------------------------------------------
    /** require (f <= 0) */
    public static void leqZero(float f, String name, int numCallsBelowRequirer) {
        if (!(f <= 0)) {
            fail(name + " <= 0", ExceptionMessageUtil.valueAndName(f, name), numCallsBelowRequirer + 1);
        }
    }

    public static void leqZero(float f, String name) {
        leqZero(f, name, 1);
    }

    //----------------------------------------------------------------
    /** require (d <= 0) */
    public static void leqZero(double d, String name, int numCallsBelowRequirer) {
        if (!(d <= 0)) {
            fail(name + " <= 0", ExceptionMessageUtil.valueAndName(d, name), numCallsBelowRequirer + 1);
        }
    }

    public static void leqZero(double d, String name) {
        leqZero(d, name, 1);
    }

    //################################################################
    // gtZero (primitiveValue > 0)

    //----------------------------------------------------------------
    /** require (b > 0) */
    public static void gtZero(byte b, String name, int numCallsBelowRequirer) {
        if (!(b > 0)) {
            fail(name + " > 0", ExceptionMessageUtil.valueAndName(b, name), numCallsBelowRequirer + 1);
        }
    }

    public static void gtZero(byte b, String name) {
        gtZero(b, name, 1);
    }

    //----------------------------------------------------------------
    /** require (s > 0) */
    public static void gtZero(short s, String name, int numCallsBelowRequirer) {
        if (!(s > 0)) {
            fail(name + " > 0", ExceptionMessageUtil.valueAndName(s, name), numCallsBelowRequirer + 1);
        }
    }

    public static void gtZero(short s, String name) {
        gtZero(s, name, 1);
    }

    //----------------------------------------------------------------
    /** require (i > 0) */
    public static void gtZero(int i, String name, int numCallsBelowRequirer) {
        if (!(i > 0)) {
            fail(name + " > 0", ExceptionMessageUtil.valueAndName(i, name), numCallsBelowRequirer + 1);
        }
    }

    public static void gtZero(int i, String name) {
        gtZero(i, name, 1);
    }

    //----------------------------------------------------------------
    /** require (l > 0) */
    public static void gtZero(long l, String name, int numCallsBelowRequirer) {
        if (!(l > 0)) {
            fail(name + " > 0", ExceptionMessageUtil.valueAndName(l, name), numCallsBelowRequirer + 1);
        }
    }

    public static void gtZero(long l, String name) {
        gtZero(l, name, 1);
    }

    //----------------------------------------------------------------
    /** require (f > 0) */
    public static void gtZero(float f, String name, int numCallsBelowRequirer) {
        if (!(f > 0)) {
            fail(name + " > 0", ExceptionMessageUtil.valueAndName(f, name), numCallsBelowRequirer + 1);
        }
    }

    public static void gtZero(float f, String name) {
        gtZero(f, name, 1);
    }

    //----------------------------------------------------------------
    /** require (d > 0) */
    public static void gtZero(double d, String name, int numCallsBelowRequirer) {
        if (!(d > 0)) {
            fail(name + " > 0", ExceptionMessageUtil.valueAndName(d, name), numCallsBelowRequirer + 1);
        }
    }

    public static void gtZero(double d, String name) {
        gtZero(d, name, 1);
    }

    //################################################################
    // geqZero (primitiveValue >= 0)

    //----------------------------------------------------------------
    /** require (b >= 0) */
    public static void geqZero(byte b, String name, int numCallsBelowRequirer) {
        if (!(b >= 0)) {
            fail(name + " >= 0", ExceptionMessageUtil.valueAndName(b, name), numCallsBelowRequirer + 1);
        }
    }

    public static void geqZero(byte b, String name) {
        geqZero(b, name, 1);
    }

    //----------------------------------------------------------------
    /** require (s >= 0) */
    public static void geqZero(short s, String name, int numCallsBelowRequirer) {
        if (!(s >= 0)) {
            fail(name + " >= 0", ExceptionMessageUtil.valueAndName(s, name), numCallsBelowRequirer + 1);
        }
    }

    public static void geqZero(short s, String name) {
        geqZero(s, name, 1);
    }

    //----------------------------------------------------------------
    /** require (i >= 0) */
    public static void geqZero(int i, String name, int numCallsBelowRequirer) {
        if (!(i >= 0)) {
            fail(name + " >= 0", ExceptionMessageUtil.valueAndName(i, name), numCallsBelowRequirer + 1);
        }
    }

    public static void geqZero(int i, String name) {
        geqZero(i, name, 1);
    }

    //----------------------------------------------------------------
    /** require (l >= 0) */
    public static void geqZero(long l, String name, int numCallsBelowRequirer) {
        if (!(l >= 0)) {
            fail(name + " >= 0", ExceptionMessageUtil.valueAndName(l, name), numCallsBelowRequirer + 1);
        }
    }

    public static void geqZero(long l, String name) {
        geqZero(l, name, 1);
    }

    //----------------------------------------------------------------
    /** require (f >= 0) */
    public static void geqZero(float f, String name, int numCallsBelowRequirer) {
        if (!(f >= 0)) {
            fail(name + " >= 0", ExceptionMessageUtil.valueAndName(f, name), numCallsBelowRequirer + 1);
        }
    }

    public static void geqZero(float f, String name) {
        geqZero(f, name, 1);
    }

    //----------------------------------------------------------------
    /** require (d >= 0) */
    public static void geqZero(double d, String name, int numCallsBelowRequirer) {
        if (!(d >= 0)) {
            fail(name + " >= 0", ExceptionMessageUtil.valueAndName(d, name), numCallsBelowRequirer + 1);
        }
    }

    public static void geqZero(double d, String name) {
        geqZero(d, name, 1);
    }

    //################################################################
    // eq, neq (Object ==/!= Object)

    //----------------------------------------------------------------
    /** require (o0 == o1) */
    public static void eq(Object o0, String name0, Object o1, String name1, int numCallsBelowRequirer) {
        if (!(o0 == o1)) {
            fail(name0 + " == " + name1, ExceptionMessageUtil.valueAndName(o0, name0, o1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(Object o0, String name0, Object o1, String name1) {
        eq(o0, name0, o1, name1, 1);
    }

    public static void eq(Object o0, String name0, Object o1, int numCallsBelowRequirer) {
        if (!(o0 == o1)) {
            fail(name0 + " == " + ExceptionMessageUtil.valueString(o1), ExceptionMessageUtil.valueAndName(o0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void eq(Object o0, String name0, Object o1) {
        eq(o0, name0, o1, 1);
    }

    //----------------------------------------------------------------
    /** require (o0 != o1) */
    public static void neq(Object o0, String name0, Object o1, String name1, int numCallsBelowRequirer) {
        if (!(o0 != o1)) {
            fail(name0 + " != " + name1, ExceptionMessageUtil.valueAndName(o0, name0, o1, name1), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(Object o0, String name0, Object o1, String name1) {
        neq(o0, name0, o1, name1, 1);
    }

    public static void neq(Object o0, String name0, Object o1, int numCallsBelowRequirer) {
        if (!(o0 != o1)) {
            fail(name0 + " != " + ExceptionMessageUtil.valueString(o1), ExceptionMessageUtil.valueAndName(o0, name0), numCallsBelowRequirer + 1);
        }
    }

    public static void neq(Object o0, String name0, Object o1) {
        neq(o0, name0, o1, 1);
    }


    //################################################################
    // eqNull, neqNull (Object ==/!= null)

    //----------------------------------------------------------------
    /** require (o == null) */
    public static void eqNull(Object o, String name, int numCallsBelowRequirer) {
        if (!(null == o)) {
            fail(name + " == null", ExceptionMessageUtil.valueAndName(o, name), numCallsBelowRequirer + 1);
        }
    }

    public static void eqNull(Object o, String name) {
        eqNull(o, name, 1);
    }

    //----------------------------------------------------------------
    /** require (o != null) */
    public static <T> T neqNull(T o, String name, int numCallsBelowRequirer) {
        if (!(null != o)) {
            fail(name + " != null", ExceptionMessageUtil.valueAndName(o, name), numCallsBelowRequirer + 1);
        }
        return o;
    }

    public static <T> T neqNull(T o, String name) {
        return neqNull(o, name, 1);
    }

    //################################################################
    // equals (Object.equals(Object))

    //----------------------------------------------------------------
    /** require (o0 != null && o1 != null && o0.equals(o1)) */
    public static void equals(Object o0, String name0, Object o1, String name1, int numCallsBelowRequirer) {
        neqNull(o0, name0, numCallsBelowRequirer + 1);
        neqNull(o1, name1, numCallsBelowRequirer + 1);
        if (!(o0.equals(o1))) {
            fail(name0 + ".equals(" + name1 + ")", ExceptionMessageUtil.valueAndName(o0, name0, o1, name1), numCallsBelowRequirer + 1);
        }
    }

    /** require (o0 != null && o1 != null && o0.equals(o1)) */
    public static void equals(Object o0, String name0, Object o1, String name1) {
        equals(o0, name0, o1, name1, 1);
    }

    /** require (o0 != null && o1 != null && o0.equals(o1)) */
    public static void equals(Object o0, String name0, Object o1, int numCallsBelowRequirer) {
        neqNull(o0, name0, numCallsBelowRequirer + 1);
        neqNull(o1, "o1", numCallsBelowRequirer + 1);
        if (!(o0.equals(o1))) {
            fail(name0 + ".equals(" + ExceptionMessageUtil.valueString(o1) + ")", ExceptionMessageUtil.valueAndName(o0, name0), numCallsBelowRequirer + 1);
        }
    }

    /** require (o0 != null && o1 != null && o0.equals(o1)) */
    public static void equals(Object o0, String name0, Object o1) {
        equals(o0, name0, o1, 1);
    }

    //----------------------------------------------------------------
    /** require (o0 != null && o1 != null && !o0.equals(o1)) */
    public static void notEquals(Object o0, String name0, Object o1, String name1, int numCallsBelowRequirer) {
        neqNull(o0, name0, numCallsBelowRequirer + 1);
        neqNull(o1, name1, numCallsBelowRequirer + 1);
        if (o0.equals(o1)) {
            fail("!" + name0 + ".equals(" + name1 + ")", ExceptionMessageUtil.valueAndName(o0, name0, o1, name1), numCallsBelowRequirer + 1);
        }
    }

    /** require (o0 != null && o1 != null && !o0.equals(o1)) */
    public static void notEquals(Object o0, String name0, Object o1, String name1) {
        notEquals(o0, name0, o1, name1, 1);
    }

    /** require (o0 != null && o1 != null && !o0.equals(o1)) */
    public static void notEquals(Object o0, String name0, Object o1, int numCallsBelowRequirer) {
        neqNull(o0, name0, numCallsBelowRequirer + 1);
        neqNull(o1, "o1", numCallsBelowRequirer + 1);
        if (o0.equals(o1)) {
            fail("!" + name0 + ".equals(" + ExceptionMessageUtil.valueString(o1) + ")", ExceptionMessageUtil.valueAndName(o0, name0), numCallsBelowRequirer + 1);
        }
    }

    /** require (o0 != null && o1 != null && !o0.equals(o1)) */
    public static void notEquals(Object o0, String name0, Object o1) {
        notEquals(o0, name0, o1, 1);
    }

    //################################################################
    // nonempty (String.equals(nonempty))

    //----------------------------------------------------------------
    /** require (s != null && s.length() > 0) */
    public static void nonempty(String s, String name, int numCallsBelowRequirer) {
        neqNull(s, name, numCallsBelowRequirer + 1);
        if (!(s.length() > 0)) {
            fail(name + ".length() > 0", ExceptionMessageUtil.valueAndName(s, name), numCallsBelowRequirer + 1);
        }
    }

    public static void nonempty(String s, String name) {
        nonempty(s, name, 1);
    }
}