package edu.wustl.cse.mobilab.openworkflow.workflow;

import edu.wustl.cse.mobilab.openworkflow.comm.FragmentObservable;
import edu.wustl.cse.mobilab.openworkflow.shell.PersistenceProvider;
import edu.wustl.cse.mobilab.openworkflow.util.SimpleXmlSerializer;
import edu.wustl.cse.mobilab.openworkflow.verify.Assert;
import edu.wustl.cse.mobilab.openworkflow.verify.Require;
import edu.wustl.cse.mobilab.openworkflow.comm.FragmentQueryable;
import org.apache.log4j.Logger;
import org.jdom.Element;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

//--------------------------------------------------------------------
/** (description of class)
 * @author Created by Louis Thomas on Nov 6, 2008. */
public class FragmentManagerImpl implements FragmentManager {

    private final PersistenceProvider m_persistenceProvider;
    private final Logger m_logger;
    private List<CompositeWorkflow> m_compositeWorkflows = new ArrayList<CompositeWorkflow>();
    private Set<FragmentObservable.Observer> m_observers = new HashSet<FragmentObservable.Observer>();
    private final Object m_lock = new Object();

    //----------------------------------------------------------------
    public FragmentManagerImpl(PersistenceProvider persistenceProvider, Logger logger) {
        Require.neqNull(persistenceProvider, "persistenceProvider");
        Require.neqNull(logger, "logger");
        m_persistenceProvider = persistenceProvider;
        m_logger = logger;

        // Read configuration info from disk.
        Element element;
        try {
            element = m_persistenceProvider.loadXmlByName(FRAGMENTSTORE_NAME);
        } catch (PersistenceProvider.NameNotFoundException nameNotFoundException) {
            throw Assert.exceptionNeverCaught(nameNotFoundException);
        }

        // Convert XML code into actual workflow objects
        SimpleXmlSerializer simpleXmlSerializer = new SimpleXmlSerializer();
        for (Object o : element.getContent()) {
            try {
                m_compositeWorkflows.add(simpleXmlSerializer.deserializeCompositeWorkflow((Element) o));
            } catch (XmlDeserializationException xmlDeserializationException) {
                throw Assert.exceptionNeverCaught(xmlDeserializationException);
            }
        }
    }

    //----------------------------------------------------------------
    public void dispose() {
        // nothing to do yet
    }

    //----------------------------------------------------------------
    public void unsubscribeFragmentsWithConfiguration(ConfigurationNode configurationNode, FragmentQueryable.Listener listener) {
        // nothing to do
        m_logger.info("FragmentManagerImpl: Subscription removed for " + configurationNode + ".");
    }

    //----------------------------------------------------------------
    public void subscribeFragmentsWithConfiguration(ConfigurationNode configurationNode, FragmentQueryable.Listener listener) {
        // Keep track of this listener and if a new fragment is added whose precondition or postcondition matches this
        // configurationNode, then notify this listener of that fragment.

        m_logger.info("FragmentManagerImpl: Subscription added for " + configurationNode + ".");

        // ISSUE: WHO/WHAT ADDS FRAGMENTS DYNAMICALLY TO THE LIST AS THEY BECOME AVAILABLE??  IS FRAGMENTMANAGER NOTIFIED OF THIS??

        // Look through all the tasks in all the composite workflows & return those that contain the param configurationNode
        for (CompositeWorkflow compositeWorkflow : m_compositeWorkflows) {
            taskloop:
            for (TaskNode taskNode : compositeWorkflow.getTaskNodes()) {
                for (ConfigurationNode node : taskNode.getPreconditions()) {
                    if (node.equals(configurationNode)) {
                        m_logger.debug("FragmentManagerImpl: " + configurationNode + " found in " + compositeWorkflow + ".");
                        listener.handleFragment(compositeWorkflow);
                        break taskloop;
                    }
                }
                for (ConfigurationNode node : taskNode.getPostconditions()) {
                    if (node.equals(configurationNode)) {
                        m_logger.debug("FragmentManagerImpl: " + configurationNode + " found in " + compositeWorkflow + ".");
                        listener.handleFragment(compositeWorkflow);
                        break taskloop;
                    }
                }
            }
        }
    }

    public void unsubscribe(FragmentObservable.Observer listener) {
        synchronized (m_lock) {
            m_observers.remove(listener); // Remove the listener.
            for (CompositeWorkflow compositeWorkflow : m_compositeWorkflows) {
                listener.update(compositeWorkflow, null); // Add the workflows.
            }
        }
    }

    public void subscribe(FragmentObservable.Observer listener) {

        synchronized (m_lock) {
            m_observers.add(listener); // Remove the listener.
            for (CompositeWorkflow compositeWorkflow : m_compositeWorkflows) {
                listener.update(null, compositeWorkflow); // Remove the workflows.
            }
        }
    }
}
