package edu.wustl.cse.mobilab.openworkflow.workflow;

import edu.wustl.cse.mobilab.openworkflow.schedule.Location;
import edu.wustl.cse.mobilab.openworkflow.schedule.TimePeriod;
import edu.wustl.cse.mobilab.openworkflow.verify.Require;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

//--------------------------------------------------------------------
/** Default implementation of {@link TaskImpl}.
 * @author Created by Sean Fellows on Dec 10, 2008. */
public class TaskImpl implements Task {

    // NOTE: this object is used as a key, so all fields must be immutable!

    private final TaskNode m_taskNode;
    private final TimePeriod m_timePeriod;
    private final Location m_location;
    private final Map<ExecutionDataName, ParameterName> m_inputSubscriptionNameToParameterNameMap;
    private final Map<ExecutionDataName, ParameterName> m_outputSubscriptionNameToParameterNameMap;

    //----------------------------------------------------------------
    public TaskImpl(TaskNode taskNode, TimePeriod timePeriod, Location location, Map<ExecutionDataName, ParameterName> inputSubscriptionNameToParameterNameMap, Map<ExecutionDataName, ParameterName> outputSubscriptionNameToParameterNameMap) {
        Require.neqNull(taskNode, "taskNode");
        Require.neqNull(timePeriod, "timePeriod");
        Require.neqNull(location, "location");
        Require.neqNull(inputSubscriptionNameToParameterNameMap, "inputSubscriptionNameToParameterNameMap");
        Require.neqNull(outputSubscriptionNameToParameterNameMap, "outputSubscriptionNameToParameterNameMap");
        m_taskNode=taskNode;
        m_timePeriod=timePeriod;
        m_location=location;
        // copy map for safety
        m_inputSubscriptionNameToParameterNameMap=Collections.unmodifiableMap(new LinkedHashMap<ExecutionDataName, ParameterName>(inputSubscriptionNameToParameterNameMap));
        m_outputSubscriptionNameToParameterNameMap=Collections.unmodifiableMap(new LinkedHashMap<ExecutionDataName, ParameterName>(outputSubscriptionNameToParameterNameMap));
    }

    //----------------------------------------------------------------
    // from TaskNode
    public String getContractLabel() {
        return m_taskNode.getContractLabel();
    }

    //----------------------------------------------------------------
    // from TaskNode
    public Set<ConfigurationNode> getPreconditions() {
        return m_taskNode.getPreconditions();
    }

    //----------------------------------------------------------------
    // from TaskNode
    public Set<ConfigurationNode> getPostconditions() {
        return m_taskNode.getPostconditions();
    }

    //----------------------------------------------------------------
    // from TaskNode
    public Junction getJunction() {
        return m_taskNode.getJunction();
    }

    //----------------------------------------------------------------
    // from TaskConstraint
    public Location getLocation() {
        return m_location;
    }

    //----------------------------------------------------------------
    // from TaskConstraint
    public TimePeriod getTimePeriod() {
        return m_timePeriod;
    }

    //----------------------------------------------------------------
    // from Task
    public Map<ExecutionDataName, ParameterName> getInputSubscriptionNameToParameterNameMap() {
        return m_inputSubscriptionNameToParameterNameMap;
    }

    //----------------------------------------------------------------
    // from Task
    public Map<ExecutionDataName, ParameterName> getOutputSubscriptionNameToParameterNameMap() {
        return m_outputSubscriptionNameToParameterNameMap;
    }

    //----------------------------------------------------------------
    @Override
    public boolean equals(Object that) {
        if (this==that) { return true; }
        if (that==null || getClass()!=that.getClass()) { return false; }
        TaskImpl task=(TaskImpl)that;
        if (!m_inputSubscriptionNameToParameterNameMap.equals(task.m_inputSubscriptionNameToParameterNameMap)) { return false; }
        if (!m_location.equals(task.m_location)) { return false; }
        if (!m_outputSubscriptionNameToParameterNameMap.equals(task.m_outputSubscriptionNameToParameterNameMap)) { return false; }
        if (!m_taskNode.equals(task.m_taskNode)) { return false; }
        if (!m_timePeriod.equals(task.m_timePeriod)) { return false; }
        return true;
    }

    //----------------------------------------------------------------
    @Override
    public int hashCode() {
        int result=m_taskNode.hashCode();
        result=31*result+m_timePeriod.hashCode();
        result=31*result+m_location.hashCode();
        result=31*result+m_inputSubscriptionNameToParameterNameMap.hashCode();
        result=31*result+m_outputSubscriptionNameToParameterNameMap.hashCode();
        return result;
    }

    //----------------------------------------------------------------
    @Override
    public String toString() {
        return "{Task "+m_taskNode
            +" timePeriod:"+m_timePeriod
            +" location:"+m_location
            +" inputs:"+m_inputSubscriptionNameToParameterNameMap
            +" outputs:"+m_outputSubscriptionNameToParameterNameMap
            +"}";
    }
}
