package edu.wustl.cse.mobilab.openworkflow.workflow;

import edu.wustl.cse.mobilab.openworkflow.auction.AuctionManager;
import edu.wustl.cse.mobilab.openworkflow.comm.FragmentQueryable;
import edu.wustl.cse.mobilab.openworkflow.comm.TaskFeasibilityQueryable;
import edu.wustl.cse.mobilab.openworkflow.schedule.Duration;
import edu.wustl.cse.mobilab.openworkflow.schedule.DurationImpl;
import edu.wustl.cse.mobilab.openworkflow.schedule.Location;
import edu.wustl.cse.mobilab.openworkflow.schedule.LocationImpl;
import edu.wustl.cse.mobilab.openworkflow.schedule.Time;
import edu.wustl.cse.mobilab.openworkflow.schedule.TimeImpl;
import edu.wustl.cse.mobilab.openworkflow.util.Reference;
import edu.wustl.cse.mobilab.openworkflow.util.Tuple2;
import edu.wustl.cse.mobilab.openworkflow.verify.Assert;
import edu.wustl.cse.mobilab.openworkflow.verify.Require;
import org.apache.log4j.Logger;

import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

//--------------------------------------------------------------------
/** Default implementation of {@link WorkflowManager}.
 * @author Created by Louis Thomas on Nov 6, 2008. */
public class WorkflowManagerImpl implements WorkflowManager {

    /** This flag indicates whether we will initiate a red search
     * from a disjunctive node. When this flag is false, we still meet
     * the guarantee of finding one solution. However, when false we
     * may miss solutions with fewer nodes due to a distributed race
     * condition, which can lead to confusing results. Consider a
     * red disjunctive node. If input fragments A and B are added,
     * both will turn red and be searched. However, if A is added,
     * turns green, turns the node green, and then B is added, B
     * will not be searched. Should it be? */
    public static final boolean INITIATE_RED_ON_DISJUNCTION=false;

    private static final String TERMINAL_PSEUDOTASK_ID="TERMINAL";

    private final FragmentQueryable m_fragmentQueryable;
    private final TaskFeasibilityQueryable m_taskFeasibilityQueryable;
    private final AuctionManager m_auctionManager;
    private final Logger m_logger;
    private final Set<Listener> m_listeners=new LinkedHashSet<Listener>();
    private final Set<WorkspaceImpl> m_workspaces=new LinkedHashSet<WorkspaceImpl>();

    //----------------------------------------------------------------
    public WorkflowManagerImpl(FragmentQueryable fragmentQueryable, TaskFeasibilityQueryable taskFeasibilityQueryable, AuctionManager auctionManager, Logger logger) {
        Require.neqNull(fragmentQueryable, "fragmentQueryable");
        Require.neqNull(taskFeasibilityQueryable, "taskFeasibilityQueryable");
        Require.neqNull(auctionManager, "auctionManager");
        Require.neqNull(logger, "logger");
        m_fragmentQueryable=fragmentQueryable;
        m_taskFeasibilityQueryable=taskFeasibilityQueryable;
        m_auctionManager=auctionManager;
        m_logger=logger;
    }

    //----------------------------------------------------------------
    public void dispose() {
        // nothing to do yet
    }

    //----------------------------------------------------------------
    public void addProblem(AlphaFunction alphaFunction, BetaFunction betaFunction, Set<ConfigurationNode> initialConfigurationNodes, String workflowID) {
        Require.neqNull(alphaFunction, "alphaFunction");
        Require.neqNull(betaFunction, "betaFunction");
        Require.neqNull(initialConfigurationNodes, "initialConfigurationNodes");
        Require.gtZero(initialConfigurationNodes.size(), "initialConfigurationNodes.size()");

        m_logger.info("WorkflowManagerImpl: Problem added.");

        WorkspaceImpl workspace=new WorkspaceImpl(alphaFunction, betaFunction, new LinkedHashSet<ConfigurationNode>(initialConfigurationNodes), workflowID);
        m_workspaces.add(workspace);

        // notify listeners
        for (Listener listener : m_listeners) {
            listener.handleProblemAdded(workspace);
        }

        workspace.start();
    }

    //----------------------------------------------------------------
    public void addListener(Listener listener) {
        Require.neqNull(listener, "listener");
        Require.eqFalse(m_listeners.contains(listener), "m_listeners.contains(listener)");
        m_listeners.add(listener);

        // send initial update to listener
        for (WorkspaceImpl workspace : m_workspaces) {
            listener.handleProblemAdded(workspace);
        }
    }

    //----------------------------------------------------------------
    public void removeListener(Listener listener) {
        Require.neqNull(listener, "listener");
        Require.eqTrue(m_listeners.contains(listener), "m_listeners.contains(listener)");
        m_listeners.remove(listener);
    }

    //################################################################

    private enum Phase {EXPLORING, PRUNING}

    //----------------------------------------------------------------
    private class WorkspaceImpl implements Workspace {

        private final AlphaFunction m_alphaFunction;
        private final BetaFunction m_betaFunction;
        private final Set<ConfigurationNode> m_initialConfigurationNodes;
        private final String m_sWorkflowId;

        private final Set<Node> m_inspectionNeedingNodes=new LinkedHashSet<Node>();
        private final Map<Node, NodeAnnotation> m_nodeToNodeAnnotationMap=new LinkedHashMap<Node, NodeAnnotation>();
        private Phase m_phase;
        private Set<ConfigurationNode> m_finalConfigurationNodes;
        private final Set<ConfigurationNode> m_greenConfigurationNodes=new LinkedHashSet<ConfigurationNode>();
        private final Set<ConfigurationNode> m_unmodifiableGreenConfigurationNodes=Collections.unmodifiableSet(m_greenConfigurationNodes);
        private final Set<Node> m_purpleNodes=new LinkedHashSet<Node>();
        private final Set<Tuple2<Node, Node>> m_blueEdges=new LinkedHashSet<Tuple2<Node, Node>>();

        private final Set<Listener> m_workspaceListeners=new LinkedHashSet<Listener>();

        //------------------------------------------------------------
        private FragmentQueryable.Listener m_fragmentQueryableListener=new FragmentQueryable.Listener() {
            public void handleFragment(CompositeWorkflow compositeWorkflow) {
                m_logger.debug("WorkflowManagerImpl: Handling fragment "+compositeWorkflow+".");
                for (TaskNode taskNode : compositeWorkflow.getTaskNodes()) {
                    if (!m_nodeToNodeAnnotationMap.containsKey(taskNode)) {
                        createAnnotation(taskNode);
                        m_inspectionNeedingNodes.add(taskNode);
                        for (ConfigurationNode configurationNode : taskNode.getPostconditions()) {
                            getOrCreateAnnotation(configurationNode).getParentTasks().add(taskNode);
                            m_inspectionNeedingNodes.add(configurationNode);
                        }
                        for (ConfigurationNode configurationNode : taskNode.getPreconditions()) {
                            getOrCreateAnnotation(configurationNode).getChildTasks().add(taskNode);
                            m_inspectionNeedingNodes.add(configurationNode);
                        }
                    }
                    for (Listener workspaceListener : m_workspaceListeners) {
                        workspaceListener.taskNodeAdded(taskNode);
                    }
                }
                process();
            }
        };

        //------------------------------------------------------------
        public WorkspaceImpl(AlphaFunction alphaFunction, BetaFunction betaFunction, Set<ConfigurationNode> initialConfigurationNodes, String sWorkflowId) {
            Require.neqNull(alphaFunction, "alphaFunction");
            Require.neqNull(betaFunction, "betaFunction");
            Require.neqNull(initialConfigurationNodes, "initialConfigurationNodes");
            Require.eqFalse(initialConfigurationNodes.isEmpty(), "initialConfigurationNodes.isEmpty()");
            Require.nonempty(sWorkflowId, "sWorkflowId");
            m_alphaFunction=alphaFunction;
            m_betaFunction=betaFunction;
            m_initialConfigurationNodes=initialConfigurationNodes;
            m_sWorkflowId=sWorkflowId;
            m_phase=Phase.EXPLORING;

            // create annotations for initial nodes
            for (ConfigurationNode configurationNode : m_initialConfigurationNodes) {
                markRed(getOrCreateAnnotation(configurationNode));
            }
        }

        //------------------------------------------------------------
        public void start() {
            process();
        }

        //------------------------------------------------------------
        // from WorkflowManager.Workspace
        public AlphaFunction getAlphaFunction() {
            return m_alphaFunction;
        }

        //------------------------------------------------------------
        // from WorkflowManager.Workspace
        public BetaFunction getBetaFunction() {
            return m_betaFunction;
        }

        //------------------------------------------------------------
        // from WorkflowManager.Workspace
        public Set<ConfigurationNode> getInitialConfigurationNodes() {
            return m_initialConfigurationNodes;
        }

        //------------------------------------------------------------
        // from WorkflowManager.Workspace
        public String getWorkflowId() {
            return m_sWorkflowId;
        }

        //------------------------------------------------------------
        // from WorkflowManager.Workspace
        public void addListener(Listener listener) {
            Require.neqNull(listener, "listener");
            Require.eqFalse(m_workspaceListeners.contains(listener), "m_workspaceListeners.contains(listener)");
            m_workspaceListeners.add(listener);
        }

        //------------------------------------------------------------
        // from WorkflowManager.Workspace
        public void removeListener(Listener listener) {
            Require.neqNull(listener, "listener");
            Require.eqTrue(m_workspaceListeners.contains(listener), "m_workspaceListeners.contains(listener)");
            m_workspaceListeners.remove(listener);
        }

        //------------------------------------------------------------
        // from WorkflowManager.Workspace
        public Set<TaskNode> getTaskNodes() {
            Set<TaskNode> taskNodes=new LinkedHashSet<TaskNode>();
            for (Node node : m_nodeToNodeAnnotationMap.keySet()) {
                if (node instanceof TaskNode) {
                    TaskNode taskNode=(TaskNode)node;
                    taskNodes.add(taskNode);
                }
            }
            return Collections.unmodifiableSet(taskNodes);
        }

        //------------------------------------------------------------
        // from WorkflowManager.Workspace
        public Color getNodeColor(Node node) {
            Require.neqNull(node, "node");
            NodeAnnotation nodeAnnotation=m_nodeToNodeAnnotationMap.get(node);
            Require.neqNull(nodeAnnotation, "nodeAnnotation");
            return nodeAnnotation.getColor();
        }

        //------------------------------------------------------------
        // from WorkflowManager.Workspace
        public int getNodeDistance(Node node) {
            Require.neqNull(node, "node");
            NodeAnnotation nodeAnnotation=m_nodeToNodeAnnotationMap.get(node);
            Require.neqNull(nodeAnnotation, "nodeAnnotation");
            return nodeAnnotation.getDistance();
        }

        //------------------------------------------------------------
        // from WorkflowManager.Workspace
        public Color getEdgeColor(Tuple2<Node, Node> edge) {
            Require.neqNull(edge, "edge");
            return m_blueEdges.contains(edge)?Color.BLUE : Color.UNCOLORED;
        }

        //------------------------------------------------------------
        private ConfigurationAnnotation getOrCreateAnnotation(ConfigurationNode configurationNode) {
            ConfigurationAnnotation configurationAnnotation=(ConfigurationAnnotation)m_nodeToNodeAnnotationMap.get(configurationNode);
            if (null==configurationAnnotation) {
                configurationAnnotation=new ConfigurationAnnotation(configurationNode);
                m_nodeToNodeAnnotationMap.put(configurationNode, configurationAnnotation);
            }
            return configurationAnnotation;
        }

        //------------------------------------------------------------
        private void createAnnotation(TaskNode taskNode) {
            TaskAnnotation taskAnnotation=new TaskAnnotation(taskNode);
            m_nodeToNodeAnnotationMap.put(taskNode, taskAnnotation);
        }

        //------------------------------------------------------------
        private NodeAnnotation getAnnotation(Node node) {
            NodeAnnotation nodeAnnotation=m_nodeToNodeAnnotationMap.get(node);
            Assert.neqNull(nodeAnnotation, "nodeAnnotation");
            return nodeAnnotation;
        }

        //------------------------------------------------------------
        private void process() {

            while (!m_inspectionNeedingNodes.isEmpty()) {
                Iterator<Node> nodeIterator=m_inspectionNeedingNodes.iterator();
                Node node = nodeIterator.next();
                nodeIterator.remove();

                final NodeAnnotation nodeAnnotation=getAnnotation(node);

                // if this node is uncolored and it has a red child, turn red
                if (Color.UNCOLORED==nodeAnnotation.getColor()) {
                    final Reference<Boolean> hasRedChildRef=new Reference<Boolean>(false);
                    nodeAnnotation.visitChildren(new NodeVisitor() {
                        public void visit(Node node) {
                            if (Color.RED==getAnnotation(node).getColor()) {
                                hasRedChildRef.setValue(true);
                            }
                        }
                    });
                    if (true==hasRedChildRef.getValue()) {
                        markRed(nodeAnnotation);
                    }
                }

                // check for any red nodes that should turn green
                if (Color.RED==nodeAnnotation.getColor() && node instanceof ConfigurationNode) {
                    if (m_alphaFunction.isInitialConfiguration(((ConfigurationAnnotation)nodeAnnotation).getNode())) {
                        markGreen(nodeAnnotation, 0);
                    }
                }

                // disjunctive green propagation
                if (node.getJunction()==Junction.DISJUNCT) {
                    final Reference<Integer> minGreenParentDistanceRef=new Reference<Integer>();
                    nodeAnnotation.visitParents(new NodeVisitor() {
                        public void visit(Node parentNode) {
                            NodeAnnotation parentNodeAnnotation=getAnnotation(parentNode);
                            if (parentNodeAnnotation.getColor().ordinal()<=Color.GREEN.ordinal()
                                && (null==minGreenParentDistanceRef.getValue() || minGreenParentDistanceRef.getValue()>parentNodeAnnotation.getDistance())
                            ) {
                                minGreenParentDistanceRef.setValue(parentNodeAnnotation.getDistance());
                            }
                        }
                    });
                    if (null!=minGreenParentDistanceRef.getValue()
                        && (nodeAnnotation.getColor().ordinal()>Color.GREEN.ordinal()
                            || (nodeAnnotation.getColor().ordinal()==Color.GREEN.ordinal() && nodeAnnotation.getDistance()>minGreenParentDistanceRef.getValue()+1))
                    ) {
                        markGreen(nodeAnnotation, minGreenParentDistanceRef.getValue()+1);
                    }
                }

                // conjunctive green propagation
                if (node.getJunction()==Junction.CONJUNCT) {
                    final Reference<Boolean> hasNonGreenParentRef=new Reference<Boolean>(false);
                    final Reference<Integer> maxGreenParentDistanceRef=new Reference<Integer>();
                    nodeAnnotation.visitParents(new NodeVisitor() {
                        public void visit(Node parentNode) {
                            if (hasNonGreenParentRef.getValue()) {
                                return;
                            }
                            NodeAnnotation parentNodeAnnotation=getAnnotation(parentNode);
                            if (parentNodeAnnotation.getColor().ordinal()>Color.GREEN.ordinal()) {
                                hasNonGreenParentRef.setValue(true);
                                return;
                            }
                            if (null==maxGreenParentDistanceRef.getValue() || maxGreenParentDistanceRef.getValue()<parentNodeAnnotation.getDistance()) {
                                maxGreenParentDistanceRef.setValue(parentNodeAnnotation.getDistance());
                            }
                        }
                    });
                    if (false==hasNonGreenParentRef.getValue()
                        && null!=maxGreenParentDistanceRef.getValue() // detect no parents
                        && (nodeAnnotation.getColor().ordinal()>Color.GREEN.ordinal()
                            || (Color.GREEN==nodeAnnotation.getColor() && nodeAnnotation.getDistance()>maxGreenParentDistanceRef.getValue()+1))
                        ) {
                        markGreen(nodeAnnotation, maxGreenParentDistanceRef.getValue()+1);
                    }
                }

                // disjunctive red initiation
                if (INITIATE_RED_ON_DISJUNCTION && node.getJunction()==Junction.DISJUNCT && Color.GREEN==nodeAnnotation.getColor()) {
                    nodeAnnotation.visitParents(new NodeVisitor() {
                        public void visit(Node parentNode) {
                            NodeAnnotation parentNodeAnnotation=getAnnotation(parentNode);
                            if (parentNodeAnnotation.getColor()==Color.UNCOLORED) {
                                markRed(parentNodeAnnotation);
                            }
                        }
                    });
                }

                // conjunctive red initiation
                if ((INITIATE_RED_ON_DISJUNCTION || node.getJunction()==Junction.CONJUNCT) && Color.UNCOLORED==nodeAnnotation.getColor()) {
                    final Reference<Boolean> hasGreenParentRef=new Reference<Boolean>(false);
                    nodeAnnotation.visitParents(new NodeVisitor() {
                        public void visit(Node parentNode) {
                            if (getAnnotation(parentNode).getColor()==Color.GREEN) {
                                hasGreenParentRef.setValue(true);
                            }
                        }
                    });
                    if (true==hasGreenParentRef.getValue()) {
                        markRed(nodeAnnotation);
                    }
                }

                // purple propagation
                if (Color.PURPLE==nodeAnnotation.getColor()) {
                    if (0==nodeAnnotation.getDistance()) {
                        // no parents to color

                    } else if (node.getJunction()==Junction.DISJUNCT) {
                        // disjunctive - find best parent
                        final Reference<NodeAnnotation> minParentNodeAnnotationRef=new Reference<NodeAnnotation>();
                        nodeAnnotation.visitParents(new NodeVisitor() {
                            public void visit(Node parentNode) {
                                NodeAnnotation parentNodeAnnotation=getAnnotation(parentNode);
                                if (parentNodeAnnotation.getColor().ordinal()<=Color.GREEN.ordinal()
                                    && (null==minParentNodeAnnotationRef.getValue() || minParentNodeAnnotationRef.getValue().getDistance()>parentNodeAnnotation.getDistance())
                                ) {
                                    minParentNodeAnnotationRef.setValue(parentNodeAnnotation);
                                }
                            }
                        });
                        NodeAnnotation minParentNodeAnnotation=minParentNodeAnnotationRef.getValue();
                        Assert.neqNull(minParentNodeAnnotation, "minParentNodeAnnotation");
                        markEdgeBlue(minParentNodeAnnotation, nodeAnnotation);
                        if (minParentNodeAnnotation.getColor()==Color.GREEN) {
                            markPurple(minParentNodeAnnotation);
                        }

                    } else {
                        // conjunctive - mark all parents
                        //Assert.statementNeverExecuted();  // todo (need to test)
                        Assert.eq(node.getJunction(), "node.getJunction()", Junction.CONJUNCT, "Junction.CONJUNCT");
                        nodeAnnotation.visitParents(new NodeVisitor() {
                            public void visit(Node parentNode) {
                                NodeAnnotation parentNodeAnnotation=getAnnotation(parentNode);
                                Assert.leq(parentNodeAnnotation.getColor().ordinal(), "parentNodeAnnotation.getColor().ordinal()", Color.GREEN.ordinal(), "Color.GREEN.ordinal()");
                                markEdgeBlue(parentNodeAnnotation, nodeAnnotation);
                                if (parentNodeAnnotation.getColor()==Color.GREEN) {
                                    markPurple(parentNodeAnnotation);
                                }
                            }
                        });
                    }

                    // parents have been marked, so mark this node blue
                    markBlue(nodeAnnotation);
                }

                // add any missing fragment subscriptions
                if (Phase.EXPLORING==m_phase && node instanceof ConfigurationNode) {
                    ConfigurationAnnotation configurationAnnotation=(ConfigurationAnnotation)nodeAnnotation;
                    if (Color.UNCOLORED!=configurationAnnotation.getColor() && false==configurationAnnotation.isSubscribedForFragments()) {
                        configurationAnnotation.setSubscribedForFragments(true);
                        m_logger.debug("WorkflowManagerImpl: Subscribing to "+configurationAnnotation.getNode()+".");
                        m_fragmentQueryable.subscribeFragmentsWithConfiguration(configurationAnnotation.getNode(), m_fragmentQueryableListener);
                    }
                }

            } // <- end node inspection loop
        }

        //------------------------------------------------------------
        private void markRed(NodeAnnotation nodeAnnotation) {
            Assert.eq(nodeAnnotation.getColor(), "nodeAnnotation.getColor()", Color.UNCOLORED, "Color.UNCOLORED");

            nodeAnnotation.setColor(Color.RED);

            for (Listener workspaceListener : m_workspaceListeners) {
                workspaceListener.nodeProperyChanged(nodeAnnotation.getNode());
            }

            nodeAnnotation.visitParents(new NodeVisitor() {
                public void visit(Node node) {
                    m_inspectionNeedingNodes.add(node);
                }
            });
            m_inspectionNeedingNodes.add(nodeAnnotation.getNode());
        }

        //------------------------------------------------------------
        private void markGreen(NodeAnnotation nodeAnnotation, int number) {
            // verify preconditions
            boolean bWasGreen=false;
            if (Color.GREEN==nodeAnnotation.getColor()) {
                Assert.lt(number, "number", nodeAnnotation.getDistance(), "nodeAnnotation.getNumber()");
                bWasGreen=true;
            } else {
                Assert.gt(nodeAnnotation.getColor().ordinal(), "nodeAnnotation.getColor().ordinal()", Color.GREEN.ordinal(), "Color.GREEN.ordinal()");
            }

            nodeAnnotation.setColor(Color.GREEN);
            nodeAnnotation.setDistance(number);

            // notify listeners of the state change
            Node node=nodeAnnotation.getNode();
            for (Listener workspaceListener : m_workspaceListeners) {
                workspaceListener.nodeProperyChanged(node);
            }

            // mark that all children need to be checked for possible transitions
            nodeAnnotation.visitChildren(new NodeVisitor() {
                public void visit(Node node) {
                    m_inspectionNeedingNodes.add(node);
                }
            });
            m_inspectionNeedingNodes.add(node);

            // update the list of green nodes and check against beta
            if (false==bWasGreen && node instanceof ConfigurationNode) {
                m_greenConfigurationNodes.add((ConfigurationNode)node);
                if (Phase.EXPLORING==m_phase) {
                    m_finalConfigurationNodes=m_betaFunction.identifyFinalConfigurations(m_unmodifiableGreenConfigurationNodes);
                    if (null!=m_finalConfigurationNodes) {
                        // beta accepted - turn all nodes purple
                        m_finalConfigurationNodes=new LinkedHashSet<ConfigurationNode>(m_finalConfigurationNodes); // make copy for safety
                        Assert.assertion(!m_finalConfigurationNodes.isEmpty(), "!finalConfigurationNodes.isEmpty()");
                        Assert.assertion(m_greenConfigurationNodes.containsAll(m_finalConfigurationNodes), "m_greenConfigurationNodes.containsAll(finalConfigurationNodes)");
                        m_phase=Phase.PRUNING;
                        m_logger.info("WorkflowManagerImpl: Entering pruning phase.");
                        for (Node finalConfigurationNode : m_finalConfigurationNodes) {
                            markPurple(getAnnotation(finalConfigurationNode));
                        }
                    } // <- end if beta accepted
                }
            } // <- end if green configuration node
        }

        //------------------------------------------------------------
        private void markPurple(NodeAnnotation nodeAnnotation) {
            // verify preconditions
            Assert.neqNull(nodeAnnotation, "nodeAnnotation");
            Assert.eq(nodeAnnotation.getColor(), "nodeAnnotation.getColor()", Color.GREEN, "Color.GREEN");

            nodeAnnotation.setColor(Color.PURPLE);

            // notify listeners of the state change
            Node node=nodeAnnotation.getNode();
            for (Listener workspaceListener : m_workspaceListeners) {
                workspaceListener.nodeProperyChanged(node);
            }

            m_purpleNodes.add(nodeAnnotation.getNode());

            // mark that node needs to be checked for possible transitions
            m_inspectionNeedingNodes.add(nodeAnnotation.getNode());
        }

        //------------------------------------------------------------
        private void markEdgeBlue(NodeAnnotation parentNodeAnnotation, NodeAnnotation childNodeAnnotation) {
            Assert.neqNull(parentNodeAnnotation, "parentNodeAnnotation");
            Assert.neqNull(childNodeAnnotation, "childNodeAnnotation");
            Tuple2<Node, Node> edge=new Tuple2<Node, Node>(parentNodeAnnotation.getNode(), childNodeAnnotation.getNode());
            boolean bWasInSet=!m_blueEdges.add(edge);
            Assert.eqFalse(bWasInSet, "bWasInSet");

            // notify listeners of the state change
            for (Listener workspaceListener : m_workspaceListeners) {
                workspaceListener.edgeProperyChanged(edge);
            }
        }

        //------------------------------------------------------------
        private void markBlue(NodeAnnotation nodeAnnotation) {

            // verify preconditions
            Assert.eq(nodeAnnotation.getColor(), "nodeAnnotation.getColor()", Color.PURPLE, "Color.PURPLE");

            nodeAnnotation.setColor(Color.BLUE);

            // notify listeners of the state change
            Node node=nodeAnnotation.getNode();
            for (Listener workspaceListener : m_workspaceListeners) {
                workspaceListener.nodeProperyChanged(node);
            }

            m_purpleNodes.remove(nodeAnnotation.getNode());
            if (m_purpleNodes.isEmpty()) {
                allocateSolution();
            }
        }

        //----------------------------------------------------------------
        private void allocateSolution() {
            m_logger.info("WorkflowManagerImpl: Solution found!");
            Set<ExecutionDataName> initialExecutionDataNames=new LinkedHashSet<ExecutionDataName>();
            Set<ExecutionDataName> terminalExecutionDataNames=new LinkedHashSet<ExecutionDataName>();

            List<Task> tasks=new LinkedList<Task>();
            for (NodeAnnotation nodeAnnotation : m_nodeToNodeAnnotationMap.values()) {

                // just need to clean up subscriptions on configuration nodes
                if (nodeAnnotation instanceof ConfigurationAnnotation) {
                    ConfigurationAnnotation configurationAnnotation=(ConfigurationAnnotation)nodeAnnotation;
                    if (configurationAnnotation.isSubscribedForFragments()) {
                        m_fragmentQueryable.unsubscribeFragmentsWithConfiguration(configurationAnnotation.getNode(), m_fragmentQueryableListener);
                        configurationAnnotation.setSubscribedForFragments(false);
                    }
                }

                // blue task nodes get converted to actual workflow tasks
                if (nodeAnnotation instanceof TaskAnnotation) {
                    TaskAnnotation taskAnnotation=(TaskAnnotation)nodeAnnotation;
                    if (Color.BLUE==taskAnnotation.getColor()) {

                        int nDistance=taskAnnotation.getDistance();
                        Time startTime=new TimeImpl(nDistance*10000);
                        Duration duration=new DurationImpl(3000);
                        Location location=new LocationImpl("Anywhere");

                        // Find all the input subscriptions
                        TaskNode taskNode=taskAnnotation.getNode();
                        Map<ExecutionDataName, ParameterName> inputSubscriptionNameToParameterNameMap=new LinkedHashMap<ExecutionDataName, ParameterName>();
                        for (ConfigurationNode configurationNode : taskNode.getPreconditions()) {
                            if (Color.BLUE==getEdgeColor(new Tuple2<Node, Node>(configurationNode, taskNode))) {
                                // this edge is blue
                                ExecutionDataName executionDataName=new ExecutionDataNameImpl(m_sWorkflowId+"."+configurationNode.getPredicateLabel()+"."+taskNode.getContractLabel());
                                ParameterName parameterName=new ParameterNameImpl(configurationNode.getPredicateLabel());
                                if (0==getAnnotation(configurationNode).getDistance()) {
                                    // this is an initial condition (an condition-task link)
                                    //   There is no parent task for this condition as it is an initial condition.
                                    //   (Currently we track this locally but don't set up a subscription for it.)
                                    Assert.eqFalse(initialExecutionDataNames.contains(executionDataName), "initialExecutionDataNames.contains(executionDataName)");
                                    initialExecutionDataNames.add(executionDataName);
                                } else {
                                    // this is a middle condition (a task-condition-task link)
                                    //   The parent task is implicit since a node must only have one blue parent.
                                    //   We don't have to figure out the parent task. We just need to identify the last half of the link, which we have.
                                    Assert.eqFalse(inputSubscriptionNameToParameterNameMap.containsKey(executionDataName), "inputSubscriptionNameToParameterNameMap.containsKey(executionDataName)");
                                    inputSubscriptionNameToParameterNameMap.put(executionDataName, parameterName);
                                }
                            }
                        }

                        // Find all the output subscriptions
                        final Map<ExecutionDataName, ParameterName> outputSubscriptionNameToParameterNameMap=new LinkedHashMap<ExecutionDataName, ParameterName>();
                        for (final ConfigurationNode configurationNode : taskNode.getPostconditions()) {
                            if (Color.BLUE==getEdgeColor(new Tuple2<Node, Node>(taskNode, configurationNode))) {
                                // this edge is blue - need to find all the dependents of this configuration node
                                getAnnotation(configurationNode).visitChildren(new NodeVisitor() {
                                    public void visit(Node node) {
                                        if (Color.BLUE==getEdgeColor(new Tuple2<Node, Node>(configurationNode, node))) {
                                            // this edge is blue - we found a task-condition-task link
                                            ExecutionDataName executionDataName=new ExecutionDataNameImpl(m_sWorkflowId+"."+configurationNode.getPredicateLabel()+"."+((TaskNode)node).getContractLabel());
                                            ParameterName parameterName=new ParameterNameImpl(configurationNode.getPredicateLabel());
                                            Assert.eqFalse(outputSubscriptionNameToParameterNameMap.containsKey(executionDataName), "outputSubscriptionNameToParameterNameMap.containsKey(executionDataName)");
                                            outputSubscriptionNameToParameterNameMap.put(executionDataName, parameterName);
                                        }
                                    }
                                });
                                if (m_finalConfigurationNodes.contains(configurationNode)) {
                                    // this edge is blue - we found a terminal task-condition link
                                    //   (Currently we track this locally but don't set up a subscription for it.)
                                    ExecutionDataName executionDataName=new ExecutionDataNameImpl(m_sWorkflowId+"."+configurationNode.getPredicateLabel()+"."+TERMINAL_PSEUDOTASK_ID);
                                    Assert.eqFalse(terminalExecutionDataNames.contains(executionDataName), "terminalExecutionDataNames.contains(executionDataName)");
                                    terminalExecutionDataNames.add(executionDataName);
                                }
                            }
                        }

                        // finally, build the tasl
                        TaskImpl task=new TaskImpl(taskNode, startTime.periodOfDuration(duration), location, inputSubscriptionNameToParameterNameMap, outputSubscriptionNameToParameterNameMap);
                        tasks.add(task);
                    }
                }
            } // <- end node inspection loop

            m_logger.debug("WorkflowManagerImpl: initialExecutionDataNames:"+initialExecutionDataNames);
            m_logger.debug("WorkflowManagerImpl: terminalExecutionDataNames:"+terminalExecutionDataNames);

            for (Task task : tasks) {
                m_auctionManager.allocateTask(task);
            }
        }

    }

    //----------------------------------------------------------------
    private abstract static class NodeAnnotation {

        private Workspace.Color m_color=Workspace.Color.UNCOLORED;
        private int m_distance;

        //------------------------------------------------------------
        public void setColor(Workspace.Color color) {
            m_color=color;
        }

        //------------------------------------------------------------
        public Workspace.Color getColor() {
            return m_color;
        }

        //------------------------------------------------------------
        public int getDistance() {
            Assert.neq(m_color, "m_color", Workspace.Color.UNCOLORED, "Workspace.Color.UNCOLORED");
            Assert.neq(m_color, "m_color", Workspace.Color.RED, "Workspace.Color.RED");
            return m_distance;
        }

        //------------------------------------------------------------
        public void setDistance(int distance) {
            m_distance=distance;
        }

        //------------------------------------------------------------
        public abstract void visitChildren(NodeVisitor nodeVisitor);

        //------------------------------------------------------------
        public abstract void visitParents(NodeVisitor nodeVisitor);

        //------------------------------------------------------------
        public abstract Node getNode();
    }

    //----------------------------------------------------------------
    private static interface NodeVisitor {
        void visit(Node node);
    }

    //----------------------------------------------------------------
    private static class ConfigurationAnnotation extends NodeAnnotation {

        private final ConfigurationNode m_configurationNode;
        private boolean m_bIsSubscribedForFragments;
        private final Set<TaskNode> m_parentTaskNodes=new LinkedHashSet<TaskNode>();
        private final Set<TaskNode> m_childTaskNodes=new LinkedHashSet<TaskNode>();

        //------------------------------------------------------------
        public ConfigurationAnnotation(ConfigurationNode configurationNode) {
            m_configurationNode=configurationNode;
            Require.neqNull(configurationNode, "configurationNode");
        }

        //------------------------------------------------------------
        @Override
        public ConfigurationNode getNode() {
            return m_configurationNode;
        }

        //------------------------------------------------------------
        public void setSubscribedForFragments(boolean bIsSubscribedForFragments) {
            m_bIsSubscribedForFragments=bIsSubscribedForFragments;
        }

        //------------------------------------------------------------
        public boolean isSubscribedForFragments() {
            return m_bIsSubscribedForFragments;
        }

        //------------------------------------------------------------
        public Set<TaskNode> getParentTasks() {
            return m_parentTaskNodes;
        }

        //------------------------------------------------------------
        public Set<TaskNode> getChildTasks() {
            return m_childTaskNodes;
        }

        //------------------------------------------------------------
        @Override
        public void visitChildren(NodeVisitor nodeVisitor) {
            for (TaskNode childTaskNode : m_childTaskNodes) {
                nodeVisitor.visit(childTaskNode);
            }
        }

        //------------------------------------------------------------
        @Override
        public void visitParents(NodeVisitor nodeVisitor) {
            for (TaskNode parentTaskNode : m_parentTaskNodes) {
                nodeVisitor.visit(parentTaskNode);
            }
        }
    }

    //----------------------------------------------------------------
    private static class TaskAnnotation extends NodeAnnotation {

        private final TaskNode m_taskNode;

        //------------------------------------------------------------
        private TaskAnnotation(TaskNode taskNode) {
            Require.neqNull(taskNode, "taskNode");
            m_taskNode=taskNode;
        }

        //------------------------------------------------------------
        @Override
        public TaskNode getNode() {
            return m_taskNode;
        }

        //------------------------------------------------------------
        @Override
        public void visitChildren(NodeVisitor nodeVisitor) {
            for (ConfigurationNode configurationNode : m_taskNode.getPostconditions()) {
                nodeVisitor.visit(configurationNode);
            }
        }

        //------------------------------------------------------------
        @Override
        public void visitParents(NodeVisitor nodeVisitor) {
            for (ConfigurationNode configurationNode : m_taskNode.getPreconditions()) {
                nodeVisitor.visit(configurationNode);
            }
        }
    }
}
