package edu.wustl.cse.mobilab.openworkflow.auction;

import edu.wustl.cse.mobilab.openworkflow.comm.AuctionQueryable;
import edu.wustl.cse.mobilab.openworkflow.comm.HostId;
import edu.wustl.cse.mobilab.openworkflow.comm.StringHostId;
import edu.wustl.cse.mobilab.openworkflow.schedule.CommitmentImpl;
import edu.wustl.cse.mobilab.openworkflow.schedule.Location;
import edu.wustl.cse.mobilab.openworkflow.schedule.LocationImpl;
import edu.wustl.cse.mobilab.openworkflow.schedule.ScheduleManager;
import edu.wustl.cse.mobilab.openworkflow.schedule.ScheduleManager.CommitmentListener;
import edu.wustl.cse.mobilab.openworkflow.schedule.ServiceInvocationDescriptor;
import edu.wustl.cse.mobilab.openworkflow.schedule.Time;
import edu.wustl.cse.mobilab.openworkflow.schedule.TimeImpl;
import edu.wustl.cse.mobilab.openworkflow.service.ServiceDescriptorImpl;
import edu.wustl.cse.mobilab.openworkflow.service.ServiceManager;
import edu.wustl.cse.mobilab.openworkflow.shell.SystemTimeProvider;
import edu.wustl.cse.mobilab.openworkflow.util.BaseTestCase;
import edu.wustl.cse.mobilab.openworkflow.verify.RequirementFailure;
import edu.wustl.cse.mobilab.openworkflow.workflow.Task;
import org.apache.log4j.Logger;
import org.easymock.Capture;
import org.easymock.EasyMock;
import static org.easymock.EasyMock.*;

import java.util.HashSet;
import java.util.Set;

//--------------------------------------------------------------------
/** Tests for AuctionParticipationManagerImpl.
 * @author sf7 */
public class TestAuctionParticipationManagerImpl extends BaseTestCase {

    ScheduleManager m_mockScheduleManager;
    ServiceManager m_mockServiceManager;
    SystemTimeProvider m_mockSystemTimeProvider;
    Logger m_logger;
    LoggingRecorder m_loggingRecorder;
    Set<Object> m_mocks;

    //----------------------------------------------------------------
    @Override
    protected void tearDown() {
        if (null!=m_loggingRecorder) {
            m_loggingRecorder.detatch();
        }
        super.tearDown();
    }

    //----------------------------------------------------------------
    private void makeConstructorMocks() {
        m_mocks=new HashSet<Object>();
        m_mockScheduleManager=createMock(ScheduleManager.class);
        m_mockServiceManager=createMock(ServiceManager.class);
        m_mockSystemTimeProvider=createMock(SystemTimeProvider.class);
        m_logger=Logger.getLogger(AuctionParticipationManagerImpl.class);
        m_loggingRecorder=new LoggingRecorder(m_logger);
    }

    //----------------------------------------------------------------
    private <T> T createMock(Class<T> c) {
        T mock=EasyMock.createMock(c);
        m_mocks.add(mock);
        return mock;
    }

    //----------------------------------------------------------------
    /** This test ensures that the AuctionParticipationManagerImpl
     * does not bid on tasks it has no matching service for. */
    public void testNoBidOnNoService() {
        // Set up mocks
        makeConstructorMocks();
        Task mockTask = createMock(Task.class);
        AuctionQueryable.Listener listener = createMock(AuctionQueryable.Listener.class);

        expect(m_mockServiceManager.getImplementingService(mockTask))
            .andReturn(null);
       m_mockScheduleManager.subscribeCommitments((CommitmentListener) anyObject());

        // Do test
        replay(m_mocks.toArray());
        AuctionParticipationManagerImpl auctionParticipationManager=
            new AuctionParticipationManagerImpl(m_mockScheduleManager,
                m_mockServiceManager, m_mockSystemTimeProvider, m_logger, new StringHostId("TestHost"));

        auctionParticipationManager.startSolicitingBids(mockTask, listener);

        // End test
        verify(m_mocks.toArray());
    }

    //----------------------------------------------------------------
    /** This test ensures that the AuctionParticipationManagerImpl
     * does not bid on tasks it cannot fit in its schedule. */
    public void testNoBidOnConflict() {
        // Set up mocks
        makeConstructorMocks();
        Task mockTask = createMock(Task.class);
        AuctionQueryable.Listener listener = createMock(AuctionQueryable.Listener.class);

        expect(m_mockServiceManager.getImplementingService(mockTask))
            .andReturn(new ServiceDescriptorImpl("name", "contract", "javaClass", null));
        expect(m_mockScheduleManager.addToSchedule((ServiceInvocationDescriptor) anyObject()))
            .andReturn(null);
       m_mockScheduleManager.subscribeCommitments((CommitmentListener) anyObject());

        // Do test
        replay(m_mocks.toArray());
        AuctionParticipationManagerImpl auctionParticipationManager=
            new AuctionParticipationManagerImpl(m_mockScheduleManager,
                m_mockServiceManager, m_mockSystemTimeProvider, m_logger, new StringHostId("TestHost"));

        auctionParticipationManager.startSolicitingBids(mockTask, listener);

        // End test
        verify(m_mocks.toArray());
    }

    //----------------------------------------------------------------
    /** This test ensures that the AuctionParticipationManagerImpl
     * bids when all preconditions are satisfied. This test also
     * verifies the data submitted in the bid.
     * TODO: This test does not yet verify the deadline supplied in the task bid. */
    public void testBidOnGoodTask() {
        // Set up mocks
        makeConstructorMocks();
        Task mockTask = createMock(Task.class);
        Time systemTime = new TimeImpl(10);
        Time deadline = new TimeImpl(100);
        Location currentLocation = new LocationImpl("Test Land");
        CommitmentImpl bidWindow = new CommitmentImpl(systemTime.periodToTime(deadline), currentLocation);
        ScheduleManager.ServiceCommitment mockCommitment =
            createMock(ScheduleManager.ServiceCommitment.class);
        AuctionQueryable.Listener mockListener = createMock(AuctionQueryable.Listener.class);

        expect(m_mockSystemTimeProvider.getSystemTime()).andReturn(systemTime).anyTimes();
       m_mockScheduleManager.subscribeCommitments((CommitmentListener) anyObject());

        // Ensure that the participator will bid.
        expect(m_mockServiceManager.getImplementingService(mockTask))
            .andReturn(new ServiceDescriptorImpl("name", "contract", "javaClass", null));
        expect(m_mockScheduleManager.addToSchedule((ServiceInvocationDescriptor) anyObject()))
            .andReturn(mockCommitment);
        expect(m_mockServiceManager.getCapabilityFraction(mockTask)).andReturn(0.5);
        expect(m_mockScheduleManager.getLocation(systemTime)).andReturn(currentLocation);
        expect(m_mockScheduleManager.getEndOfLoiteringWindow(systemTime, currentLocation)).andReturn(deadline);
        expect(m_mockScheduleManager.addToSchedule(eq(bidWindow))).andReturn(true);

        Capture<TaskBid> bidCapture = new Capture<TaskBid>();
        mockListener.handleBid(
                (AuctionQueryable.Bidder)anyObject(),
                (TaskBid)and(anyObject(), capture(bidCapture)));

        // Do test
        replay(m_mocks.toArray());
        HostId hostID = new StringHostId("TestHost");
        AuctionParticipationManagerImpl auctionParticipationManager=
            new AuctionParticipationManagerImpl(m_mockScheduleManager,
                m_mockServiceManager, m_mockSystemTimeProvider, m_logger, hostID);

        auctionParticipationManager.startSolicitingBids(mockTask, mockListener);
        if (!bidCapture.hasCaptured()) {
            fail("Did not capture bid.");
        }
        assertEquals(mockTask, bidCapture.getValue().getTask());
        assertEquals(0.5, bidCapture.getValue().getCapabilityFraction());
        assertEquals(hostID, bidCapture.getValue().getHostId());
        assertEquals(deadline, bidCapture.getValue().getDeadline());

        // End test
        verify(m_mocks.toArray());
    }

    //----------------------------------------------------------------
    public void testAbortBidOnBidWindowConflict() {
        // Set up mocks
        makeConstructorMocks();
        Task mockTask = createMock(Task.class);
        Time systemTime = new TimeImpl(10);
        Time deadline = new TimeImpl(100);
        Location currentLocation = new LocationImpl("Test Land");
        CommitmentImpl bidWindow = new CommitmentImpl(systemTime.periodToTime(deadline), currentLocation);
        ScheduleManager.ServiceCommitment mockCommitment =
            createMock(ScheduleManager.ServiceCommitment.class);
        AuctionQueryable.Listener mockListener = createMock(AuctionQueryable.Listener.class);

        expect(m_mockSystemTimeProvider.getSystemTime()).andReturn(systemTime).anyTimes();
       m_mockScheduleManager.subscribeCommitments((CommitmentListener) anyObject());

        // Ensure that the participator will bid.
        expect(m_mockServiceManager.getImplementingService(mockTask))
            .andReturn(new ServiceDescriptorImpl("name", "contract", "javaClass", null));
        expect(m_mockScheduleManager.addToSchedule((ServiceInvocationDescriptor) anyObject()))
            .andReturn(mockCommitment);
        expect(m_mockServiceManager.getCapabilityFraction(mockTask)).andReturn(0.5);
        expect(m_mockScheduleManager.getLocation(systemTime)).andReturn(currentLocation);
        expect(m_mockScheduleManager.getEndOfLoiteringWindow(systemTime, currentLocation)).andReturn(deadline);
        expect(m_mockScheduleManager.addToSchedule(eq(bidWindow))).andReturn(false);
        m_mockScheduleManager.removeFromSchedule(mockCommitment);

        // Do test
        replay(m_mocks.toArray());
        HostId hostID = new StringHostId("TestHost");
        AuctionParticipationManagerImpl auctionParticipationManager=
            new AuctionParticipationManagerImpl(m_mockScheduleManager,
                m_mockServiceManager, m_mockSystemTimeProvider, m_logger, hostID);

        auctionParticipationManager.startSolicitingBids(mockTask, mockListener);

        // End test
        verify(m_mocks.toArray());
    }

    //----------------------------------------------------------------
    /** This test ensures that the AuctionParticipationManagerImpl
     * removes a rejected bid's obligation from its schedule. This
     * test also checks that an error is reported if a previously
     * rejected bid is then confirmed. */
    public void testRemoveObligationOnReject() {
        // Set up mocks
        makeConstructorMocks();
        Task mockTask = createMock(Task.class);
        Time systemTime = new TimeImpl(10);
        Time deadline = new TimeImpl(100);
        Location currentLocation = new LocationImpl("Test Land");
        CommitmentImpl bidWindow = new CommitmentImpl(systemTime.periodToTime(deadline), currentLocation);
        ScheduleManager.ServiceCommitment mockCommitment =
            createMock(ScheduleManager.ServiceCommitment.class);
        AuctionQueryable.Listener mockListener = createMock(AuctionQueryable.Listener.class);

        expect(m_mockSystemTimeProvider.getSystemTime()).andReturn(systemTime).anyTimes();
       m_mockScheduleManager.subscribeCommitments((CommitmentListener) anyObject());

        // Ensure that the participator will bid.
        expect(m_mockServiceManager.getImplementingService(mockTask))
            .andReturn(new ServiceDescriptorImpl("name", "contract", "javaClass", null));
        expect(m_mockScheduleManager.addToSchedule((ServiceInvocationDescriptor) anyObject()))
            .andReturn(mockCommitment);
        expect(m_mockServiceManager.getCapabilityFraction(mockTask)).andReturn(0.5);

        Capture<TaskBid> bidCapture = new Capture<TaskBid>();
        Capture<AuctionQueryable.Bidder> bidderCapture =
            new Capture<AuctionQueryable.Bidder>();
        mockListener.handleBid(
                (AuctionQueryable.Bidder)and(anyObject(), capture(bidderCapture)),
                (TaskBid)and(anyObject(), capture(bidCapture)));

        expect(m_mockScheduleManager.getLocation(systemTime)).andReturn(currentLocation);
        expect(m_mockScheduleManager.getEndOfLoiteringWindow(systemTime, currentLocation)).andReturn(deadline);
        expect(m_mockScheduleManager.addToSchedule(eq(bidWindow))).andReturn(true);
       m_mockScheduleManager.removeFromSchedule(mockCommitment);
       m_mockScheduleManager.removeFromSchedule(eq(bidWindow));
       m_mockScheduleManager.unsubscribeCommitments((CommitmentListener) anyObject());

        // Do test
        replay(m_mocks.toArray());
        AuctionParticipationManagerImpl auctionParticipationManager=
            new AuctionParticipationManagerImpl(m_mockScheduleManager,
                m_mockServiceManager, m_mockSystemTimeProvider, m_logger, new StringHostId("TestHost"));

        auctionParticipationManager.startSolicitingBids(mockTask, mockListener);
        if (!bidCapture.hasCaptured()) {
            fail("Did not capture bid.");
        }
        if (!bidderCapture.hasCaptured()) {
            fail("Did not capture bidder.");
        }
        bidderCapture.getValue().rejectBid(bidCapture.getValue());

        try {
            bidderCapture.getValue().confirmBid(bidCapture.getValue());
            fail("Should have reported requirement failure.");
        } catch (RequirementFailure rf) {
            assertTrue(rf.isThisStackFrameCulprit(0));
        }

        // End test
        verify(m_mocks.toArray());
    }

    //----------------------------------------------------------------
    /** This test ensures that the AuctionParticipationManagerImpl
     * reports an error if a previously confirmed bid is rejected. */
    public void testErrorOnConfirmThenReject() {
        // Set up mocks
        makeConstructorMocks();
        Task mockTask = createMock(Task.class);
        Time systemTime = new TimeImpl(10);
        Time deadline = new TimeImpl(100);
        Location currentLocation = new LocationImpl("Test Land");
        CommitmentImpl bidWindow = new CommitmentImpl(systemTime.periodToTime(deadline), currentLocation);
        ScheduleManager.ServiceCommitment mockCommitment =
            createMock(ScheduleManager.ServiceCommitment.class);
        AuctionQueryable.Listener mockListener = createMock(AuctionQueryable.Listener.class);

        expect(m_mockSystemTimeProvider.getSystemTime()).andReturn(systemTime).anyTimes();
       m_mockScheduleManager.subscribeCommitments((CommitmentListener) anyObject());

        // Ensure that the participator will bid.
        expect(m_mockServiceManager.getImplementingService(mockTask))
            .andReturn(new ServiceDescriptorImpl("name", "contract", "javaClass", null));
        expect(m_mockScheduleManager.addToSchedule((ServiceInvocationDescriptor) anyObject()))
            .andReturn(mockCommitment);
        expect(m_mockServiceManager.getCapabilityFraction(mockTask)).andReturn(0.5);

        Capture<TaskBid> bidCapture = new Capture<TaskBid>();
        Capture<AuctionQueryable.Bidder> bidderCapture =
            new Capture<AuctionQueryable.Bidder>();
        mockListener.handleBid(
                (AuctionQueryable.Bidder)and(anyObject(), capture(bidderCapture)),
                (TaskBid)and(anyObject(), capture(bidCapture)));
        expect(m_mockScheduleManager.getLocation(systemTime)).andReturn(currentLocation);
        expect(m_mockScheduleManager.getEndOfLoiteringWindow(systemTime, currentLocation)).andReturn(deadline);
        expect(m_mockScheduleManager.addToSchedule(eq(bidWindow))).andReturn(true);
       m_mockScheduleManager.removeFromSchedule(eq(bidWindow));
       m_mockScheduleManager.unsubscribeCommitments((CommitmentListener) anyObject());

        // Do test
        replay(m_mocks.toArray());
        AuctionParticipationManagerImpl auctionParticipationManager=
            new AuctionParticipationManagerImpl(m_mockScheduleManager,
                m_mockServiceManager, m_mockSystemTimeProvider, m_logger, new StringHostId("TestHost"));

        auctionParticipationManager.startSolicitingBids(mockTask, mockListener);
        if (!bidCapture.hasCaptured()) {
            fail("Did not capture bid.");
        }
        if (!bidderCapture.hasCaptured()) {
            fail("Did not capture bidder.");
        }
        bidderCapture.getValue().confirmBid(bidCapture.getValue());

        try {
            bidderCapture.getValue().rejectBid(bidCapture.getValue());
            fail("Should have reported requirement failure.");
        } catch (RequirementFailure rf) {
            assertTrue(rf.isThisStackFrameCulprit(0));
        }

        // End test
        verify(m_mocks.toArray());
    }

    //----------------------------------------------------------------
    /** This test ensures that the AuctionParticipationManagerImpl
     * reports an error if an unknown task is provided to
     * stopSolicitingBids. */
    public void testErrorOnStopBeforeStart() {
        // Set up mocks
        makeConstructorMocks();
        Task mockTask = createMock(Task.class);
        AuctionQueryable.Listener mockListener = createMock(AuctionQueryable.Listener.class);

        // Do test
        replay(m_mocks.toArray());
        AuctionParticipationManagerImpl auctionParticipationManager=
            new AuctionParticipationManagerImpl(m_mockScheduleManager,
                m_mockServiceManager, m_mockSystemTimeProvider, m_logger, new StringHostId("TestHost"));
        try{
            auctionParticipationManager.stopSolicitingBids(mockTask, mockListener);
            fail("Should have reported requirement failure.");
        } catch (RequirementFailure rf) {
            assertTrue(rf.isThisStackFrameCulprit(0));
        }

        // End test
        verify(m_mocks.toArray());
    }


    //----------------------------------------------------------------
    /** This test ensures that the AuctionParticipationManagerImpl
     * rebids on tasks which are still solicited when a schedule
     * change occurs.
     *
     * In this test, we have mockTasks A and B. We first solicit A and
     * bid. We then solicit B and do not bid due to schedule
     * constraints. We then reject A and expect that B will be bid on
     * again. We do not care particularly whether B is bid on or not,
     * but we do expect that B will be considered exactly twice and A
     * only once. */
    public void testRebidOnScheduleChange() {
        // Set up mocks
        makeConstructorMocks();
        Task mockTaskA = createMock(Task.class);
        Task mockTaskB = createMock(Task.class);
        Time systemTime = new TimeImpl(10);
        Time deadlineA = new TimeImpl(100);
        Location currentLocation = new LocationImpl("Test Land");
        CommitmentImpl bidWindowA = new CommitmentImpl(systemTime.periodToTime(deadlineA), currentLocation);
        ScheduleManager.ServiceCommitment mockCommitmentA =
            createMock(ScheduleManager.ServiceCommitment.class);
        AuctionQueryable.Listener mockListenerA = createMock(AuctionQueryable.Listener.class);
        AuctionQueryable.Listener mockListenerB = createMock(AuctionQueryable.Listener.class);

        expect(m_mockSystemTimeProvider.getSystemTime()).andReturn(systemTime).anyTimes();

        Capture<CommitmentListener> scheduleListenerCapture = new Capture<CommitmentListener>();
       m_mockScheduleManager.subscribeCommitments(and((CommitmentListener) anyObject(), capture(scheduleListenerCapture)));
        expectLastCall().times(2);

        expect(m_mockServiceManager.getImplementingService(mockTaskA))
            .andReturn(new ServiceDescriptorImpl("name", "contract", "javaClass", null));
        expect(m_mockServiceManager.getImplementingService(mockTaskB))
            .andReturn(null).times(2);
        expect(m_mockScheduleManager.addToSchedule((ServiceInvocationDescriptor) anyObject()))
            .andReturn(mockCommitmentA);
        expect(m_mockServiceManager.getCapabilityFraction(mockTaskA)).andReturn(0.5);
        expect(m_mockScheduleManager.getLocation(systemTime)).andReturn(currentLocation);
        expect(m_mockScheduleManager.getEndOfLoiteringWindow(systemTime, currentLocation)).andReturn(deadlineA);
        Capture<TaskBid> bidCaptureA = new Capture<TaskBid>();
        Capture<AuctionQueryable.Bidder> bidderCaptureA =
            new Capture<AuctionQueryable.Bidder>();
        mockListenerA.handleBid(
                (AuctionQueryable.Bidder)and(anyObject(), capture(bidderCaptureA)),
                (TaskBid)and(anyObject(), capture(bidCaptureA)));
        expect(m_mockScheduleManager.addToSchedule(eq(bidWindowA))).andReturn(true);
       m_mockScheduleManager.removeFromSchedule(eq(bidWindowA));
       m_mockScheduleManager.removeFromSchedule(mockCommitmentA);
       m_mockScheduleManager.unsubscribeCommitments((CommitmentListener) anyObject());

        // Do test
        replay(m_mocks.toArray());
        HostId hostID = new StringHostId("TestHost");
        AuctionParticipationManagerImpl auctionParticipationManager=
            new AuctionParticipationManagerImpl(m_mockScheduleManager,
                m_mockServiceManager, m_mockSystemTimeProvider, m_logger, hostID);

        auctionParticipationManager.startSolicitingBids(mockTaskA, mockListenerA);
        if (!bidCaptureA.hasCaptured()) {
            fail("Did not capture bid.");
        }
        if (!bidderCaptureA.hasCaptured()) {
            fail("Did not capture bidder.");
        }
        auctionParticipationManager.startSolicitingBids(mockTaskB, mockListenerB);
        bidderCaptureA.getValue().rejectBid(bidCaptureA.getValue());

        if (!scheduleListenerCapture.hasCaptured()) {
            fail("Did not capture listener.");
        }
        scheduleListenerCapture.getValue().updateCommitment(mockCommitmentA, null);

        // End test
        verify(m_mocks.toArray());
    }
}
