package edu.wustl.cse.mobilab.openworkflow.util;

import edu.wustl.cse.mobilab.openworkflow.verify.Assert;
import edu.wustl.cse.mobilab.openworkflow.verify.Require;
import junit.framework.AssertionFailedError;
import junit.framework.ComparisonFailure;
import junit.framework.TestCase;
import org.apache.log4j.Appender;
import org.apache.log4j.Layout;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.spi.ErrorHandler;
import org.apache.log4j.spi.Filter;
import org.apache.log4j.spi.LoggingEvent;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Pattern;

//--------------------------------------------------------------------
/** Provides some base functionality for all JUnit tests.
 * @author Created by Louis Thomas on Dec 3, 2008. */
public abstract class BaseTestCase extends TestCase {

    //----------------------------------------------------------------
    /** Sets up testing environment (no-op). (Removes the ability to
     * throw non-runtime-{@link Exception}s. */
    @Override protected void setUp() {
        try {
            super.setUp();
        } catch (Exception e) {
            Assert.exceptionNeverCaught(e);
        }
    }

    //----------------------------------------------------------------
    /** Cleans up the testing environment (no-op). (Removes the
     * ability to throw non-runtime-{@link Exception}s. */
    @Override protected void tearDown() {
        try {
            super.tearDown();
        } catch (Exception e) {
            Assert.exceptionNeverCaught(e);
        }
    }


    ////----------------------------------------------------------------
    ///** Starts a thread that will execute the given runnable. Returns
    // * an exact {@link Stepper.ThreadReference} that needs to be
    // * released before the runnable will start executing. */
    //public Stepper.ThreadReference startTestingThread(final Stepper stepper, final Runnable runnable, String sRunnableName) {
    //    Thread thread=new Thread(new Runnable() {
    //        public void run() {
    //            stepper.step("ready");
    //            runnable.run();
    //        }
    //    }, "test-"+sRunnableName);
    //    Stepper.ThreadReference threadReference=stepper.getThreadReferenceExact(thread);
    //    thread.start();
    //    stepper.catchThread(threadReference, "ready");
    //    return threadReference;
    //}

    //----------------------------------------------------------------
    /** Serializes then deserializes an object. */
    public static <T> T serializeDeserialize(T t) throws IOException, ClassNotFoundException {
        ByteArrayOutputStream byteArrayOutputStream=new ByteArrayOutputStream();
        ObjectOutputStream objectOutputStream=new ObjectOutputStream(byteArrayOutputStream);
        objectOutputStream.writeObject(t);
        objectOutputStream.close();
        ByteArrayInputStream byteArrayInputStream=new ByteArrayInputStream(byteArrayOutputStream.toByteArray());
        ObjectInputStream objectInputStream=new ObjectInputStream(byteArrayInputStream);
        //noinspection unchecked
        T deserializedT=(T)objectInputStream.readObject();
        objectInputStream.close();
        return deserializedT;
    }

    //----------------------------------------------------------------
    /** Asserts that the given collection contains exactly the given
     * elements (in any order). */
    public <E> void assertCollectionContainsExactly(Collection<E> collectionToSearch, E... itemsToFind) {
        assertCollectionContainsExactly(null, collectionToSearch, itemsToFind);
    }

    //----------------------------------------------------------------
    /** Asserts that the given collection contains exactly the given
     * elements (in any order). */
    public <E> void assertCollectionContainsExactly(String sMessage, Collection<E> collectionToSearch, E... itemsToFind) {
        try {
            String sPrefix=null==sMessage?"":sMessage+" ";
            if (null==itemsToFind) {
                assertNull(sPrefix+"Expected collectionToSearch to be null.", collectionToSearch);
            } else {
                assertNotNull(sPrefix+"Expected collectionToSearch to be non-null.", collectionToSearch);
                assertEquals(sPrefix+"Expected collectionToSearch and itemsToFind to be the same size.", itemsToFind.length, collectionToSearch.size());
                for (E item : itemsToFind) {
                    assertTrue(sPrefix+"Expected collectionToSeach to contain \""+item+"\".", collectionToSearch.contains(item));
                }
            }
        } catch (AssertionFailedError e) {
            System.err.println("Expected ("+itemsToFind.length+" items):");
            for (E item : itemsToFind) {
                System.err.println("    "+item);
            }
            System.err.println("Actual ("+collectionToSearch.size()+" items):");
            for (E item : collectionToSearch) {
                System.err.println("    "+item);
            }
            throw e;
        }
    }

    //----------------------------------------------------------------
    /** Asserts that the given string contains the given substring. */
    public static void assertStringContains(String sWhole, String sFragment) {
        if (!sWhole.contains(sFragment)) {
            throw new ComparisonFailure("Could not find fragment(expected) in whole(actual).", sFragment, sWhole);
        }
    }

    //----------------------------------------------------------------
    /** Asserts that the given string contains the given substring. */
    public static void assertStringContains(String sTestDescription, String sWhole, String sFragment) {
        if (!sWhole.contains(sFragment)) {
            throw new ComparisonFailure(sTestDescription+" Could not find fragment(expected) in whole(actual).", sFragment, sWhole);
        }
    }
    //################################################################

    //----------------------------------------------------------------
    /** Temporarily intercepts logging for a given {@link Logger} and
     * saves all logged
     * messages to memory so that a unit test can verify the correct
     * messages were logged.
     * <P> Usage:
     * <LI> To start capturing:
     * <Pre>m_loggingRecorder=new LoggingRecorder(Logger.getLogger(Target.class));</Pre>
     * <LI> To stop capturing:
     * <Pre>m_loggingRecorder.detatch()</Pre>
     * <LI> To check results:
     * <Pre>m_loggingRecorder.assertNoMessagesLogged();</Pre>
     * &nbsp; &nbsp; &nbsp; or
     * <Pre>m_loggingRecorder.assertMessagesLogged(<BR>    new CheckedMessage("Hello, World!", Level.INFO)<BR>);</Pre>
     * */
    public static class LoggingRecorder {

        private static final CheckedMessage[] NO_CHECKED_MESSAGES=new CheckedMessage[0];

        private final List<LoggingEvent> m_loggingEvents=new LinkedList<LoggingEvent>();

        private final Logger m_logger;
        private final boolean m_bOriginalAdditivity;

        private final Appender m_appender=new Appender() {
            public void addFilter(Filter filter) {
                Assert.statementNeverExecuted();
            }

            public Filter getFilter() {
                Assert.statementNeverExecuted();
                return null;
            }

            public void clearFilters() {
                Assert.statementNeverExecuted();
            }

            public void close() {
                Assert.statementNeverExecuted();
            }

            public void doAppend(LoggingEvent loggingEvent) {
                loggingEvent.getRenderedMessage();
                m_loggingEvents.add(loggingEvent);
            }

            public String getName() {
                return "LoggingRecorder";
            }

            public void setErrorHandler(ErrorHandler errorHandler) {
                Assert.statementNeverExecuted();
            }

            public ErrorHandler getErrorHandler() {
                Assert.statementNeverExecuted();
                return null;
            }

            public void setLayout(Layout layout) {
                Assert.statementNeverExecuted();
            }

            public Layout getLayout() {
                Assert.statementNeverExecuted();
                return null;
            }

            public void setName(String string) {
                Assert.statementNeverExecuted();
            }

            public boolean requiresLayout() {
                Assert.statementNeverExecuted();
                return false;
            }
        };

        //------------------------------------------------------------
        public LoggingRecorder(Logger logger) {
            m_logger=logger;
            m_bOriginalAdditivity=m_logger.getAdditivity();
            m_logger.setAdditivity(false);
            m_logger.addAppender(m_appender);
        }

        //------------------------------------------------------------
        public void detatch() {
            m_logger.removeAppender(m_appender);
            m_logger.setAdditivity(m_bOriginalAdditivity);
        }

        //------------------------------------------------------------
        public List<LoggingEvent> getReportAndReset() {
            LoggingEvent[] loggingEvents=m_loggingEvents.toArray(new LoggingEvent[m_loggingEvents.size()]);
            m_loggingEvents.clear();
            return Arrays.asList(loggingEvents);
        }

        //------------------------------------------------------------
        public void assertNoMessagesLogged() {
            assertMessagesLogged(NO_CHECKED_MESSAGES);
        }

        //------------------------------------------------------------
        public void assertOneWarningLogged(String sMessageFragment) {
            Require.nonempty(sMessageFragment, "sMessageFragment");
            assertOneMessageLogged(sMessageFragment, Level.WARN);
        }

        //------------------------------------------------------------
        public void assertOneMessageLogged(String sMessageFragment, Level level) {
            Require.nonempty(sMessageFragment, "sMessageFragment");
            assertMessagesLogged(new CheckedMessage(sMessageFragment, level));
        }

        //------------------------------------------------------------
        public void assertMessagesLogged(CheckedMessage... checkedMessages) {
            List<LoggingEvent> messages=getReportAndReset();
            if (checkedMessages.length!=messages.size()) {
                dumpLogMessages(messages);
            }
            assertEquals(checkedMessages.length, messages.size());
            for (int nIndex=0; nIndex<checkedMessages.length; nIndex++) {
                CheckedMessage checkedMessage=checkedMessages[nIndex];
                LoggingEvent message=messages.get(nIndex);
                checkedMessage.checkMessage(message.getRenderedMessage());

                if (null!=checkedMessage.getDetailFragment()) {
                    assertStringContains(message.getThrowableInformation().getThrowable().toString(), checkedMessage.getDetailFragment());
                }
                assertEquals(checkedMessage.getLevel(), message.getLevel());
            }
        }

        //------------------------------------------------------------
        private static void dumpLogMessages(List<LoggingEvent> messages) {
            for (LoggingEvent message : messages) {
                System.err.println("Possibly unexpected log message: ["+message.getLevel()+"] "+message.getRenderedMessage()
                    +(null!=message.getThrowableInformation()?" ("+message.getThrowableInformation().getThrowable().toString()+")":""));
            }
        }
    }

    //################################################################

    //----------------------------------------------------------------
    public static class CheckedMessage {
        private final String m_sMessageFragment;
        private final String m_sDetailFragment;
        private final Level m_level;


        public CheckedMessage(String sMessageFragment, Level level) {
            Require.nonempty(sMessageFragment, "sMessageFragment");
            m_sMessageFragment=sMessageFragment;
            m_sDetailFragment=null;
            m_level=level;
        }

        public CheckedMessage(String sMessageFragment, String sDetailFragment, Level level) {
            Require.nonempty(sMessageFragment, "sMessageFragment");
            Require.nonempty(sDetailFragment, "sDetailFragment");
            m_sMessageFragment=sMessageFragment;
            m_sDetailFragment=sDetailFragment;
            m_level=level;
        }

        public String getDetailFragment() {
            return m_sDetailFragment;
        }

        public Level getLevel() {
            return m_level;
        }

        public void checkMessage(String sRenderedMessage) {
            assertStringContains(sRenderedMessage, m_sMessageFragment);
        }
    }

    //----------------------------------------------------------------
    public static class CheckedMessageRegexp extends CheckedMessage {

        private final Pattern m_pattern;

        public CheckedMessageRegexp(Pattern pattern, Level level) {
            super(pattern.toString(), level);
            m_pattern=pattern;
        }

        @Override
        public void checkMessage(String sRenderedMessage) {
            if (!m_pattern.matcher(sRenderedMessage).matches()) {
                throw new ComparisonFailure("Could not match pattern.", m_pattern.toString(), sRenderedMessage);
            }
        }
    }


}
