package edu.wustl.cse.mobilab.openworkflow.workflow;

import edu.wustl.cse.mobilab.openworkflow.auction.MockAuctionManager;
import edu.wustl.cse.mobilab.openworkflow.comm.MockFragmentQueryable;
import edu.wustl.cse.mobilab.openworkflow.comm.MockTaskFeasibilityQueryable;
import edu.wustl.cse.mobilab.openworkflow.util.BaseTestCase;
import edu.wustl.cse.mobilab.openworkflow.util.CollectionUtil;
import edu.wustl.cse.mobilab.openworkflow.util.RecordingMockObject;
import edu.wustl.cse.mobilab.openworkflow.util.Tuple2;
import edu.wustl.cse.mobilab.openworkflow.verify.RequirementFailure;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

//--------------------------------------------------------------------
/** Tests for {@link WorkflowManagerImpl}
 * @author Created by Louis Thomas on Jan 15, 2009. */
public class TestWorkflowManagerImpl extends BaseTestCase {

    private WorkflowManager m_workflowManager;
    private MockFragmentQueryable m_mockFragmentQueryable;
    private MockTaskFeasibilityQueryable m_mockTaskFeasibilityQueryable;
    private MockAuctionManager m_mockAuctionManager;
    private MockWorkflowManagerListener m_mockWorkflowManagerListener;
    private MockWorkspaceListener m_mockWorkspaceListener;
    private LoggingRecorder m_loggingRecorder;

    //----------------------------------------------------------------
    @Override
    protected void setUp() {
        super.setUp();
        m_loggingRecorder=new LoggingRecorder(Logger.getLogger(WorkflowManagerImpl.class));
    }

    //----------------------------------------------------------------
    @Override
    protected void tearDown() {
        m_loggingRecorder.detatch();
        super.tearDown();
    }

    //----------------------------------------------------------------
    public void testAddRemoveListener() {
        setUpWorkflowManager();

        m_mockWorkflowManagerListener=new MockWorkflowManagerListener();

        // check good add
        m_workflowManager.addListener(m_mockWorkflowManagerListener);
        checkNoOtherActivity();

        // check bad add
        try {
            m_workflowManager.addListener(m_mockWorkflowManagerListener);
            fail("Expected dual add to fail");
        } catch (RequirementFailure requirementFailure) {
            assertTrue(requirementFailure.isThisStackFrameCulprit(0));
        }
        checkNoOtherActivity();

        // check good remove
        m_workflowManager.removeListener(m_mockWorkflowManagerListener);
        checkNoOtherActivity();

        // check bad remove
        try {
            m_workflowManager.removeListener(m_mockWorkflowManagerListener);
            fail("Expected dual remove to fail");
        } catch (RequirementFailure requirementFailure) {
            assertTrue(requirementFailure.isThisStackFrameCulprit(0));
        }
        checkNoOtherActivity();
    }

    //----------------------------------------------------------------
    public void testAddListenerInitialUpdate() {
        setUpWorkflowManager();

        AlphaFunction alphaFunction=new SetAlphaFunction(CollectionUtil.setFromArray((ConfigurationNode)new ConfigurationNodeImpl("A")));
        BetaFunction betaFunction=new SetBetaFunction(CollectionUtil.setFromArray((ConfigurationNode)new ConfigurationNodeImpl("B")));
        Set<ConfigurationNode> initialConfigurationNodes=CollectionUtil.<ConfigurationNode>setFromArray(new ConfigurationNodeImpl("C"));
        m_workflowManager.addProblem(alphaFunction, betaFunction, initialConfigurationNodes, "workflowID");
        assertEquals("subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"C\"},listener)", m_mockFragmentQueryable.getActivityRecordAndReset());
        m_loggingRecorder.assertMessagesLogged(new CheckedMessage("WorkflowManagerImpl: Problem added.", Level.INFO),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"C\"}.", Level.DEBUG));
        checkNoOtherActivity();

        m_mockWorkflowManagerListener=new MockWorkflowManagerListener();

        // check added listener gets initial update
        m_workflowManager.addListener(m_mockWorkflowManagerListener);
        assertEquals("handleProblemAdded({Workspace alpha:{SetAlphaFunction [{ConfigurationNode predicateLabel:\"A\"}]} beta:{SetBetaFunction [{ConfigurationNode predicateLabel:\"B\"}]} initial:[{ConfigurationNode predicateLabel:\"C\"}]})", m_mockWorkflowManagerListener.getActivityRecordAndReset());
        checkNoOtherActivity();
    }

    //----------------------------------------------------------------
    public void testAddProblem() {
        setUpWorkflowManager();

        m_mockWorkflowManagerListener=new MockWorkflowManagerListener();
        m_workflowManager.addListener(m_mockWorkflowManagerListener);

        // check bad add
        try {
            m_workflowManager.addProblem(null, null, null, "workflowID");
            fail("Expected bad params to fail");
        } catch (RequirementFailure requirementFailure) {
            assertTrue(requirementFailure.isThisStackFrameCulprit(0));
        }
        checkNoOtherActivity();

        ConfigurationNode configurationNodeA=new ConfigurationNodeImpl("A");
        AlphaFunction alphaFunction=new SetAlphaFunction(CollectionUtil.setFromArray(configurationNodeA));

        // check bad add
        try {
            m_workflowManager.addProblem(alphaFunction, null, null, "workflowID");
            fail("Expected bad params to fail");
        } catch (RequirementFailure requirementFailure) {
            assertTrue(requirementFailure.isThisStackFrameCulprit(0));
        }
        checkNoOtherActivity();

        ConfigurationNode configurationNodeB=new ConfigurationNodeImpl("B");
        BetaFunction betaFunction=new SetBetaFunction(CollectionUtil.setFromArray(configurationNodeB));

        // check bad add
        try {
            m_workflowManager.addProblem(alphaFunction, betaFunction, null, "workflowID");
            fail("Expected bad params to fail");
        } catch (RequirementFailure requirementFailure) {
            assertTrue(requirementFailure.isThisStackFrameCulprit(0));
        }
        checkNoOtherActivity();

        Set<ConfigurationNode> initialConfigurationNodes=new LinkedHashSet<ConfigurationNode>();

        // check bad add
        try {
            m_workflowManager.addProblem(alphaFunction, betaFunction, initialConfigurationNodes, "workflowID");
            fail("Expected bad params to fail");
        } catch (RequirementFailure requirementFailure) {
            assertTrue(requirementFailure.isThisStackFrameCulprit(0));
        }
        checkNoOtherActivity();

        ConfigurationNode configurationNodeC=new ConfigurationNodeImpl("C");
        initialConfigurationNodes.add(configurationNodeC);

        // check good add
        m_workflowManager.addProblem(alphaFunction, betaFunction, initialConfigurationNodes, "workflowID");
        assertEquals("subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"C\"},listener)", m_mockFragmentQueryable.getActivityRecordAndReset());
        assertEquals("handleProblemAdded({Workspace alpha:{SetAlphaFunction [{ConfigurationNode predicateLabel:\"A\"}]} beta:{SetBetaFunction [{ConfigurationNode predicateLabel:\"B\"}]} initial:[{ConfigurationNode predicateLabel:\"C\"}]})", m_mockWorkflowManagerListener.getActivityRecordAndReset());
        m_loggingRecorder.assertMessagesLogged(new CheckedMessage("WorkflowManagerImpl: Problem added.", Level.INFO),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"C\"}.", Level.DEBUG));
        checkNoOtherActivity();
    }

    //----------------------------------------------------------------
    public void testAddProblemStartQueries() {
        setUpWorkflowManager();

        ConfigurationNode configurationNodeA=new ConfigurationNodeImpl("A");
        ConfigurationNode configurationNodeB=new ConfigurationNodeImpl("B");
        ConfigurationNode configurationNodeC=new ConfigurationNodeImpl("C");

        AlphaFunction alphaFunction=new SetAlphaFunction(CollectionUtil.setFromArray(configurationNodeA));
        BetaFunction betaFunction=new SetBetaFunction(CollectionUtil.setFromArray(configurationNodeB));
        Set<ConfigurationNode> initialConfigurationNodes=CollectionUtil.setFromArray(configurationNodeC);

        TaskNode taskNode1=new TaskNodeImpl("T1", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeA), CollectionUtil.setFromArray(configurationNodeC));
        TaskNode taskNode2=new TaskNodeImpl("T2", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeC), CollectionUtil.setFromArray(configurationNodeB));

        WorkflowManager.Workspace workspace=setUpProblem(alphaFunction, betaFunction, initialConfigurationNodes,
            "subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"C\"},listener)",
            "WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"C\"}.",
            "handleProblemAdded({Workspace alpha:{SetAlphaFunction [{ConfigurationNode predicateLabel:\"A\"}]} beta:{SetBetaFunction [{ConfigurationNode predicateLabel:\"B\"}]} initial:[{ConfigurationNode predicateLabel:\"C\"}]})",
            "[][{ConfNode C: RED}][]"
        );

        // add first fragment
        m_mockFragmentQueryable.getLastListener().handleFragment(makeCompositeWorkflow(taskNode1));
        assertEquals("subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"A\"},listener)", m_mockFragmentQueryable.getActivityRecordAndReset());
        assertEquals("taskNodeAdded({TaskNode T1})nodeProperyChanged({TaskNode T1})nodeProperyChanged({ConfNode A})nodeProperyChanged({ConfNode A})nodeProperyChanged({TaskNode T1})nodeProperyChanged({ConfNode C})", m_mockWorkspaceListener.getActivityRecordAndReset());
        m_loggingRecorder.assertMessagesLogged(
            new CheckedMessage("WorkflowManagerImpl: Handling fragment {CompositeWorkflowBuilder [{TaskNode contractLabel:\"T1\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"A\"}] postcondition:[{ConfigurationNode predicateLabel:\"C\"}]}]}.", Level.DEBUG),
            new CheckedMessage("Subscribing to {ConfigurationNode predicateLabel:\"A\"}.", Level.DEBUG)
        );
        checkNoOtherActivity();
        assertEquals("[{TaskNode T1: GREEN 1}][{ConfNode A: GREEN 0}, {ConfNode C: GREEN 2}][]", getWorkspaceStateAsString(workspace));


        // add second fragment
        m_mockFragmentQueryable.getLastListener().handleFragment(makeCompositeWorkflow(taskNode2));
        assertEquals("unsubscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"C\"},listener)"
            +"unsubscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"A\"},listener)", m_mockFragmentQueryable.getActivityRecordAndReset());
        assertEquals("taskNodeAdded({TaskNode T2})nodeProperyChanged({TaskNode T2})nodeProperyChanged({ConfNode B})"
            +"nodeProperyChanged({ConfNode B})"
            +"edgeProperyChanged({Edge {TaskNode T2} -> {ConfNode B}})nodeProperyChanged({TaskNode T2})nodeProperyChanged({ConfNode B})"
            +"edgeProperyChanged({Edge {ConfNode C} -> {TaskNode T2}})nodeProperyChanged({ConfNode C})nodeProperyChanged({TaskNode T2})"
            +"edgeProperyChanged({Edge {TaskNode T1} -> {ConfNode C}})nodeProperyChanged({TaskNode T1})nodeProperyChanged({ConfNode C})"
            +"edgeProperyChanged({Edge {ConfNode A} -> {TaskNode T1}})nodeProperyChanged({ConfNode A})nodeProperyChanged({TaskNode T1})"
            +"nodeProperyChanged({ConfNode A})", m_mockWorkspaceListener.getActivityRecordAndReset());
        m_loggingRecorder.assertMessagesLogged(
            new CheckedMessage("WorkflowManagerImpl: Handling fragment {CompositeWorkflowBuilder [{TaskNode contractLabel:\"T2\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"C\"}] postcondition:[{ConfigurationNode predicateLabel:\"B\"}]}]}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Entering pruning phase.", Level.INFO),
            new CheckedMessage("WorkflowManagerImpl: Solution found!", Level.INFO),
            new CheckedMessage("WorkflowManagerImpl: initialExecutionDataNames:[{ExecutionDataName \"workflowID.A.T1\"}]", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: terminalExecutionDataNames:[{ExecutionDataName \"workflowID.B.TERMINAL\"}]", Level.DEBUG)
        );
        assertEquals("solicitNewTask({Task {TaskNode contractLabel:\"T1\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"A\"}] postcondition:[{ConfigurationNode predicateLabel:\"C\"}]} timePeriod:{TimePeriod {Time 10000} to {Time 13000}} location:{Location \"Anywhere\"} inputs:{} outputs:{{ExecutionDataName \"workflowID.C.T2\"}={ParameterName \"C\"}}})"
            +"solicitNewTask({Task {TaskNode contractLabel:\"T2\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"C\"}] postcondition:[{ConfigurationNode predicateLabel:\"B\"}]} timePeriod:{TimePeriod {Time 30000} to {Time 33000}} location:{Location \"Anywhere\"} inputs:{{ExecutionDataName \"workflowID.C.T2\"}={ParameterName \"C\"}} outputs:{}})", m_mockAuctionManager.getActivityRecordAndReset());
        checkNoOtherActivity();
        assertEquals("[{TaskNode T1: BLUE 1}, {TaskNode T2: BLUE 3}][{ConfNode A: BLUE 0}, {ConfNode B: BLUE 4}, {ConfNode C: BLUE 2}]"
            +"[{Edge {ConfNode A} -> {TaskNode T1}}, {Edge {ConfNode C} -> {TaskNode T2}}, {Edge {TaskNode T1} -> {ConfNode C}}, {Edge {TaskNode T2} -> {ConfNode B}}]", getWorkspaceStateAsString(workspace));
    }

    //----------------------------------------------------------------
    public void testDisjunctiveGreenPropagation() {
        setUpWorkflowManager();

        ConfigurationNode configurationNodeA0=new ConfigurationNodeImpl("A0");
        ConfigurationNode configurationNodeA2=new ConfigurationNodeImpl("A2");
        ConfigurationNode configurationNodeA4=new ConfigurationNodeImpl("A4");
        ConfigurationNode configurationNodeB0=new ConfigurationNodeImpl("B0");
        ConfigurationNode configurationNodeB2=new ConfigurationNodeImpl("B2");
        ConfigurationNode configurationNodeC0=new ConfigurationNodeImpl("C0");
        ConfigurationNode configurationNodeC2=new ConfigurationNodeImpl("C2");
        ConfigurationNode configurationNodeC4=new ConfigurationNodeImpl("C4");
        ConfigurationNode configurationNodeC6=new ConfigurationNodeImpl("C6");
        ConfigurationNode configurationNodeZ=new ConfigurationNodeImpl("Z");

        TaskNode taskNodeA1=new TaskNodeImpl("A1", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeA0), CollectionUtil.setFromArray(configurationNodeA2));
        TaskNode taskNodeA3=new TaskNodeImpl("A3", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeA2), CollectionUtil.setFromArray(configurationNodeA4));
        TaskNode taskNodeB1=new TaskNodeImpl("B1", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeB0), CollectionUtil.setFromArray(configurationNodeB2));
        TaskNode taskNodeC1=new TaskNodeImpl("C1", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeC0), CollectionUtil.setFromArray(configurationNodeC2));
        TaskNode taskNodeC3=new TaskNodeImpl("C3", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeC2), CollectionUtil.setFromArray(configurationNodeC4));
        TaskNode taskNodeC5=new TaskNodeImpl("C5", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeC4), CollectionUtil.setFromArray(configurationNodeC6));
        TaskNode taskNodeY=new TaskNodeImpl("Y", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeA4, configurationNodeB2, configurationNodeC6),
            CollectionUtil.setFromArray(configurationNodeZ));

        AlphaFunction alphaFunction=new SetAlphaFunction(CollectionUtil.setFromArray(configurationNodeA0, configurationNodeB0, configurationNodeC0));
        BetaFunction betaFunction=new NoBetaFunction();
        Set<ConfigurationNode> initialConfigurationNodes=CollectionUtil.setFromArray(configurationNodeZ);

        WorkflowManager.Workspace workspace=setUpProblem(alphaFunction, betaFunction, initialConfigurationNodes,
            "subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"Z\"},listener)",
            "WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"Z\"}.",
            "handleProblemAdded({Workspace alpha:{SetAlphaFunction [{ConfigurationNode predicateLabel:\"A0\"}, {ConfigurationNode predicateLabel:\"B0\"}, {ConfigurationNode predicateLabel:\"C0\"}]} beta:{NoBetaFunction} initial:[{ConfigurationNode predicateLabel:\"Z\"}]})",
            "[][{ConfNode Z: RED}][]");

        // add first fragment
        m_mockFragmentQueryable.getLastListener().handleFragment(makeCompositeWorkflow(taskNodeY));
        assertEquals("subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"A4\"},listener)"
            +"subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"B2\"},listener)"
            +"subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"C6\"},listener)",
            m_mockFragmentQueryable.getActivityRecordAndReset());
        assertEquals("taskNodeAdded({TaskNode Y})"
            +"nodeProperyChanged({TaskNode Y})"
            +"nodeProperyChanged({ConfNode A4})"
            +"nodeProperyChanged({ConfNode B2})"
            +"nodeProperyChanged({ConfNode C6})", m_mockWorkspaceListener.getActivityRecordAndReset());
        m_loggingRecorder.assertMessagesLogged(
            new CheckedMessage("WorkflowManagerImpl: Handling fragment {CompositeWorkflowBuilder [{TaskNode contractLabel:\"Y\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"A4\"}, {ConfigurationNode predicateLabel:\"B2\"}, {ConfigurationNode predicateLabel:\"C6\"}] postcondition:[{ConfigurationNode predicateLabel:\"Z\"}]}]}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"A4\"}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"B2\"}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"C6\"}.", Level.DEBUG)
        );
        checkNoOtherActivity();
        assertEquals("[{TaskNode Y: RED}][{ConfNode A4: RED}, {ConfNode B2: RED}, {ConfNode C6: RED}, {ConfNode Z: RED}][]", getWorkspaceStateAsString(workspace));

        // test first green parent turns child green
        m_mockFragmentQueryable.getLastListener().handleFragment(makeCompositeWorkflow(taskNodeA1, taskNodeA3));
        assertEquals("subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"A2\"},listener)"
            +"subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"A0\"},listener)", m_mockFragmentQueryable.getActivityRecordAndReset());
        assertEquals("taskNodeAdded({TaskNode A1})"
            +"taskNodeAdded({TaskNode A3})"
            +"nodeProperyChanged({TaskNode A3})"
            +"nodeProperyChanged({ConfNode A2})"
            +"nodeProperyChanged({TaskNode A1})"
            +"nodeProperyChanged({ConfNode A0})"
            +"nodeProperyChanged({ConfNode A0})"
            +"nodeProperyChanged({TaskNode A1})"
            +"nodeProperyChanged({ConfNode A2})"
            +"nodeProperyChanged({TaskNode A3})"
            +"nodeProperyChanged({ConfNode A4})"
            +"nodeProperyChanged({TaskNode Y})"
            +"nodeProperyChanged({ConfNode Z})", m_mockWorkspaceListener.getActivityRecordAndReset());
        m_loggingRecorder.assertMessagesLogged(
            new CheckedMessage("WorkflowManagerImpl: Handling fragment {CompositeWorkflowBuilder [{TaskNode contractLabel:\"A1\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"A0\"}] postcondition:[{ConfigurationNode predicateLabel:\"A2\"}]}, {TaskNode contractLabel:\"A3\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"A2\"}] postcondition:[{ConfigurationNode predicateLabel:\"A4\"}]}]}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"A2\"}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"A0\"}.", Level.DEBUG)
        );
        checkNoOtherActivity();
        assertEquals("[{TaskNode A1: GREEN 1}, {TaskNode A3: GREEN 3}, {TaskNode Y: GREEN 5}]"
            +"[{ConfNode A0: GREEN 0}, {ConfNode A2: GREEN 2}, {ConfNode A4: GREEN 4}, {ConfNode B2: RED}, {ConfNode C6: RED}, {ConfNode Z: GREEN 6}][]", getWorkspaceStateAsString(workspace));

        // test new better green parent inproves child
        m_mockFragmentQueryable.getLastListener().handleFragment(makeCompositeWorkflow(taskNodeB1));
        assertEquals("subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"B0\"},listener)", m_mockFragmentQueryable.getActivityRecordAndReset());
        assertEquals("taskNodeAdded({TaskNode B1})"
            +"nodeProperyChanged({TaskNode B1})"
            +"nodeProperyChanged({ConfNode B0})"
            +"nodeProperyChanged({ConfNode B0})"
            +"nodeProperyChanged({TaskNode B1})"
            +"nodeProperyChanged({ConfNode B2})"
            +"nodeProperyChanged({TaskNode Y})"
            +"nodeProperyChanged({ConfNode Z})", m_mockWorkspaceListener.getActivityRecordAndReset());
        m_loggingRecorder.assertMessagesLogged(
            new CheckedMessage("WorkflowManagerImpl: Handling fragment {CompositeWorkflowBuilder [{TaskNode contractLabel:\"B1\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"B0\"}] postcondition:[{ConfigurationNode predicateLabel:\"B2\"}]}]}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"B0\"}.", Level.DEBUG)
        );
        checkNoOtherActivity();
        assertEquals("[{TaskNode A1: GREEN 1}, {TaskNode A3: GREEN 3}, {TaskNode B1: GREEN 1}, {TaskNode Y: GREEN 3}]"
            +"[{ConfNode A0: GREEN 0}, {ConfNode A2: GREEN 2}, {ConfNode A4: GREEN 4}, {ConfNode B0: GREEN 0}, {ConfNode B2: GREEN 2}, {ConfNode C6: RED}, {ConfNode Z: GREEN 4}][]", getWorkspaceStateAsString(workspace));

        // test new worse green parent does not affect child
        m_mockFragmentQueryable.getLastListener().handleFragment(makeCompositeWorkflow(taskNodeC1, taskNodeC3, taskNodeC5));
        assertEquals("subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"C4\"},listener)"
            +"subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"C2\"},listener)"
            +"subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"C0\"},listener)", m_mockFragmentQueryable.getActivityRecordAndReset());
        assertEquals("taskNodeAdded({TaskNode C1})"
            +"taskNodeAdded({TaskNode C3})"
            +"taskNodeAdded({TaskNode C5})"
            +"nodeProperyChanged({TaskNode C5})"
            +"nodeProperyChanged({ConfNode C4})"
            +"nodeProperyChanged({TaskNode C3})"
            +"nodeProperyChanged({ConfNode C2})"
            +"nodeProperyChanged({TaskNode C1})"
            +"nodeProperyChanged({ConfNode C0})"
            +"nodeProperyChanged({ConfNode C0})"
            +"nodeProperyChanged({TaskNode C1})"
            +"nodeProperyChanged({ConfNode C2})"
            +"nodeProperyChanged({TaskNode C3})"
            +"nodeProperyChanged({ConfNode C4})"
            +"nodeProperyChanged({TaskNode C5})"
            +"nodeProperyChanged({ConfNode C6})", m_mockWorkspaceListener.getActivityRecordAndReset());
        m_loggingRecorder.assertMessagesLogged(
            new CheckedMessage("WorkflowManagerImpl: Handling fragment {CompositeWorkflowBuilder [{TaskNode contractLabel:\"C1\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"C0\"}] postcondition:[{ConfigurationNode predicateLabel:\"C2\"}]}, {TaskNode contractLabel:\"C3\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"C2\"}] postcondition:[{ConfigurationNode predicateLabel:\"C4\"}]}, {TaskNode contractLabel:\"C5\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"C4\"}] postcondition:[{ConfigurationNode predicateLabel:\"C6\"}]}]}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"C4\"}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"C2\"}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"C0\"}.", Level.DEBUG)
        );
        checkNoOtherActivity();
        assertEquals("[{TaskNode A1: GREEN 1}, {TaskNode A3: GREEN 3}, {TaskNode B1: GREEN 1}, {TaskNode C1: GREEN 1}, {TaskNode C3: GREEN 3}, {TaskNode C5: GREEN 5}, {TaskNode Y: GREEN 3}]"
            +"[{ConfNode A0: GREEN 0}, {ConfNode A2: GREEN 2}, {ConfNode A4: GREEN 4}, {ConfNode B0: GREEN 0}, {ConfNode B2: GREEN 2}, {ConfNode C0: GREEN 0}, {ConfNode C2: GREEN 2}, {ConfNode C4: GREEN 4}, {ConfNode C6: GREEN 6}, {ConfNode Z: GREEN 4}][]", getWorkspaceStateAsString(workspace));
    }

    //----------------------------------------------------------------
    public void testConjunctiveGreenPropagation() {
        setUpWorkflowManager();

        ConfigurationNode configurationNodeA0=new ConfigurationNodeImpl("A0");
        ConfigurationNode configurationNodeA2=new ConfigurationNodeImpl("A2");
        ConfigurationNode configurationNodeA4=new ConfigurationNodeImpl("A4");
        ConfigurationNode configurationNodeA6=new ConfigurationNodeImpl("A6");
        ConfigurationNode configurationNodeB0=new ConfigurationNodeImpl("B0");
        ConfigurationNode configurationNodeB2=new ConfigurationNodeImpl("B2");
        ConfigurationNode configurationNodeC0=new ConfigurationNodeImpl("C0");
        ConfigurationNode configurationNodeC2=new ConfigurationNodeImpl("C2");
        ConfigurationNode configurationNodeC4=new ConfigurationNodeImpl("C4");
        ConfigurationNode configurationNodeC6=new ConfigurationNodeImpl("C6");
        ConfigurationNode configurationNodeC8=new ConfigurationNodeImpl("C8");
        ConfigurationNode configurationNodeD0=new ConfigurationNodeImpl("D0");
        ConfigurationNode configurationNodeD2=new ConfigurationNodeImpl("D2");
        ConfigurationNode configurationNodeE0=new ConfigurationNodeImpl("E0");
        ConfigurationNode configurationNodeE2=new ConfigurationNodeImpl("E2");
        ConfigurationNode configurationNodeZ=new ConfigurationNodeImpl("Z");

        TaskNode taskNodeA1=new TaskNodeImpl("A1", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeA0), CollectionUtil.setFromArray(configurationNodeA2));
        TaskNode taskNodeA3=new TaskNodeImpl("A3", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeA2), CollectionUtil.setFromArray(configurationNodeA4));
        TaskNode taskNodeA5=new TaskNodeImpl("A5", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeA4, configurationNodeD2), CollectionUtil.setFromArray(configurationNodeA6));
        TaskNode taskNodeB1=new TaskNodeImpl("B1", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeB0), CollectionUtil.setFromArray(configurationNodeB2));
        TaskNode taskNodeC1=new TaskNodeImpl("C1", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeC0), CollectionUtil.setFromArray(configurationNodeC2));
        TaskNode taskNodeC3=new TaskNodeImpl("C3", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeC2), CollectionUtil.setFromArray(configurationNodeC4));
        TaskNode taskNodeC5=new TaskNodeImpl("C5", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeC4, configurationNodeE2), CollectionUtil.setFromArray(configurationNodeC6));
        TaskNode taskNodeC7=new TaskNodeImpl("C7", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeC6), CollectionUtil.setFromArray(configurationNodeC8));
        TaskNode taskNodeD1=new TaskNodeImpl("D1", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeD0), CollectionUtil.setFromArray(configurationNodeD2));
        TaskNode taskNodeE1=new TaskNodeImpl("E1", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeE0), CollectionUtil.setFromArray(configurationNodeE2));
        TaskNode taskNodeY=new TaskNodeImpl("Y", Junction.CONJUNCT, CollectionUtil.setFromArray(configurationNodeA6, configurationNodeB2, configurationNodeC8),
            CollectionUtil.setFromArray(configurationNodeZ));

        AlphaFunction alphaFunction=new SetAlphaFunction(CollectionUtil.setFromArray(configurationNodeA0, configurationNodeB0, configurationNodeC0, configurationNodeD0, configurationNodeE0));
        BetaFunction betaFunction=new NoBetaFunction();
        Set<ConfigurationNode> initialConfigurationNodes=CollectionUtil.setFromArray(configurationNodeZ);

        WorkflowManager.Workspace workspace=setUpProblem(alphaFunction, betaFunction, initialConfigurationNodes,
            "subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"Z\"},listener)",
            "WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"Z\"}.",
            "handleProblemAdded({Workspace alpha:{SetAlphaFunction [{ConfigurationNode predicateLabel:\"A0\"}, {ConfigurationNode predicateLabel:\"B0\"}, {ConfigurationNode predicateLabel:\"C0\"}, {ConfigurationNode predicateLabel:\"D0\"}, {ConfigurationNode predicateLabel:\"E0\"}]} beta:{NoBetaFunction} initial:[{ConfigurationNode predicateLabel:\"Z\"}]})",
            "[][{ConfNode Z: RED}][]");

        m_mockFragmentQueryable.getLastListener().handleFragment(makeCompositeWorkflow(taskNodeY));
        assertEquals("subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"A6\"},listener)"
            +"subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"B2\"},listener)"
            +"subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"C8\"},listener)",
            m_mockFragmentQueryable.getActivityRecordAndReset());
        assertEquals("taskNodeAdded({TaskNode Y})"
            +"nodeProperyChanged({TaskNode Y})"
            +"nodeProperyChanged({ConfNode A6})"
            +"nodeProperyChanged({ConfNode B2})"
            +"nodeProperyChanged({ConfNode C8})", m_mockWorkspaceListener.getActivityRecordAndReset());
        m_loggingRecorder.assertMessagesLogged(
            new CheckedMessage("WorkflowManagerImpl: Handling fragment {CompositeWorkflowBuilder [{TaskNode contractLabel:\"Y\" junction:CONJUNCT precondition:[{ConfigurationNode predicateLabel:\"A6\"}, {ConfigurationNode predicateLabel:\"B2\"}, {ConfigurationNode predicateLabel:\"C8\"}] postcondition:[{ConfigurationNode predicateLabel:\"Z\"}]}]}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"A6\"}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"B2\"}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"C8\"}.", Level.DEBUG)
        );
        checkNoOtherActivity();
        assertEquals("[{TaskNode Y: RED}][{ConfNode A6: RED}, {ConfNode B2: RED}, {ConfNode C8: RED}, {ConfNode Z: RED}][]", getWorkspaceStateAsString(workspace));

        // test first green parent not enough to change child
        m_mockFragmentQueryable.getLastListener().handleFragment(makeCompositeWorkflow(taskNodeA1, taskNodeA3, taskNodeA5));
        assertEquals("subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"D2\"},listener)"
            +"subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"A4\"},listener)"
            +"subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"A2\"},listener)"
            +"subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"A0\"},listener)", m_mockFragmentQueryable.getActivityRecordAndReset());
        assertEquals("taskNodeAdded({TaskNode A1})"
            +"taskNodeAdded({TaskNode A3})"
            +"taskNodeAdded({TaskNode A5})"
            +"nodeProperyChanged({TaskNode A5})"
            +"nodeProperyChanged({ConfNode D2})"
            +"nodeProperyChanged({ConfNode A4})"
            +"nodeProperyChanged({TaskNode A3})"
            +"nodeProperyChanged({ConfNode A2})"
            +"nodeProperyChanged({TaskNode A1})"
            +"nodeProperyChanged({ConfNode A0})"
            +"nodeProperyChanged({ConfNode A0})"
            +"nodeProperyChanged({TaskNode A1})"
            +"nodeProperyChanged({ConfNode A2})"
            +"nodeProperyChanged({TaskNode A3})"
            +"nodeProperyChanged({ConfNode A4})"
            +"nodeProperyChanged({TaskNode A5})"
            +"nodeProperyChanged({ConfNode A6})", m_mockWorkspaceListener.getActivityRecordAndReset());
        m_loggingRecorder.assertMessagesLogged(
            new CheckedMessage("WorkflowManagerImpl: Handling fragment {CompositeWorkflowBuilder [{TaskNode contractLabel:\"A1\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"A0\"}] postcondition:[{ConfigurationNode predicateLabel:\"A2\"}]}, {TaskNode contractLabel:\"A3\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"A2\"}] postcondition:[{ConfigurationNode predicateLabel:\"A4\"}]}, {TaskNode contractLabel:\"A5\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"A4\"}, {ConfigurationNode predicateLabel:\"D2\"}] postcondition:[{ConfigurationNode predicateLabel:\"A6\"}]}]}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"D2\"}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"A4\"}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"A2\"}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"A0\"}.", Level.DEBUG)
        );
        checkNoOtherActivity();
        assertEquals("[{TaskNode A1: GREEN 1}, {TaskNode A3: GREEN 3}, {TaskNode A5: GREEN 5}, {TaskNode Y: RED}]"
            +"[{ConfNode A0: GREEN 0}, {ConfNode A2: GREEN 2}, {ConfNode A4: GREEN 4}, {ConfNode A6: GREEN 6}, {ConfNode B2: RED}, {ConfNode C8: RED}, {ConfNode D2: RED}, {ConfNode Z: RED}][]", getWorkspaceStateAsString(workspace));

        // test second green parent not enough to change child
        m_mockFragmentQueryable.getLastListener().handleFragment(makeCompositeWorkflow(taskNodeB1));
        assertEquals("subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"B0\"},listener)", m_mockFragmentQueryable.getActivityRecordAndReset());
        assertEquals("taskNodeAdded({TaskNode B1})"
            +"nodeProperyChanged({TaskNode B1})"
            +"nodeProperyChanged({ConfNode B0})"
            +"nodeProperyChanged({ConfNode B0})"
            +"nodeProperyChanged({TaskNode B1})"
            +"nodeProperyChanged({ConfNode B2})", m_mockWorkspaceListener.getActivityRecordAndReset());
        m_loggingRecorder.assertMessagesLogged(
            new CheckedMessage("WorkflowManagerImpl: Handling fragment {CompositeWorkflowBuilder [{TaskNode contractLabel:\"B1\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"B0\"}] postcondition:[{ConfigurationNode predicateLabel:\"B2\"}]}]}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"B0\"}.", Level.DEBUG)
        );
        checkNoOtherActivity();
        assertEquals("[{TaskNode A1: GREEN 1}, {TaskNode A3: GREEN 3}, {TaskNode A5: GREEN 5}, {TaskNode B1: GREEN 1}, {TaskNode Y: RED}]"
            +"[{ConfNode A0: GREEN 0}, {ConfNode A2: GREEN 2}, {ConfNode A4: GREEN 4}, {ConfNode A6: GREEN 6}, {ConfNode B0: GREEN 0}, {ConfNode B2: GREEN 2}, {ConfNode C8: RED}, {ConfNode D2: RED}, {ConfNode Z: RED}][]", getWorkspaceStateAsString(workspace));

        // test final green parent changes child; child gets max distance plus one
        m_mockFragmentQueryable.getLastListener().handleFragment(makeCompositeWorkflow(taskNodeC1, taskNodeC3, taskNodeC5, taskNodeC7));
        assertEquals("subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"C6\"},listener)"
            +"subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"C4\"},listener)"
            +"subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"E2\"},listener)"
            +"subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"C2\"},listener)"
            +"subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"C0\"},listener)", m_mockFragmentQueryable.getActivityRecordAndReset());
        assertEquals("taskNodeAdded({TaskNode C1})"
            +"taskNodeAdded({TaskNode C3})"
            +"taskNodeAdded({TaskNode C5})"
            +"taskNodeAdded({TaskNode C7})"
            +"nodeProperyChanged({TaskNode C7})"
            +"nodeProperyChanged({ConfNode C6})"
            +"nodeProperyChanged({TaskNode C5})"
            +"nodeProperyChanged({ConfNode C4})"
            +"nodeProperyChanged({ConfNode E2})"
            +"nodeProperyChanged({TaskNode C3})"
            +"nodeProperyChanged({ConfNode C2})"
            +"nodeProperyChanged({TaskNode C1})"
            +"nodeProperyChanged({ConfNode C0})"
            +"nodeProperyChanged({ConfNode C0})"
            +"nodeProperyChanged({TaskNode C1})"
            +"nodeProperyChanged({ConfNode C2})"
            +"nodeProperyChanged({TaskNode C3})"
            +"nodeProperyChanged({ConfNode C4})"
            +"nodeProperyChanged({TaskNode C5})"
            +"nodeProperyChanged({ConfNode C6})"
            +"nodeProperyChanged({TaskNode C7})"
            +"nodeProperyChanged({ConfNode C8})"
            +"nodeProperyChanged({TaskNode Y})"
            +"nodeProperyChanged({ConfNode Z})", m_mockWorkspaceListener.getActivityRecordAndReset());
        m_loggingRecorder.assertMessagesLogged(
            new CheckedMessage("WorkflowManagerImpl: Handling fragment {CompositeWorkflowBuilder [{TaskNode contractLabel:\"C1\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"C0\"}] postcondition:[{ConfigurationNode predicateLabel:\"C2\"}]}, {TaskNode contractLabel:\"C3\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"C2\"}] postcondition:[{ConfigurationNode predicateLabel:\"C4\"}]}, {TaskNode contractLabel:\"C5\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"C4\"}, {ConfigurationNode predicateLabel:\"E2\"}] postcondition:[{ConfigurationNode predicateLabel:\"C6\"}]}, {TaskNode contractLabel:\"C7\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"C6\"}] postcondition:[{ConfigurationNode predicateLabel:\"C8\"}]}]}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"C6\"}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"C4\"}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"E2\"}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"C2\"}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"C0\"}.", Level.DEBUG)
        );
        checkNoOtherActivity();
        assertEquals("[{TaskNode A1: GREEN 1}, {TaskNode A3: GREEN 3}, {TaskNode A5: GREEN 5}, {TaskNode B1: GREEN 1}, {TaskNode C1: GREEN 1}, {TaskNode C3: GREEN 3}, {TaskNode C5: GREEN 5}, {TaskNode C7: GREEN 7}, {TaskNode Y: GREEN 9}]"
            +"[{ConfNode A0: GREEN 0}, {ConfNode A2: GREEN 2}, {ConfNode A4: GREEN 4}, {ConfNode A6: GREEN 6}, {ConfNode B0: GREEN 0}, {ConfNode B2: GREEN 2}, {ConfNode C0: GREEN 0}, {ConfNode C2: GREEN 2}, {ConfNode C4: GREEN 4}, {ConfNode C6: GREEN 6}, {ConfNode C8: GREEN 8}, {ConfNode D2: RED}, {ConfNode E2: RED}, {ConfNode Z: GREEN 10}][]", getWorkspaceStateAsString(workspace));

        // test improvement to non-worst parent doesn't change child
        m_mockFragmentQueryable.getLastListener().handleFragment(makeCompositeWorkflow(taskNodeD1));
        assertEquals("subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"D0\"},listener)", m_mockFragmentQueryable.getActivityRecordAndReset());
        assertEquals("taskNodeAdded({TaskNode D1})"
            +"nodeProperyChanged({TaskNode D1})"
            +"nodeProperyChanged({ConfNode D0})"
            +"nodeProperyChanged({ConfNode D0})"
            +"nodeProperyChanged({TaskNode D1})"
            +"nodeProperyChanged({ConfNode D2})"
            +"nodeProperyChanged({TaskNode A5})"
            +"nodeProperyChanged({ConfNode A6})", m_mockWorkspaceListener.getActivityRecordAndReset());
        m_loggingRecorder.assertMessagesLogged(
            new CheckedMessage("WorkflowManagerImpl: Handling fragment {CompositeWorkflowBuilder [{TaskNode contractLabel:\"D1\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"D0\"}] postcondition:[{ConfigurationNode predicateLabel:\"D2\"}]}]}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"D0\"}.", Level.DEBUG)
        );
        checkNoOtherActivity();
        assertEquals("[{TaskNode A1: GREEN 1}, {TaskNode A3: GREEN 3}, {TaskNode A5: GREEN 3}, {TaskNode B1: GREEN 1}, {TaskNode C1: GREEN 1}, {TaskNode C3: GREEN 3}, {TaskNode C5: GREEN 5}, {TaskNode C7: GREEN 7}, {TaskNode D1: GREEN 1}, {TaskNode Y: GREEN 9}]"
            +"[{ConfNode A0: GREEN 0}, {ConfNode A2: GREEN 2}, {ConfNode A4: GREEN 4}, {ConfNode A6: GREEN 4}, {ConfNode B0: GREEN 0}, {ConfNode B2: GREEN 2}, {ConfNode C0: GREEN 0}, {ConfNode C2: GREEN 2}, {ConfNode C4: GREEN 4}, {ConfNode C6: GREEN 6}, {ConfNode C8: GREEN 8}, {ConfNode D0: GREEN 0}, {ConfNode D2: GREEN 2}, {ConfNode E2: RED}, {ConfNode Z: GREEN 10}][]", getWorkspaceStateAsString(workspace));

        // test improvement to worst parent changes child
        m_mockFragmentQueryable.getLastListener().handleFragment(makeCompositeWorkflow(taskNodeE1));
        assertEquals("subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"E0\"},listener)", m_mockFragmentQueryable.getActivityRecordAndReset());
        assertEquals("taskNodeAdded({TaskNode E1})"
            +"nodeProperyChanged({TaskNode E1})"
            +"nodeProperyChanged({ConfNode E0})"
            +"nodeProperyChanged({ConfNode E0})"
            +"nodeProperyChanged({TaskNode E1})"
            +"nodeProperyChanged({ConfNode E2})"
            +"nodeProperyChanged({TaskNode C5})"
            +"nodeProperyChanged({ConfNode C6})"
            +"nodeProperyChanged({TaskNode C7})"
            +"nodeProperyChanged({ConfNode C8})"
            +"nodeProperyChanged({TaskNode Y})"
            +"nodeProperyChanged({ConfNode Z})", m_mockWorkspaceListener.getActivityRecordAndReset());
        m_loggingRecorder.assertMessagesLogged(
            new CheckedMessage("WorkflowManagerImpl: Handling fragment {CompositeWorkflowBuilder [{TaskNode contractLabel:\"E1\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"E0\"}] postcondition:[{ConfigurationNode predicateLabel:\"E2\"}]}]}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"E0\"}.", Level.DEBUG)
        );
        checkNoOtherActivity();
        assertEquals("[{TaskNode A1: GREEN 1}, {TaskNode A3: GREEN 3}, {TaskNode A5: GREEN 3}, {TaskNode B1: GREEN 1}, {TaskNode C1: GREEN 1}, {TaskNode C3: GREEN 3}, {TaskNode C5: GREEN 3}, {TaskNode C7: GREEN 5}, {TaskNode D1: GREEN 1}, {TaskNode E1: GREEN 1}, {TaskNode Y: GREEN 7}]"
            +"[{ConfNode A0: GREEN 0}, {ConfNode A2: GREEN 2}, {ConfNode A4: GREEN 4}, {ConfNode A6: GREEN 4}, {ConfNode B0: GREEN 0}, {ConfNode B2: GREEN 2}, {ConfNode C0: GREEN 0}, {ConfNode C2: GREEN 2}, {ConfNode C4: GREEN 4}, {ConfNode C6: GREEN 4}, {ConfNode C8: GREEN 6}, {ConfNode D0: GREEN 0}, {ConfNode D2: GREEN 2}, {ConfNode E0: GREEN 0}, {ConfNode E2: GREEN 2}, {ConfNode Z: GREEN 8}][]", getWorkspaceStateAsString(workspace));
    }

    //----------------------------------------------------------------
    public void testDisjunctiveRedInitiation() {
        setUpWorkflowManager();

        ConfigurationNode configurationNodeA0=new ConfigurationNodeImpl("A0");
        ConfigurationNode configurationNodeB0=new ConfigurationNodeImpl("B0");
        ConfigurationNode configurationNodeZ=new ConfigurationNodeImpl("Z");

        TaskNode taskNodeA1=new TaskNodeImpl("A1", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeA0), CollectionUtil.setFromArray(configurationNodeZ));
        TaskNode taskNodeB1=new TaskNodeImpl("B1", Junction.DISJUNCT, CollectionUtil.setFromArray(configurationNodeB0), CollectionUtil.setFromArray(configurationNodeZ));

        AlphaFunction alphaFunction=new SetAlphaFunction(CollectionUtil.setFromArray(configurationNodeA0));
        BetaFunction betaFunction=new NoBetaFunction();
        Set<ConfigurationNode> initialConfigurationNodes=CollectionUtil.setFromArray(configurationNodeZ);

        WorkflowManager.Workspace workspace=setUpProblem(alphaFunction, betaFunction, initialConfigurationNodes,
            "subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"Z\"},listener)",
            "WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"Z\"}.",
            "handleProblemAdded({Workspace alpha:{SetAlphaFunction [{ConfigurationNode predicateLabel:\"A0\"}]} beta:{NoBetaFunction} initial:[{ConfigurationNode predicateLabel:\"Z\"}]})",
            "[][{ConfNode Z: RED}][]");

        m_mockFragmentQueryable.getLastListener().handleFragment(makeCompositeWorkflow(taskNodeA1));
        assertEquals("subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"A0\"},listener)", m_mockFragmentQueryable.getActivityRecordAndReset());
        assertEquals("taskNodeAdded({TaskNode A1})"
            +"nodeProperyChanged({TaskNode A1})"
            +"nodeProperyChanged({ConfNode A0})"
            +"nodeProperyChanged({ConfNode A0})"
            +"nodeProperyChanged({TaskNode A1})"
            +"nodeProperyChanged({ConfNode Z})", m_mockWorkspaceListener.getActivityRecordAndReset());
        m_loggingRecorder.assertMessagesLogged(
            new CheckedMessage("WorkflowManagerImpl: Handling fragment {CompositeWorkflowBuilder [{TaskNode contractLabel:\"A1\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"A0\"}] postcondition:[{ConfigurationNode predicateLabel:\"Z\"}]}]}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"A0\"}.", Level.DEBUG)
        );
        checkNoOtherActivity();
        assertEquals("[{TaskNode A1: GREEN 1}][{ConfNode A0: GREEN 0}, {ConfNode Z: GREEN 2}][]", getWorkspaceStateAsString(workspace));

        if (WorkflowManagerImpl.INITIATE_RED_ON_DISJUNCTION) {
            // test fragment turns red
            m_mockFragmentQueryable.getLastListener().handleFragment(makeCompositeWorkflow(taskNodeB1));
            assertEquals("subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"B0\"},listener)", m_mockFragmentQueryable.getActivityRecordAndReset());
            assertEquals("taskNodeAdded({TaskNode B1})"
                +"nodeProperyChanged({TaskNode B1})"
                +"nodeProperyChanged({ConfNode B0})", m_mockWorkspaceListener.getActivityRecordAndReset());
            checkNoOtherActivity();
            assertEquals("[{TaskNode A1: GREEN 1}, {TaskNode B1: RED}]"
                +"[{ConfNode A0: GREEN 0}, {ConfNode B0: RED}, {ConfNode Z: GREEN 2}][]", getWorkspaceStateAsString(workspace));
        } else {
            // test nothing happens
            m_mockFragmentQueryable.getLastListener().handleFragment(makeCompositeWorkflow(taskNodeB1));
            assertEquals("taskNodeAdded({TaskNode B1})", m_mockWorkspaceListener.getActivityRecordAndReset());
            m_loggingRecorder.assertMessagesLogged(
                new CheckedMessage("WorkflowManagerImpl: Handling fragment {CompositeWorkflowBuilder [{TaskNode contractLabel:\"B1\" junction:DISJUNCT precondition:[{ConfigurationNode predicateLabel:\"B0\"}] postcondition:[{ConfigurationNode predicateLabel:\"Z\"}]}]}.", Level.DEBUG)
            );
            checkNoOtherActivity();
            assertEquals("[{TaskNode A1: GREEN 1}, {TaskNode B1: UNCOLORED}]"
                +"[{ConfNode A0: GREEN 0}, {ConfNode B0: UNCOLORED}, {ConfNode Z: GREEN 2}][]", getWorkspaceStateAsString(workspace));
        }
    }

    //----------------------------------------------------------------
    public void testConjunctiveRedInitiation() {
        setUpWorkflowManager();

        ConfigurationNode configurationNodeA=new ConfigurationNodeImpl("A");
        ConfigurationNode configurationNodeB=new ConfigurationNodeImpl("B");
        ConfigurationNode configurationNodeZ=new ConfigurationNodeImpl("Z");

        TaskNode taskNodeY=new TaskNodeImpl("Y", Junction.CONJUNCT, CollectionUtil.setFromArray(configurationNodeA, configurationNodeB),CollectionUtil.setFromArray(configurationNodeZ));

        AlphaFunction alphaFunction=new SetAlphaFunction(CollectionUtil.setFromArray(configurationNodeA));
        BetaFunction betaFunction=new NoBetaFunction();
        Set<ConfigurationNode> initialConfigurationNodes=CollectionUtil.setFromArray(configurationNodeA);

        WorkflowManager.Workspace workspace=setUpProblem(alphaFunction, betaFunction, initialConfigurationNodes,
            "subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"A\"},listener)",
            "WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"A\"}.",
            "handleProblemAdded({Workspace alpha:{SetAlphaFunction [{ConfigurationNode predicateLabel:\"A\"}]} beta:{NoBetaFunction} initial:[{ConfigurationNode predicateLabel:\"A\"}]})",
            "[][{ConfNode A: GREEN 0}][]");

        // test red initiation
        m_mockFragmentQueryable.getLastListener().handleFragment(makeCompositeWorkflow(taskNodeY));
        assertEquals("subscribeFragmentsWithConfiguration({ConfigurationNode predicateLabel:\"B\"},listener)", m_mockFragmentQueryable.getActivityRecordAndReset());
        assertEquals("taskNodeAdded({TaskNode Y})"
            +"nodeProperyChanged({TaskNode Y})"
            +"nodeProperyChanged({ConfNode B})", m_mockWorkspaceListener.getActivityRecordAndReset());
        m_loggingRecorder.assertMessagesLogged(
            new CheckedMessage("WorkflowManagerImpl: Handling fragment {CompositeWorkflowBuilder [{TaskNode contractLabel:\"Y\" junction:CONJUNCT precondition:[{ConfigurationNode predicateLabel:\"A\"}, {ConfigurationNode predicateLabel:\"B\"}] postcondition:[{ConfigurationNode predicateLabel:\"Z\"}]}]}.", Level.DEBUG),
            new CheckedMessage("WorkflowManagerImpl: Subscribing to {ConfigurationNode predicateLabel:\"B\"}.", Level.DEBUG)
        );
        checkNoOtherActivity();
        assertEquals("[{TaskNode Y: RED}][{ConfNode A: GREEN 0}, {ConfNode B: RED}, {ConfNode Z: UNCOLORED}][]", getWorkspaceStateAsString(workspace));
    }


    //################################################################

    //----------------------------------------------------------------
    private void setUpWorkflowManager() {
        m_mockFragmentQueryable=new MockFragmentQueryable();
        m_mockTaskFeasibilityQueryable=new MockTaskFeasibilityQueryable();
        m_mockAuctionManager=new MockAuctionManager();
        m_workflowManager=new WorkflowManagerImpl(m_mockFragmentQueryable, m_mockTaskFeasibilityQueryable, m_mockAuctionManager, Logger.getLogger(WorkflowManagerImpl.class));
        checkNoOtherActivity();
    }

    //----------------------------------------------------------------
    private void checkNoOtherActivity() {
        assertEquals("", m_mockFragmentQueryable.getActivityRecordAndReset());
        assertEquals("", m_mockTaskFeasibilityQueryable.getActivityRecordAndReset());
        assertEquals("", m_mockAuctionManager.getActivityRecordAndReset());
        if (null!=m_mockWorkflowManagerListener) {
            assertEquals("", m_mockWorkflowManagerListener.getActivityRecordAndReset());
        }
        if (null!=m_mockWorkspaceListener) {
            assertEquals("", m_mockWorkspaceListener.getActivityRecordAndReset());
        }
        m_loggingRecorder.assertNoMessagesLogged();
    }

    //----------------------------------------------------------------
    private String getWorkspaceStateAsString(WorkflowManager.Workspace workspace) {
        Set<TaskNode> taskNodes=new LinkedHashSet<TaskNode>();
        Set<ConfigurationNode> configurationNodes=new LinkedHashSet<ConfigurationNode>();
        Set<Tuple2<Node, Node>> edges=new LinkedHashSet<Tuple2<Node, Node>>();
        configurationNodes.addAll(workspace.getInitialConfigurationNodes());
        for (TaskNode taskNode : workspace.getTaskNodes()) {
            taskNodes.add(taskNode);
            for (ConfigurationNode configurationNode : taskNode.getPreconditions()) {
                configurationNodes.add(configurationNode);
                edges.add(new Tuple2<Node, Node>(configurationNode, taskNode));
            }
            for (ConfigurationNode configurationNode : taskNode.getPostconditions()) {
                configurationNodes.add(configurationNode);
                edges.add(new Tuple2<Node, Node>(taskNode, configurationNode));
            }
        }

        StringBuilder stringBuilder=new StringBuilder();
        List<String> nodes=new ArrayList<String>();
        for (TaskNode taskNode : taskNodes) {
            nodes.add("{TaskNode "+taskNode.getContractLabel()+": "+workspace.getNodeColor(taskNode)
                +(workspace.getNodeColor(taskNode).ordinal()<=WorkflowManager.Workspace.Color.GREEN.ordinal()?" "+workspace.getNodeDistance(taskNode):"")+"}");
        }
        Collections.sort(nodes);
        stringBuilder.append(nodes);

        nodes.clear();
        for (ConfigurationNode configurationNode : configurationNodes) {
            nodes.add("{ConfNode "+configurationNode.getPredicateLabel()+": "+workspace.getNodeColor(configurationNode)
                +(workspace.getNodeColor(configurationNode).ordinal()<=WorkflowManager.Workspace.Color.GREEN.ordinal()?" "+workspace.getNodeDistance(configurationNode):"")+"}");
        }
        Collections.sort(nodes);
        stringBuilder.append(nodes);

        nodes.clear();
        for (Tuple2<Node, Node> edge : edges) {
            if (WorkflowManager.Workspace.Color.BLUE==workspace.getEdgeColor(edge)) {
                nodes.add(MockWorkspaceListener.getThumbprint(edge));
            }
        }
        Collections.sort(nodes);
        stringBuilder.append(nodes);

        return stringBuilder.toString();
    }

    //----------------------------------------------------------------
    private CompositeWorkflow makeCompositeWorkflow(TaskNode... taskNodes) {
        CompositeWorkflowBuilder compositeWorkflowBuilder=new CompositeWorkflowBuilderImpl();
        for (TaskNode taskNode : taskNodes) {
            compositeWorkflowBuilder.addTaskNode(taskNode);
        }
        return compositeWorkflowBuilder;
    }

    //----------------------------------------------------------------
    private WorkflowManager.Workspace setUpProblem(AlphaFunction alphaFunction, BetaFunction betaFunction, Set<ConfigurationNode> initialConfigurationNodes, String sInitialQuery, String sInitialQueryMessage, String sProblemAdded, String sWorkspace) {

        // add problem
        m_workflowManager.addProblem(alphaFunction, betaFunction, initialConfigurationNodes, "workflowID");
        assertEquals(sInitialQuery, m_mockFragmentQueryable.getActivityRecordAndReset());
        m_loggingRecorder.assertMessagesLogged(new CheckedMessage("WorkflowManagerImpl: Problem added.", Level.INFO),
            new CheckedMessage(sInitialQueryMessage, Level.DEBUG));
        checkNoOtherActivity();

        // add workflowmanager listener
        m_mockWorkflowManagerListener=new MockWorkflowManagerListener();
        m_workflowManager.addListener(m_mockWorkflowManagerListener);
        assertEquals(sProblemAdded, m_mockWorkflowManagerListener.getActivityRecordAndReset());
        checkNoOtherActivity();

        // add workspace listener
        WorkflowManager.Workspace workspace=m_mockWorkflowManagerListener.getLastWorkspace();
        m_mockWorkspaceListener=new MockWorkspaceListener();
        workspace.addListener(m_mockWorkspaceListener);
        checkNoOtherActivity();
        assertEquals(sWorkspace, getWorkspaceStateAsString(workspace));
        return workspace;
    }


    //################################################################

    //----------------------------------------------------------------
    private static class MockWorkflowManagerListener extends RecordingMockObject implements WorkflowManager.Listener {

        private WorkflowManager.Workspace m_lastWorkspace;

        public void handleProblemAdded(WorkflowManager.Workspace workspace) {
            recordActivity("handleProblemAdded("+getThumbprint(workspace)+")");
            m_lastWorkspace=workspace;
        }

        public WorkflowManager.Workspace getLastWorkspace() {
            return m_lastWorkspace;
        }

        private String getThumbprint(WorkflowManager.Workspace workspace) {
            return "{Workspace"
                +" alpha:"+workspace.getAlphaFunction()
                +" beta:"+workspace.getBetaFunction()
                +" initial:"+CollectionUtil.formatSetStringSorted(workspace.getInitialConfigurationNodes())
                +"}";
        }
    }

    //----------------------------------------------------------------
    private static class MockWorkspaceListener extends RecordingMockObject implements WorkflowManager.Workspace.Listener {

        public void taskNodeAdded(TaskNode taskNode) {
            recordActivity("taskNodeAdded("+getThumbprint(taskNode)+")");
        }

        public void nodeProperyChanged(Node node) {
            recordActivity("nodeProperyChanged("+getThumbprint(node)+")");
        }

        public void edgeProperyChanged(Tuple2<Node, Node> edge) {
            recordActivity("edgeProperyChanged("+getThumbprint(edge)+")");
        }

        //############################################################

        public static String getThumbprint(Tuple2<Node, Node> edge) {
            return "{Edge "+getThumbprint(edge.getA())+" -> "+getThumbprint(edge.getB())+"}";
        }

        public  static String getThumbprint(TaskNode taskNode) {
            return "{TaskNode "+taskNode.getContractLabel()+"}";
        }

        public  static String getThumbprint(ConfigurationNode configurationNode) {
            return "{ConfNode "+configurationNode.getPredicateLabel()+"}";
        }

        public  static String getThumbprint(Node node) {
            if (node instanceof TaskNode) {
                return getThumbprint((TaskNode)node);
            } else {
                return getThumbprint((ConfigurationNode)node);
            }
        }
    }
}